{------------------------------------------------------------------------------}
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2005-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit ASN1Tools;

interface

uses
  pgpBase, PGPStrTools, KeyPropTypes, X509Types;

type
  TX509CertASN1Parser = Class
  private
    FCertData: PAnsiChar;
    FCertSize: Cardinal;
    FCertVersion: Cardinal;
    FIssuerCertID: RawByteString;
    FIssuerHashAlgorithm: THashAlgorithm;
    FIssuerKeyAlgorithm: TKeyAlgorithm;
    FKeyUsages: TX509KeyUsages;
  public
    procedure ParseCertificate(ASNData: PAnsiChar; Size: Cardinal);
    property CertVersion: Cardinal read FCertVersion;
    property IssuerCertID: RawByteString read FIssuerCertID;
    property IssuerHashAlgorithm: THashAlgorithm read FIssuerHashAlgorithm;
    property IssuerKeyAlgorithm: TKeyAlgorithm read FIssuerKeyAlgorithm;
    property KeyUsages: TX509KeyUsages read FKeyUsages;
  end;

function GetSerialNumber(ASNData: PAnsiChar; Size: Cardinal): RawByteString;
function GetKeyUsage(ASNData: PAnsiChar; Size, Pos: Cardinal): Word;


implementation

uses
  Windows, SysUtils;

// Basic IDs -------------------------------------------------------------------

const
  idINT		= #2;
  idOID		= #6;
  lenOID	= #3;
  idTRUE	= #$01#$01#$FF;


// Object IDs ------------------------------------------------------------------

const
  oidVersion		= #$A0;
  oidExtensions		= #$A3;
  oidDSAFilter		= #$2A#$86#$48#$CE#$38#$04;
  oidRSAFilter		= #$2A#$86#$48#$86#$F7#$0D#$01#$01;
  oidSerialNumber	= #$55#$04#$05#$13;
  oidKeyUsages		= #$55#$1D#$0F;
  oidBasicConstraints	= #$55#$1D#$13;


// Data retrieval (generic) ----------------------------------------------------

function BitSwapByte(B: Byte): Byte; register; assembler;
asm
  XOR	DL,DL

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  SHL	AL,1
  ADC	DL,0
  ROR	DL,1

  MOV	AL,DL
end;

function IDCheck(ID, ASNData: PAnsiChar; Pos: Cardinal): Longbool; register; assembler;
asm	// EAX = @ID, EDX = @ASNData, ECX = Pos
  OR	EAX,EAX
  JE	@ERROR
  OR	EDX,EDX
  JE	@ERROR
  CMP	BYTE PTR [EAX],0
  JE	@ERROR
  ADD	EDX,ECX
  CMP	BYTE PTR [EDX],0
  JE	@ERROR
  PUSH	EBX
  DEC	EAX
  DEC	EDX
  XOR	EBX,EBX
  @LOOP:
  INC	EBX
  JO	@ERROR
  MOV	CL,[EAX + EBX]
  OR	CL,CL
  JE	@END
  CMP	[EDX + EBX],CL
  JE	@LOOP
  POP	EBX
  @ERROR:
  XOR	EAX,EAX
  RET
  @END:
  POP	EBX
  MOV	EAX,TRUE
end;

function GetTypeLength(TaggedData: PAnsiChar; Size: Cardinal; var Pos: Cardinal): Cardinal;
var
  SizeLen, Index: Cardinal;
begin
  Result := 0;
  inc(Pos);
  if Pos < Size then begin
    if Ord(TaggedData[Pos]) < $80 then
      Result := Ord(TaggedData[Pos])
    else begin
      SizeLen := Ord(TaggedData[Pos]) and $7F;
      if (SizeLen <= 4) and (Pos + SizeLen < Size) then begin
	for Index := 1 to SizeLen do begin
	  inc(Pos);
	  Result := Result shl 8 + Ord(TaggedData[Pos])
	end;
      end;
    end;
    inc(Pos);
  end;
end;

function FindObjectID(OID, ASNData: PAnsiChar; Size: Cardinal; var Pos: Cardinal): Cardinal;
var
  OIDLen: Cardinal;
begin
  Result := 0;
  OIDLen := StrLen(OID);
  while Pos + OIDLen <= Size do begin
    if IDCheck(OID, ASNData, Pos) then begin
      Result := GetTypeLength(ASNData, Size, Pos);
      Break;
    end
    else if ASNData[Pos] < #$30 then begin
      Result := GetTypeLength(ASNData, Size, Pos);
      inc(Pos, Result);
      Continue;
    end;
    Result := GetTypeLength(ASNData, Size, Pos);
  end;
  if Pos >= Size then Result := 0;
end;

function GetBitStream(ASNData: PAnsiChar; Size, Len: Cardinal; var Pos: Cardinal): Cardinal;
var
  LastPos, PadBits: Cardinal;
begin
  Result := 0;
  LastPos := Pos + pred(Len);
  if LastPos > Size then Exit;
  PadBits := Ord(ASNData[Pos]);
  Result := BitSwapByte(Ord(ASNData[LastPos])) and ($FF shr PadBits);
  dec(LastPos);
  while LastPos > Pos do begin
    Result := Result shl 8 + BitSwapByte(Ord(ASNData[LastPos]));
    dec(LastPos);
  end;
  inc(Pos, Len);
end;


// Data retrieval (specific) ---------------------------------------------------

function GetSerialNumberValue(ASNData: PAnsiChar; Size: Cardinal; var Pos: Cardinal): RawByteString;
var
  Len: Integer;
begin
  Result := '';
  Len := GetTypeLength(ASNData, Size, Pos);
  if Len > 0 then begin
    SetString(Result, NIL, Len);
    Move(ASNData[Pos], Pointer(Result)^, Len);
  end;
  inc(Pos, Len);
end;

function GetKeyUsageValue(ASNData: PAnsiChar; Size: Cardinal; var Pos: Cardinal): Cardinal;
var
  Len: Cardinal;
begin
  Result := 0;
  if IDCheck(idTRUE, ASNData, Pos) then inc(Pos, Length(idTRUE));
  GetTypeLength(ASNData, Size, Pos);
  Len := GetTypeLength(ASNData, Size, Pos);
  if Len < 4 then Result := GetBitStream(ASNData, Size, Len, Pos);
end;

function GetOwnerIsCAValue(ASNData: PAnsiChar; Len: Cardinal; var Pos: Cardinal): Longbool;
begin
  Result := IDCheck(idTRUE, ASNData, Pos);
  inc(Pos, Len);
end;


// Stand-alone functions -------------------------------------------------------

function GetSerialNumber(ASNData: PAnsiChar; Size: Cardinal): RawByteString;
const
  idSerialNumber = idOID + lenOID + oidSerialNumber;
var
  Pos, Len: Cardinal;
begin
  Pos := 0;
  Result := '';
  Len := FindObjectID(idSerialNumber, ASNData, Size, Pos);
  if Len > 0 then begin
    inc(Pos, Len);
    Result := GetSerialNumberValue(ASNData, Size, Pos);
  end;
end;

function GetKeyUsage(ASNData: PAnsiChar; Size: Cardinal; Pos: Cardinal): Word;
var
  Len: Cardinal;
  OwnerIsCA: Longbool;
begin
  Result := 0;
  OwnerIsCA := FALSE;
  try
    Len := FindObjectID(oidExtensions, ASNData, Size, Pos);
    if Len = 0 then Exit;
    repeat
      Len := FindObjectID(idOID, ASNData, Size, Pos);
      case Len of
	0: Break;
	3: begin
	  if IDCheck(oidKeyUsages, ASNData, Pos) then begin
	    inc(Pos, Len);
	    Result := GetKeyUsageValue(ASNData, Size, Pos);
	  end
	  else if IDCheck(oidBasicConstraints, ASNData, Pos) then begin
	    inc(Pos, Len);
	    OwnerIsCA := IDCheck(idTRUE, ASNData, Pos);
	    inc(Pos, Len);
	  end
	  else inc(Pos, Len);
	end;
      else
	inc(Pos, Len);
      end;
    until Pos >= Size;
  finally
    if Result = 0 then begin
      Result := x509KeyUsage_digitalSignature or x509KeyUsage_nonRepudiation or
		x509KeyUsage_keyEncipherment or x509KeyUsage_dataEncipherment or
		x509KeyUsage_keyAgreement or Ord(OwnerIsCA = TRUE) * (x509KeyUsage_keyCertSign or x509KeyUsage_cRLSign);
    end;
  end;
end;


// TX509CertASN1Parser ---------------------------------------------------------

procedure TX509CertASN1Parser.ParseCertificate(ASNData: PAnsiChar; Size: Cardinal);

  function GetIssuerAlgorithms(Pos, Len: Cardinal): Integer;
  const // DSA values
    lenDSAFilter	=  7;
    idDSA		=  2;
    idSHA1WithDSA	=  3;
  const // RSA values
    lenRSAFilter	=  9;
    idRSA		=  1;
    idMD2WithRSA	=  2;
    idMD5WithRSA	=  4;
    idSHA1WithRSA	=  5;
    idSHA256WithRSA	= 11;
    idSHA384WithRSA	= 12;
    idSHA512WithRSA	= 13;
  begin
    FIssuerHashAlgorithm := HashAlgorithm_Invalid;
    FIssuerKeyAlgorithm := KeyAlgorithm_Invalid;
    Result := 0;
    case Len of
      lenDSAFilter: if IDCheck(oidDSAFilter, FCertData, Pos) then begin
	Result := Ord(FCertData[Pos + pred(Len)]);
	case Result of
	  idSHA1WithDSA: FIssuerHashAlgorithm := HashAlgorithm_SHA;
	end;
	FIssuerKeyAlgorithm := KeyAlgorithm_DSS;
      end;
      lenRSAFilter: if IDCheck(oidRSAFilter, FCertData, Pos) then begin
	Result := Ord(FCertData[Pos + pred(Len)]);
	case Result of
	  idMD2WithRSA: FIssuerHashAlgorithm := HashAlgorithm_MD2;
	  idMD5WithRSA: FIssuerHashAlgorithm := HashAlgorithm_MD5;
	  idSHA1WithRSA: FIssuerHashAlgorithm := HashAlgorithm_SHA;
	  idSHA256WithRSA: FIssuerHashAlgorithm := HashAlgorithm_SHA256;
	  idSHA384WithRSA: FIssuerHashAlgorithm := HashAlgorithm_SHA384;
	  idSHA512WithRSA: FIssuerHashAlgorithm := HashAlgorithm_SHA512;
	end;
	FIssuerKeyAlgorithm := KeyAlgorithm_RSA;
      end;
    end;
  end;

var
  Pos, Len: Cardinal;
begin
  FCertData := ASNData;
  FCertSize := Size;
  Pos := 0;
  FindObjectID(oidVersion, FCertData, FCertSize, Pos);
  Len := FindObjectID(idINT, FCertData, FCertSize, Pos);
  if Len > 0 then FCertVersion := succ(Cardinal(FCertData[Pos]));
  inc(Pos, Len);
  Len := FindObjectID(idINT, FCertData, FCertSize, Pos);
  if Len > 0 then FIssuerCertID := BytesToString(@FCertData[Pos], Len);
  inc(Pos, Len);
  Len := FindObjectID(idOID, FCertData, FCertSize, Pos);
  if Len > 0 then GetIssuerAlgorithms(Pos, Len);
  inc(Pos, Len);
  FKeyUsages := GetKeyUsageSet(GetKeyUsage(FCertData, FCertSize, Pos));
end;


end.

