{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit PGPDecode;

interface

uses
  Windows, Classes, SysUtils, UTF8, PGPStrTools, PGPStringList, PSMimeTools, KeyPropTypes, X509Types,
  PrefFuncs, KeyFuncs, X509Funcs, PGPDialogs, pgpBase, pgpErrors, pgpPubTypes, pgpUtilities,
  pgpOptionList, pgpRandomPool, pgpMemoryMgr, pgpEvents, pgpKeys, pgpTLS, pgpUI, pgpSC;

type
  TSigStatus = (
    SIGNED_NOT,
    SIGNED_BAD,
    SIGNED_GOOD,
    SIGNED_NO_KEY,
    SIGNED_UNKNOWN_ALG,
    SIGNED_SIG_EXPIRED
  );
  TSigPropsRec = Record
    sStatus: TSigStatus;
    sHexID: AnsiString;
    sUserID: AnsiString;
    sSubKeyHexID: AnsiString;
    sCreaTimeStr: AnsiString;
    sCreaTimeNum: DWord;
    sChecked: Longbool;
    sVerified: Longbool;
    sValidity: TValidityLevel;
    sRevoked: Longbool;
    sDisabled: Longbool;
    sExpired: Longbool;
    sDetached: Longbool;
    sAxiomatic: Longbool;
    sExpTimeNum: Longint;
    sExpTimeStr: AnsiString;
    sHashAlgorithm: THashAlgorithm;
    sValidCertChain: TX509ChainValidity;
  end;
  TAnalysis = (
    ENCRYPTED,
    SIGNED,
    DETACHED_SIGNATURE,
    KEY,
    UNKNOWN,
    X509_CERTIFICATE,
    X509_SECRETKEY,
    SMIME_UNRECOGNIZED,
    SMIME_ENCRYPTED,
    SMIME_SIGNED
  );
  TPassThroughOption = (
    PassThrough_Keys,
    PassThrough_ClearSigned,
    PassThrough_Unrecognized
  );
  TPassOption = (
    AnalyseOnly,
    AnalysePass,
    DecodePass
  );
  TX509Mode = (
    X509Off,
    X509Secret,
    X509Binary,
    X509Armored,
    X509IsMime,
    X509Signed,
    X509Encrypted,
    X509Processed
  );
  TPassThroughOptions = Set of TPassThroughOption;
  TOnAnalysis = procedure(SectionNumber: Longint; SectionType: TAnalysis; var SkipSection: Longbool) of Object;
  TOnGetInputFileName = procedure(var SuggestedName: AnsiString) of Object;
  TOnGetOutputFileName = procedure(var SuggestedName: AnsiString) of Object;
  TOnGetSignedFileName = procedure(var SuggestedName: AnsiString) of Object;
  TOnEnterPassphrase = procedure(const Passphrase: PAnsiChar;
				 const DecryptionKeyList: TKeyPropsList;
				 Conventional, BadPassphrase: Longbool;
				 var Cancel: Longbool) of Object;
  TOnGetSignature = procedure(const SigProps: TSigPropsRec) of Object;
  TOnShowProgress = procedure(BytesProcessed, BytesTotal: Longint) of Object;
  TOnWipePassphrase = procedure(const Passphrase: PAnsiChar) of Object;
  TPGPDecodeCustom = class(TComponent)
  private
    // internal
    FContext: pPGPContext;
    FKeySetMain: pPGPKeySet;
    FAllocatedOutputBuffer: PAnsiChar;
    FActualOutputSize: PGPSize;
    FOutputOptions: pPGPOptionList;
    FInputSize: PGPSize;
    FInputBuffer: PAnsiChar;
    FInputFileName: AnsiString;
    FFileNameBuffer: RawByteString;
    FAnalysis: TAnalysis;
    FClearSigned: Longbool;
    FEyesOnly: Longbool;
    FFileToFile: Longbool;
    FOutputExists: Longbool;
    FIsPGPMime: Longbool;
    FX509Mode: TX509Mode;
    FRecognized: Longbool;
    FRecurring: Longbool;
    FPassOption: TPassOption;
    FPassphrase: PAnsiChar;
    FPassSection: PGPUInt32;
    FRecipientsData: TPGPEventRecipientsData;
    FSectionNumber: PGPUInt32;
    FSignedDataBuffer: RawByteString;
    FDecryptionKey: pPGPKey;
    FSecureBuffer: PAnsiChar;
    FSecureBufferSize: PGPSize;
    FSignCertProps: TX509CertPropsRec;
    FValidCertChain: TX509ChainValidity;
    // protected properties
    FOutputBuffer: RawByteString;
    FFileOutput: Longbool;
    FAlwaysQueryOutputFileName: Longbool;
    FDetachedSigVerify: Longbool;
    FIgnoreKnownFlag: Longint;
    FSigPropsRec: TSigPropsRec;
    FKeyPropsList: TKeyPropsList;
    // public properties
    FCharset: AnsiString;
    FKeyAddResult: Longint;
    FParentHandle: THandle;
    // published properties
    FEnforceFileStreaming: Longbool;
    FGetKeyFromServer: Longbool;
    FLogBadSignature: AnsiString;
    FLogExpSignature: AnsiString;
    FLogKeyDisabled: AnsiString;
    FLogKeyExpired: AnsiString;
    FLogKeyRevoked: AnsiString;
    FLogKeyUnknown: AnsiString;
    FLogUnknownAlg: AnsiString;
    FPassCacheSeconds: TPassCacheSeconds;
    FPassCacheShare : Longbool;
    FPassThroughOptions: TPassThroughOptions;
    FProgressInterval: Cardinal;
    FQueryAddKeys: Longbool;
    FRecursivelyDecode: Longbool;
    FShowSigLog: Longbool;
    FKeyDlgPrompt: AnsiString;
    FPassDlgPrompt: AnsiString;
    FOutputFileName: AnsiString;
    FSignedFileName: AnsiString;
    // events
    FOnAnalysis: TOnAnalysis;
    FOnGetInputFileName: TOnGetInputFileName;
    FOnGetOutputFileName: TOnGetOutputFileName;
    FOnGetSignedFileName: TOnGetSignedFileName;
    FOnEnterPassphrase: TOnEnterPassphrase;
    FOnGetSignature: TOnGetSignature;
    FOnShowProgress: TOnShowProgress;
    FOnWipePassphrase: TOnWipePassphrase;
    procedure ClearSigPropsRec;
    procedure ClearSignCertProps;
    function  InitDecode: PGPError;
    procedure FinitDecode;
    function  HandleFileInput: PGPError;
    function  X509Test: TX509Mode;
    function  SecureMove: PGPError;
    function  ShowSigInfo: PGPError;
    function  MaxFileSize: DWord;
    function  FileSize(const FileName: RawByteString; var Bytes: DWord): PGPError;
    function  ReadInputFile: PGPError;
    function  GetOutputFileName: RawByteString;
    function  Analyze(InFileSpec: pPGPFileSpec): PGPError;
    function  FixNonAsciiBug(const Data: PAnsiChar): PGPError;
    function  GetOptionList(var OptionList: pPGPOptionList): PGPError;
    function  SetOutputOption(var OptionList: pPGPOptionList): PGPError;
    function  PrepareOutput(const Input: PAnsiChar; InputSize: DWord; const MimeHeaders, MimeBody: RawByteString): PGPError;
    function  DoRecursion(const Input: PAnsiChar; InputSize: DWord): PGPError;
    function  SecureView(const Output: PAnsiChar; OutputSize: DWord): PGPError;
    function  WriteOutputFile(const Output: PAnsiChar; OutputSize: DWord): PGPError;
    function  Decode(IsFile: Longbool): Longint;
    function  AnalyzeHandler(Event: pPGPEvent): PGPError;
    function  RecipientsHandler(Event: pPGPEvent): PGPError;
    function  PassphraseHandler(Event: pPGPEvent): PGPError;
    function  KeyFoundHandler(Event: pPGPEvent): PGPError;
    function  OutputHandler(Event: pPGPEvent): PGPError;
    function  SignatureHandler(Event: pPGPEvent): PGPError;
    function  EndLexHandler(Event: pPGPEvent): PGPError;
    function  DecryptionHandler(Event: pPGPEvent): PGPError;
  protected
    property OutputBuffer: RawByteString
      read FOutputBuffer;
    property FileOutput: Longbool
      write FFileOutput;
    property AlwaysQueryOutputFileName: Longbool
      write FAlwaysQueryOutputFileName;
    property DetachedSignature: Longbool
      read FDetachedSigVerify;
    property IgnoreKnownFlag: Longint
      write FIgnoreKnownFlag;
    property SigPropsRec: TSigPropsRec
      read FSigPropsRec;
    property KeyPropsList: TKeyPropsList
      read FKeyPropsList;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AnalyseBuffer(const DataBuffer: RawByteString): Longint; virtual;
    function AnalyseFile(const FileName: RawByteString): Longint; virtual;
    function AnalyseMimeBuffer(const DataBuffer: RawByteString): Longint; virtual;
    function AnalyseMimeFile(const FileName: RawByteString): Longint; virtual;
    function DecodeBuffer(const DataBuffer: RawByteString): Longint; virtual;
    function DecodeFile(const FileName: RawByteString): Longint; virtual;
    function DecodeMimeBuffer(const DataBuffer: RawByteString): Longint; virtual;
    function DecodeMimeFile(const FileName: RawByteString): Longint; virtual;
    property Charset: AnsiString
      read FCharset;
    property KeyAddResult: Longint
      read FKeyAddResult;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property EnforceFileStreaming: Longbool
      read FEnforceFileStreaming
      write FEnforceFileStreaming;
    property GetKeyFromServer: Longbool
      read FGetKeyFromServer
      write FGetKeyFromServer;
    property LogBadSignature: AnsiString
      read FLogBadSignature
      write FLogBadSignature;
    property LogExpSignature: AnsiString
      read FLogExpSignature
      write FLogExpSignature;
    property LogKeyDisabled: AnsiString
      read FLogKeyDisabled
      write FLogKeyDisabled;
    property LogKeyExpired: AnsiString
      read FLogKeyExpired
      write FLogKeyExpired;
    property LogKeyRevoked: AnsiString
      read FLogKeyRevoked
      write FLogKeyRevoked;
    property LogKeyUnknown: AnsiString
      read FLogKeyUnknown
      write FLogKeyUnknown;
    property LogUnknownAlg: AnsiString
      read FLogUnknownAlg
      write FLogUnknownAlg;
    property PassCacheSeconds: TPassCacheSeconds
      read FPassCacheSeconds
      write FPassCacheSeconds;
    property PassCacheShare : Longbool
      read FPassCacheShare
      write FPassCacheShare;
    property PassThroughOptions: TPassThroughOptions
      read FPassThroughOptions
      write FPassThroughOptions;
    property ProgressInterval: Cardinal
      read FProgressInterval
      write FProgressInterval;
    property RecursivelyDecode: Longbool
      read FRecursivelyDecode
      write FRecursivelyDecode;
    property QueryAddKeys: Longbool
      read FQueryAddKeys
      write FQueryAddKeys;
    property ShowSigLog: Longbool
      read FShowSigLog
      write FShowSigLog;
    property KeyDlgPrompt: AnsiString
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property PassDlgPrompt: AnsiString
      read FPassDlgPrompt
      write FPassDlgPrompt;
    property OutputFileName: AnsiString
      read FOutputFileName
      write FOutputFileName;
    property SignedFileName: AnsiString
      read FSignedFileName
      write FSignedFileName;
    property OnAnalysis: TOnAnalysis
      read FOnAnalysis
      write FOnAnalysis;
    property OnGetInputFileName: TOnGetInputFileName
      read FOnGetInputFileName
      write FOnGetInputFileName;
    property OnGetSignedFileName: TOnGetSignedFileName
      read FOnGetSignedFileName
      write FOnGetSignedFileName;
    property OnGetOutputFileName: TOnGetOutputFileName
      read FOnGetOutputFileName
      write FOnGetOutputFileName;
    property OnEnterPassphrase: TOnEnterPassphrase
      read FOnEnterPassphrase
      write FOnEnterPassphrase;
    property OnGetSignature: TOnGetSignature
      read FOnGetSignature
      write FOnGetSignature;
    property OnShowProgress: TOnShowProgress
      read FOnShowProgress
      write FOnShowProgress;
    property OnWipePassphrase: TOnWipePassphrase
      read FOnWipePassphrase
      write FOnWipePassphrase;
  end;


implementation

{$IFDEF UNICODE}
uses
  AnsiStrings;
{$ENDIF}

function DataCopy(var Output: PAnsiChar; const Input: PAnsiChar; InputSize: DWord): PGPError;
begin
  Result := kPGPError_OutOfMemory;
  if InputSize > 0 then begin
    Output := PGPNewData(PGPGetDefaultMemoryMgr, InputSize, kPGPMemoryMgrFlags_None);
    if Output <> NIL then begin
      Move(Input[0], Output[0], InputSize);
      Result := kPGPError_NoErr;
    end;
  end;
end;

function EventHandler(Context: pPGPContext; Event: pPGPEvent; UserValue: PGPUserValue): PGPError; cdecl;
begin
  Result := 0;
  with TPGPDecodeCustom(UserValue) do begin
    case Event^.EType of
      kPGPEvent_NullEvent:		if (FPassOption = DecodePass)
					and not (FDetachedSigVerify or FClearSigned) then begin
					  if TMethod(FOnShowProgress).Code <> NIL then begin
					    // BytesTotal always stays 0 => use FInputSize (at least in 6.5.X)
					    with Event^.EData.NullData do FOnShowProgress(BytesWritten, FInputSize);
					  end;
					  ProcessMessages;
					end;
      kPGPEvent_InitialEvent:		;
      kPGPEvent_FinalEvent:		;
      kPGPEvent_ErrorEvent:	  	Result := Event^.EData.ErrorData.Error;
      kPGPEvent_WarningEvent:		;
      kPGPEvent_PassphraseEvent:	Result := PassphraseHandler(Event);
      kPGPEvent_AnalyzeEvent:		Result := AnalyzeHandler(Event);
      kPGPEvent_RecipientsEvent:  	if FPassOption = DecodePass then Result := RecipientsHandler(Event);
      kPGPEvent_KeyFoundEvent:		if FPassOption = DecodePass then Result := KeyFoundHandler(Event);
      kPGPEvent_OutputEvent:		if FPassOption = DecodePass then OutputHandler(Event);
      kPGPEvent_SignatureEvent:		if FPassOption = DecodePass then Result := SignatureHandler(Event);
      kPGPEvent_BeginLexEvent:	  	FSectionNumber := succ(Event^.EData.BeginLexData.SectionNumber);
      kPGPEvent_EndLexEvent:      	Result := EndLexHandler(Event);
      kPGPEvent_DetachedSignatureEvent:	;
      kPGPEvent_DecryptionEvent:	Result := DecryptionHandler(Event);
    end;
  end;
end;

constructor TPGPDecodeCustom.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FKeyPropsList := TKeyPropsList.Create(spgpKeyPropFlag_IDFlags);
  FLogBadSignature := 'bad signature';
  FLogExpSignature := 'expired signature';
  FLogKeyDisabled := 'disabled key';
  FLogKeyExpired := 'expired key';
  FLogKeyRevoked := 'revoked key';
  FLogKeyUnknown := 'unknown key';
  FLogUnknownAlg := 'unknown algorithm';
  FProgressInterval := 1000;
end;

destructor TPGPDecodeCustom.Destroy;
begin
  FKeyPropsList.Free;
  inherited Destroy;
end;

procedure TPGPDecodeCustom.ClearSigPropsRec;
begin
  with FSigPropsRec do begin
    sStatus := SIGNED_NOT;
    sHexID := E;
    sUserID := E;
    sSubKeyHexID := E;
    sCreaTimeStr := E;
    sCreaTimeNum := 0;
    sChecked := FALSE;
    sVerified := FALSE;
    sValidity := Validity_Unknown;
    sRevoked := FALSE;
    sDisabled := FALSE;
    sExpired := FALSE;
    sDetached := FALSE;
    sAxiomatic := FALSE;
    sExpTimeNum := 0;
    sExpTimeStr := E;
    sHashAlgorithm := HashAlgorithm_Invalid;
    sValidCertChain := CertChain_Invalid;
  end;
end;

procedure TPGPDecodeCustom.ClearSignCertProps;
begin
  with FSignCertProps do begin
    x509CertificateID := E;
    x509IssuerKeyAlgorithm := KeyAlgorithm_Invalid;
    x509IssuerHashAlgorithm := HashAlgorithm_Invalid;
    x509IssuerLongName := E;
    x509IssuerSerialNumber := E;
    x509OwnerKeyHexID := E;
    x509OwnerCertUserID := E;
    x509OwnerLongName := E;
    x509OwnerSerialNumber := E;
    x509CreaTimeStr := E;
    x509ExpTimeStr := E;
    x509CreaTimeNum := 0;
    x509ExpTimeNum := 0;
    x509IsVerified := FALSE;
    x509IsCorrupt := FALSE;
    x509IsRevoked := FALSE;
    x509IsExpired := FALSE;
    x509IsRoot := FALSE;
  end;
end;

function TPGPDecodeCustom.InitDecode: PGPError;
begin
  FAnalysis := UNKNOWN;
  FSectionNumber := 0;
  FOutputBuffer := E;
  FActualOutputSize := 0;
  FAllocatedOutputBuffer := NIL;
  FSecureBuffer := NIL;
  FSecureBufferSize := 0;
  FSignedDataBuffer := E;
  FClearSigned := FALSE;
  FDetachedSigVerify := FALSE;
  FIsPGPMime := FALSE;
  FRecognized := FALSE;
  FPassSection := 0;
  FPassphrase := NIL;
  FDecryptionKey := NIL;
  if not FRecurring then begin
    FCharset := E;
    FEyesOnly := FALSE;
    FKeyAddResult := 0;
    FKeyPropsList.Clear;
    ClearSignCertProps;
    ClearSigPropsRec;
    FValidCertChain := CertChain_Invalid;
  end;
  FillChar(FRecipientsData, SizeOf(TPGPEventRecipientsData), 0);
  Result := KeyRings.InitKeyRings(FContext, FKeySetMain);
end;

procedure TPGPDecodeCustom.FinitDecode;
begin
  FX509Mode := X509Off;
  if (FKeyPropsList.Count <> 0) or (FAnalysis = X509_CERTIFICATE) then KeyRings.UpdateKeyRings;
  with FRecipientsData do begin
    PGPFreeData(KeyIDArray);
    PGPFreeKeySet(RecipientSet);
  end;
  KeyRings.FreeKeyRings;
end;

function TPGPDecodeCustom.HandleFileInput: PGPError;
begin
  Result := 0;
  FInputFileName := FInputBuffer;
  if not FileExists(FInputFileName) then begin
    if Assigned(FOnGetInputFileName) then begin
      FOnGetInputFileName(FInputFileName);
      if FInputFileName = E then Result := kPGPError_FileNotFound;
    end
    else Result := kPGPError_FileNotFound;
  end;
  if Result <> 0 then Exit;
  Result := FileSize(FInputFileName, FInputSize);
  if (SDKVersion >= PGP7XX) and ((FFileOutput and FEnforceFileStreaming) or (FInputSize > MaxFileSize)) then begin
    if FX509Mode >= X509IsMime then
      Result := kPGPError_FeatureNotAvailable
    else FFileToFile := FFileOutput;
  end
  else begin
    if FFileOutput and FEnforceFileStreaming then
      FFileToFile := TRUE
    else FFileToFile := FALSE;
  end;
end;

function TPGPDecodeCustom.X509Test: TX509Mode;
var
  Pos		: DWord;
  Error		: PGPError;
  OptionList	: pPGPOptionList;
  TestDB	: pPGPKeyDB;
begin
  Result := X509Off;
  if SDKVersion >= PGP81X then begin
    Pos := DWord(StrPos(FInputBuffer, X509Cert_Beg));
    if Pos > 0 then begin
      dec(Pos, DWord(FInputBuffer));
      if ((Pos = 0) or (FInputBuffer[pred(Pos)] = #10)) then Result := X509Armored;
    end;
    if Result = X509Off then begin
      TestDB := NIL;
      try
	// test for PKCS7 format
	OptionList := NIL;
	try
	  Error := PGPBuildOptionList(FContext, OptionList,
	    [
	      PGPOInputFormat(FContext, kPGPInputFormat_X509DataInPKCS7),
	      PGPOInputBuffer(FContext, FInputBuffer, FInputSize)
	    ]);
	  if (Error = 0) and (PGPImport(FContext, TestDB, OptionList, PGPOLastOption(FContext)) = 0) then begin
	    Result := X509Binary;
	  end;
	finally
	  PGPFreeOptionList(OptionList);
	end;
	if Result = X509Off then begin
	  // test for PKCS12 format
	  OptionList := NIL;
	  try
	    Error := PGPBuildOptionList(FContext, OptionList,
	      [
		PGPOInputFormat(FContext, kPGPInputFormat_PKCS12),
		PGPOInputBuffer(FContext, FInputBuffer, FInputSize)
	      ]);
	    if (Error = 0)
	    and (PGPImport(FContext, TestDB, OptionList, PGPOLastOption(FContext)) = kPGPError_BadPassphrase) then begin
	      Result := X509Secret;
	    end;
	  finally
	    PGPFreeOptionList(OptionList);
	  end;
	end;
      finally
	PGPFreeKeyDB(TestDB);
      end;
    end;
  end;
end;

function TPGPDecodeCustom.SecureMove: PGPError;
begin
  Result := kPGPError_NoErr;
  if FActualOutputSize <> 0 then begin
    try
      if FEyesOnly and (FSecureBuffer = NIL) then begin
	Result := kPGPError_OutOfMemory;
	FSecureBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr,
					  FSecureBufferSize + FActualOutputSize,
					  kPGPMemoryMgrFlags_None);
	if FSecureBuffer <> NIL then Result := 0;
      end
      else begin
	if (FSecureBuffer = NIL) and (FAllocatedOutputBuffer <> NIL) then begin
	  FSecureBuffer := FAllocatedOutputBuffer;
	  FSecureBufferSize := FActualOutputSize;
	  Exit;
	end;
	Result := PGPReallocData(PGPGetDefaultMemoryMgr,
				 Pointer(FSecureBuffer),
				 FSecureBufferSize + FActualOutputSize,
				 kPGPMemoryMgrFlags_None);
      end;
      if Result = 0 then begin
	Move(FAllocatedOutputBuffer[0], FSecureBuffer[FSecureBufferSize], FActualOutputSize);
	if FEyesOnly then FillChar(FAllocatedOutputBuffer^, FActualOutputSize, 0);
	inc(FSecureBufferSize, FActualOutputSize);
      end;
    finally
      if FSecureBuffer <> FAllocatedOutputBuffer then PGPFreeData(FAllocatedOutputBuffer);
      FAllocatedOutputBuffer := NIL;
      FActualOutputSize := 0;
    end;
  end;
end;

function TPGPDecodeCustom.ShowSigInfo: PGPError;
var
  Status	: RawByteString;
  Source	: RawByteString;
  Dest		: RawByteString;
  ID		: RawByteString;
  Validity	: RawByteString;
  Info		: RawByteString;
begin
  Result := 0;
  with FSigPropsRec do begin
    if sStatus <> SIGNED_NOT then begin
      // don't change these (constant) PGP strings, see SigEvent.c
      if sStatus = SIGNED_GOOD then
	Status := 'Good Signature' + LF
      else Status := 'Bad Signature' + LF;
      // these strings may be customized
      if FInputFileName <> E then begin
	if FDetachedSigVerify then
	  Source := ExtractFileName(FSignedFileName) + LF
	else if FRecurring then
	  Source := ExtractFileName(FFileNameBuffer) + LF
	else Source := ExtractFileName(FInputFileName) + LF;
	if FFileOutput and (FOutputFileName <> E) then begin
	  Dest := ExtractFileName(FOutputFileName);
	  if Dest <> E then Insert(' => ' + Dest, Source, Length(Source));
	end;
      end
      else Source := '(...)' + LF;
      if (sStatus = SIGNED_NO_KEY) or (sUserID = E) then begin
	if SDKVersion >= PGP8XX then
	  ID := '(' + sHexID + ')' + LF + ShortHexID(sHexID) + LF
	else ID := '(' + ShortHexID(sHexID) + ')' + LF;
      end
      else begin
	if SDKVersion >= PGP8XX then
	  ID := sUserID + LF + ShortHexID(sHexID) + LF
	else ID := sUserID + LF;
      end;
      // don't change these (constant) PGP strings, see SigEvent.c
      if sAxiomatic then
	Validity := 'Implicit Trust' + LF + sCreaTimeStr
      else begin
	case sValidity of
	  Validity_Unknown, Validity_Invalid: Validity := 'Invalid Key' + LF + sCreaTimeStr;
	  Validity_Marginal: Validity := 'Marginal Key' + LF + sCreaTimeStr;
	  Validity_Complete: Validity := 'Valid Key' + LF + sCreaTimeStr;
	end;
      end;
      // these strings may be customized
      if sUserID = E then
	Info := FLogKeyUnknown
      else begin
	if sStatus <> SIGNED_UNKNOWN_ALG then begin
	  if sVerified then begin
	    if sStatus = SIGNED_SIG_EXPIRED then
	      Info := FLogExpSignature
	    else begin
	      if sRevoked then Info := FLogKeyRevoked;
	      if sDisabled then Info := FLogKeyDisabled;
	      if sExpired then Info := FLogKeyExpired;
	    end;
	  end
	  else Info := FLogBadSignature;
	end
	else Info := FLogUnknownAlg;
      end;
      if Info <> E then  Info := ' (' + Info + ')';
      if not SendPGPlogMsg(FParentHandle, Status + Source + ID + Validity + Info) then begin
	Result := kPGPError_FeatureNotAvailable;
      end;
    end;
  end;
end;

function TPGPDecodeCustom.MaxFileSize: DWord;
var
  MemoryStatus	: TMemoryStatus;
begin
  MemoryStatus.dwLength := SizeOf(TMemoryStatus);
  GlobalMemoryStatus(MemoryStatus);
  Result := MemoryStatus.dwAvailPhys shr 2;
end;

function TPGPDecodeCustom.FileSize(const FileName: RawByteString; var Bytes: DWord): PGPError;
var
  FileHandle	: THandle;
  FileSizeHigh	: DWord;
begin
  Result := 0;
  FileHandle := FileOpen(String(FileName), fmOpenRead or fmShareDenyNone);
  if FileHandle <> INVALID_HANDLE_VALUE then begin
    Bytes := GetFileSize(FileHandle, @FileSizeHigh);
    FileClose(FileHandle);
  end
  else Bytes := $FFFFFFFF;
  if (Bytes = 0) or (Bytes = $FFFFFFFF) or (FileSizeHigh <> 0) then Result := kPGPError_ReadFailed;
end;

function TPGPDecodeCustom.ReadInputFile: PGPError;
var
  InputFile	: THandle;
  BytesRead	: DWord;
begin
  FInputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(FInputSize), kPGPMemoryMgrFlags_None);
  if FInputBuffer <> NIL then begin
    FInputBuffer[FInputSize] := #0;
    InputFile := CreateFileA(PAnsiChar(FInputFileName), GENERIC_READ, FILE_SHARE_READ, NIL, OPEN_EXISTING,
			     FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if InputFile <> INVALID_HANDLE_VALUE then begin
      try
	if not ReadFile(InputFile, FInputBuffer[0], FInputSize, BytesRead, NIL) or (BytesRead <> FInputSize) then
	  Result := kPGPError_ReadFailed
	else Result := kPGPError_NoErr;
      finally
	CloseHandle(InputFile);
      end;
    end
    else Result := kPGPError_CantOpenFile;
  end
  else begin
    Result := kPGPError_OutOfMemory;
    Exit;
  end;
end;

function TPGPDecodeCustom.GetOutputFileName: RawByteString;
var
  FileExtension	: RawByteString;
begin
  if FDetachedSigVerify then begin
    if FSignedFileName = E then begin
      FSignedFileName := Trim(FInputFileName);
      FileExtension := LowerCase(ExtractFileExt(FSignedFileName));
      if (FileExtension = '.sig') or (FileExtension = '.asc') then begin
	Delete(FSignedFileName, Length(FSignedFileName) - 3, 4);
      end;
    end;
    if ((FSignedFileName = FInputFileName) or not FileExists(FSignedFileName))
    and Assigned(FOnGetSignedFileName) then begin
      FOnGetSignedFileName(FSignedFileName);
    end;
    Result := FSignedFileName;
  end
  else begin
    if FOutputFileName = E then begin
      FOutputFileName := Trim(FInputFileName);
      FileExtension := LowerCase(ExtractFileExt(FOutputFileName));
      if (FileExtension = '.pgp') or (FileExtension = '.gpg') or (FileExtension = '.asc') then begin
	Delete(FOutputFileName, Length(FOutputFileName) - 3, 4);
      end;
    end;
    if FAlwaysQueryOutputFileName or (FOutputFileName = E)
    or (ExtractFileExt(FOutputFileName) = E) or FileExists(FOutputFileName) then begin
      if Assigned(FOnGetOutputFileName) then
	FOnGetOutputFileName(FOutputFileName)
      else FOutputFileName := E;
    end;
    Result := FOutputFileName;
    if Result <> E then DeleteFile(Result);
  end;
end;

function TPGPDecodeCustom.Analyze(InFileSpec: pPGPFileSpec): PGPError;
var
  OptionList	: pPGPOptionList;
begin
  OptionList := NIL;
  Result := PGPBuildOptionList(FContext, OptionList,
    [
      PGPODiscardOutput(FContext, PGPTrue),
      PGPOEventHandler(FContext, EventHandler, Self)
    ]);
  if Result <> 0 then Exit;
  if InFileSpec <> NIL then begin
    Result := PGPAppendOptionList(OptionList, [PGPOInputFile(FContext, InFileSpec)]);
  end
  else if FInputSize <> 0 then begin
    Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(FContext, FInputBuffer, FInputSize)]);
  end;
  if Result <> 0 then Exit;
  try
    Result := pgpEvents.PGPDecode(FContext, OptionList, PGPOLastOption(FContext));
    if Result = kPGPError_UserAbort then Result := 0;
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function TPGPDecodeCustom.FixNonAsciiBug(const Data: PAnsiChar): PGPError;
begin
  // try to circumvent errors with some non-ASCII data
  inc(FInputSize);
  if FEyesOnly then
    FInputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, FInputSize + 2, kPGPMemoryMgrFlags_None)
  else FInputBuffer := PGPNewData(PGPGetDefaultMemoryMgr, FInputSize + 2, kPGPMemoryMgrFlags_None);
  if FInputBuffer <> NIL then begin
    Move(Data[0], FInputBuffer[1], FInputSize);
    FInputBuffer[succ(FInputSize)] := #0;
    FInputBuffer[0] := '@';
    Result := 0;
  end
  else Result := kPGPError_OutOfMemory;
end;

function TPGPDecodeCustom.GetOptionList(var OptionList: pPGPOptionList): PGPError;
var
  X509Option	: pPGPOptionList;
begin
  Result := PGPBuildOptionList(FContext, OptionList,
    [
      PGPOKeySetRef(FContext, FKeySetMain),
      PGPOPassThroughKeys(FContext, PGPBoolean(PassThrough_Keys in PassThroughOptions)),
      // PGPOPassThroughClearSigned(FContext, PGPBoolean(PassThrough_ClearSigned in PassThroughOptions)), // buggy
      PGPOPassThroughIfUnrecognized(FContext, PGPBoolean(PassThrough_Unrecognized in PassThroughOptions)),
      PGPORecursivelyDecode(FContext, PGPBoolean(FRecursivelyDecode = TRUE)),
      PGPOSendEventIfKeyFound(FContext, PGPTrue),
      PGPOEventHandler(FContext, EventHandler, Self)
    ]);
  if Result <> 0 then Exit;
  case FX509Mode of
    X509Secret: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_PKCS12);
    X509Binary: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_X509DataInPKCS7);
    X509Armored: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_PEMEncodedX509Cert);
    X509IsMime..X509Encrypted: X509Option := PGPOInputFormat(FContext, kPGPInputFormat_SMIMEBody);
  else
    // doesn't work properly with S/MIME
    Result := PGPAppendOptionList(OptionList, [PGPOSendNullEvents(FContext, FProgressInterval)]);
    Exit;
  end;
  if FX509Mode in [X509IsMime..X509Encrypted] then begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOSMIMEMatchCriterion(FContext, kPGPSMIMEMatchCriterion_Any),
	X509Option
      ]);
  end;
end;

function TPGPDecodeCustom.SetOutputOption(var OptionList: pPGPOptionList): PGPError;
var
  FileSpec	: pPGPFileSpec;
begin
  if FDetachedSigVerify then begin
    FFileToFile := FALSE;
    if GetOutputFileName <> E then begin
      FileSpec := NIL;
      Result := PGPNewFileSpecFromFullPath(FContext, PAnsiChar(FSignedFileName), FileSpec);
      if Result <> 0 then Exit;
      try
	Result := PGPAppendOptionList(OptionList,
	  [
	    PGPODetachedSig(FContext, PGPOInputFile(FContext, FileSpec), PGPOLastOption(FContext))
	  ]);
      finally
	PGPFreeFileSpec(FileSpec);
      end;
    end
    else Result := kPGPError_DetachedSignatureFound;
  end
  else if FX509Mode = X509IsMime then begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPODetachedSig(FContext,
			PGPOInputBuffer(FContext, PAnsiChar(FSignedDataBuffer), Length(FSignedDataBuffer)),
			PGPOLastOption(FContext))
      ]);
  end
  else begin
    if FFileOutput then begin
      if GetOutputFileName = E then begin
	Result := kPGPError_CantOpenFile;
	Exit;
      end;
    end;
    if FFileToFile then begin
      FileSpec := NIL;
      Result := PGPNewFileSpecFromFullPath(FContext, PAnsiChar(FOutputFileName), FileSpec);
      try
	if Result = 0 then begin
	  Result := PGPAppendOptionList(OptionList,
	    [
	      PGPOOutputFile(FContext, FileSpec),
	      PGPOAppendOutput(FContext, PGPTrue)
	    ]);
	end;
      finally
	PGPFreeFileSpec(FileSpec);
      end;
    end
    else begin
      if FClearSigned and (PassThrough_ClearSigned in PassThroughOptions) then
	Result := PGPAppendOptionList(OptionList, [PGPODiscardOutput(FContext, PGPTrue)]) // for bugfix
      else begin
	Result := PGPAppendOptionList(OptionList,
	  [
	    PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
	  ]);
      end;
    end;
  end;
end;

function TPGPDecodeCustom.PrepareOutput(const Input: PAnsiChar; InputSize: DWord; const MimeHeaders, MimeBody: RawByteString): PGPError;
var
  HeaderSize	: DWord;
begin
  Result := 0;
  if not FFileToFile then begin
    if not (FDetachedSigVerify or FClearSigned)
    or (FIsPGPMime and not (PassThrough_ClearSigned in PassThroughOptions)) then begin
      Result := SecureMove;
      if Result <> 0 then Exit;
      HeaderSize := Length(MimeHeaders);
      if (FIsPGPMime or (FX509Mode = X509Encrypted)) and (HeaderSize > 0) then begin
	FActualOutputSize := HeaderSize + FSecureBufferSize;
	FAllocatedOutputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, FActualOutputSize, kPGPMemoryMgrFlags_None);
	if FAllocatedOutputBuffer <> NIL then begin
	  Move(MimeHeaders[1], FAllocatedOutputBuffer[0], HeaderSize);
	  Move(FSecureBuffer[0], FAllocatedOutputBuffer[HeaderSize], FSecureBufferSize);
	  PGPFreeData(FSecureBuffer);
	  FSecureBuffer := NIL;
	  FSecureBufferSize := 0;
	end
	else Result := kPGPError_OutOfMemory;
      end
      else begin
	FAllocatedOutputBuffer := FSecureBuffer;
	FActualOutputSize := FSecureBufferSize;
      end;
    end;
    // fixes for signed data
    if (FClearSigned or (FX509Mode = X509IsMime))
    and (PassThrough_ClearSigned in PassThroughOptions) then begin
      FActualOutputSize := InputSize;
      Result := DataCopy(FAllocatedOutputBuffer, Input, FActualOutputSize);
    end
    else if (FX509Mode in [X509IsMime..X509Signed]) and (FActualOutputSize = 0) then begin
      FActualOutputSize := Length(MimeHeaders) + Length(MimeBody);
      Result := DataCopy(FAllocatedOutputBuffer, PAnsiChar(MimeHeaders + MimeBody), FActualOutputSize);
    end;
  end;
end;

function TPGPDecodeCustom.DoRecursion(const Input: PAnsiChar; InputSize: DWord): PGPError;
begin
  FFileNameBuffer := FInputFileName;
  if FFileToFile then begin
    SetString(FInputFileName, NIL, MAX_PATH);
    if GetTempFileNameA(PAnsiChar(ExtractFilePath(FOutputFileName)), 'PGP', 0, PAnsiChar(FInputFileName)) <> 0 then begin
      Delete(FInputFileName, succ(StrLen(PAnsiChar(FInputFileName))), MAXINT);
      if MoveFileExA(PAnsiChar(FOutputFileName), PAnsiChar(FInputFileName), MOVEFILE_REPLACE_EXISTING) then begin
	FInputBuffer := PAnsiChar(FInputFileName);
	FInputSize := Length(FInputFileName);
	Result := Decode(TRUE);
	if Result <> 0 then DeleteFile(FInputFileName);
      end
      else begin
	Result := kPGPError_FileOpFailed;
	DeleteFile(FOutputFileName);
	DeleteFile(FInputFileName);
	Exit;
      end;
    end
    else begin
      Result := kPGPError_FileOpFailed;
      DeleteFile(FOutputFileName);
      Exit;
    end;
  end
  else begin
    FInputBuffer := Input;
    FInputSize := InputSize;
    Result := Decode(FALSE);
  end;
end;

function TPGPDecodeCustom.SecureView(const Output: PAnsiChar; OutputSize: DWord): PGPError;
var
  PGPsc		: pPGPsc;
  PGPtls	: pPGPtls;
  Utf8OutputSize: DWord;
  Utf8Output	: PAnsiChar;
begin
  PGPsc := NIL;
  PGPtls := NIL;
  if InitPGPsc(FParentHandle, PGPsc, PGPtls, 0) then begin
    try
      if SDKVersion >= PGP9XX then begin
	Utf8OutputSize := OutputSize * UTF8Factor;
	Utf8Output := PGPNewSecureData(PGPGetDefaultMemoryMgr, Utf8OutputSize, kPGPMemoryMgrFlags_None);
	try
	  Utf8OutputSize := SecureAnsiToUtf8PChar(Output, Utf8Output, Utf8OutputSize);
	  Result := TempestViewer(PGPsc, FParentHandle, Utf8Output, Utf8OutputSize, TRUE);
	finally
	  FillChar(Utf8Output^, Utf8OutputSize, 0);
	  PGPFreeData(Utf8Output);
	end;
      end
      else Result := TempestViewer(PGPsc, FParentHandle, Output, OutputSize, TRUE);
      ProcessMessages;
    finally
      UninitPGPsc(FParentHandle, PGPsc, PGPtls);
    end;
  end
  else Result := kPGPError_FeatureNotAvailable;
end;

function TPGPDecodeCustom.WriteOutputFile(const Output: PAnsiChar; OutputSize: DWord): PGPError;
var
  OutputFile	: THandle;
  BytesWritten	: DWord;
begin
  OutputFile := CreateFileA(PAnsiChar(FOutputFileName), GENERIC_WRITE, 0, NIL, CREATE_ALWAYS,
			    FILE_ATTRIBUTE_ARCHIVE or FILE_FLAG_WRITE_THROUGH, 0);
  if OutputFile <> INVALID_HANDLE_VALUE then begin
    try
      if not WriteFile(OutputFile, Output^, OutputSize, BytesWritten, NIL) or (BytesWritten <> OutputSize) then
	Result := kPGPError_WriteFailed
      else Result := kPGPError_NoErr;
    finally
      CloseHandle(OutputFile);
    end;
  end
  else Result := kPGPError_CantOpenFile;
end;

function TPGPDecodeCustom.Decode(IsFile: Longbool): Longint;
var
  BugFix	: Longbool;
  Buffered	: Longbool;
  Recursively	: Longbool;
  OptionList	: pPGPOptionList;
  InFileSpec	: pPGPFileSpec;
  Output	: PAnsiChar;
  OutputSize	: DWord;
  Input		: PAnsiChar;
  InputSize	: DWord;
  MimeHeaders	: RawByteString;
  MimeBody	: RawByteString;
  MimeSig	: RawByteString;
  Dummy		: Longbool;
  OutputBuffer	: PAnsiChar;
  BufferSize	: DWord;
begin
  Result := 0;
  BugFix := FALSE;
  Buffered := FALSE;
  Output := NIL;
  OutputSize := 0;
  if IsFile then
    Result := HandleFileInput
  else if FInputSize = 0 then Result := kPGPError_ItemNotFound;
  if Result <> 0 then Exit;
  Input := FInputBuffer;
  InputSize := FInputSize;
  try
    InFileSpec := NIL;
    OptionList := NIL;
    Result := InitDecode;
    if Result <> 0 then Exit;
    if IsFile then begin
      if (SDKVersion >= PGP7XX) and not FFileToFile then begin
	Result := ReadInputFile;
	if Result <> 0 then Exit;
	Input := FInputBuffer;
	InputSize := FInputSize;
	Buffered := TRUE;
	IsFile := FALSE;
      end
      else begin
	Result := PGPNewFileSpecFromFullPath(FContext, PAnsiChar(FInputFileName), InFileSpec);
	if Result <> 0 then Exit;
      end;
    end;
    try
      Result := Analyze(InFileSpec);
      if FPassOption in [AnalyseOnly, AnalysePass] then begin
	if Result <> kPGPError_AsciiParseIncomplete then Result := 0; // assume FAnalysis = UNKNOWN on other errors
	if (Result = 0) and not FRecognized then begin
	  if (FX509Mode = X509IsMime) and not IsFile then begin
	    case GetPSMimeType(FInputBuffer, MimeHeaders, MimeBody, MimeSig, FCharset, TRUE) of
	      psClearSigned, psSigned: FAnalysis := SMIME_SIGNED;
	      psEncrypted: FAnalysis := SMIME_ENCRYPTED;
	    else
	      FX509Mode := X509Test;
	      case FX509Mode of
		X509Secret: FAnalysis := X509_SECRETKEY;
		X509Binary..X509Armored: FAnalysis := X509_CERTIFICATE;
	      end;
	    end;
	  end;
	  if Assigned(FOnAnalysis) then FOnAnalysis(0, FAnalysis, Dummy);
	end;
	FRecognized := (FAnalysis <> UNKNOWN);
      end;
      if FPassOption <> AnalyseOnly then begin
	if (FAnalysis = UNKNOWN) and not IsFile and (FInputSize > 0) and (FInputBuffer[0] > #127) then begin
	  Result := FixNonAsciiBug(Input);
	  if Result <> 0 then Exit;
	  BugFix := TRUE;
	end;
	// X509 and S/MIME handling
	if not IsFile then begin
	  if SDKVersion >= PGP81X then begin
	    case FX509Mode of
	      X509Off: FX509Mode := X509Processed;
	      X509Secret..X509Armored: Exit;
	      X509IsMime..X509Encrypted: begin
		case GetPSMimeType(FInputBuffer, MimeHeaders, MimeBody, MimeSig, FCharset, FALSE) of
		  psClearSigned: begin
		    FAnalysis := SMIME_SIGNED;
		    FX509Mode := X509IsMime;
		    FSignedDataBuffer := MimeBody;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FInputSize := Length(MimeSig);
		    DataCopy(FInputBuffer, PAnsiChar(MimeSig), FInputSize);
		  end;
		  psEncrypted: begin
		    FAnalysis := SMIME_ENCRYPTED;
		    FX509Mode := X509Encrypted;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FInputSize := Length(MimeBody);
		    DataCopy(FInputBuffer, PAnsiChar(MimeBody), FInputSize);
		  end;
		  psSigned: begin
		    FAnalysis := SMIME_SIGNED;
		    FX509Mode := X509Signed;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FInputSize := Length(MimeBody);
		    DataCopy(FInputBuffer, PAnsiChar(MimeBody), FInputSize);
		  end;
		  psPGP: begin
		    FIsPGPMime := TRUE;
		    FX509Mode := X509Processed;
		    Input := FInputBuffer;
		    InputSize := FInputSize;
		    FInputSize := Length(MimeBody);
		    DataCopy(FInputBuffer, PAnsiChar(MimeBody), FInputSize);
		  end;
		else
		  FX509Mode := X509Processed;
		end;
	      end;
	    end;
	  end
	  else begin
	    FX509Mode := X509Processed;
	    if (GetPSMimeType(FInputBuffer, MimeHeaders, MimeBody, MimeSig, FCharset, FALSE) = psPGP) then begin
	      FIsPGPMime := TRUE;
	      Input := FInputBuffer;
	      InputSize := FInputSize;
	      FInputSize := Length(MimeBody);
	      DataCopy(FInputBuffer, PAnsiChar(MimeBody), FInputSize);
	    end;
	  end;
	end;
	// PassThrough_ClearSigned may truncate signed data,
	// RecursivelyDecode enables PassThrough_ClearSigned,
	// if we already have clear signed data just disable it
	Recursively := FRecursivelyDecode;
	if FClearSigned then FRecursivelyDecode := FALSE;
	Result := GetOptionList(OptionList);
	if Result <> 0 then Exit;
	if IsFile then
	  Result := PGPAppendOptionList(OptionList, [PGPOInputFile(FContext, InFileSpec)])
	else Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(FContext, FInputBuffer, FInputSize)]);
	try
	  if Result <> 0 then Exit;
	  if FDetachedSigVerify or FClearSigned or (FX509Mode = X509IsMime) then begin
	    Result := SetOutputOption(OptionList);
	    if Result <> 0 then Exit;
	  end
	  else FOutputExists := FALSE;
	  if not FRecurring or FClearSigned or (FX509Mode in [X509IsMime..X509Signed]) then begin
	    FPassOption := DecodePass;
	    Result := pgpEvents.PGPDecode(FContext, OptionList, PGPOLastOption(FContext));
	    if Result <> 0 then Exit;
	    if FPassOption <> AnalyseOnly then begin
	      Result := PrepareOutput(Input, InputSize, MimeHeaders, MimeBody);
	      if Result <> 0 then Exit;
	      Output := FAllocatedOutputBuffer;
	      OutputSize := FActualOutputSize;
	      if BugFix and (PassThrough_Unrecognized in PassThroughOptions) then begin // remove leading AT
		dec(OutputSize);
		inc(Output);
	      end;
	    end;
	  end;
	finally
	  try
	    if Assigned(FOnWipePassphrase) then begin
	      if SDKVersion >= PGP8XX then SecureUtf8ToAnsiPChar(FPassphrase, FPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
	      FOnWipePassphrase(FPassphrase);
	    end;
	    if FPassCacheSeconds <> 0 then CachePassphrase(FContext, FDecryptionKey, FPassCacheSeconds, FPassCacheShare, FPassphrase);
	  finally
	    PGPFreeData(FPassphrase);
	    FPassphrase := NIL;
	  end;
	end;
	// reset after bugfixing
	FRecursivelyDecode := Recursively;
      end;
    finally
      PGPFreeFileSpec(InFileSpec);
      PGPFreeOptionList(OptionList);
      if BugFix then begin
	PGPFreeData(FInputBuffer);
	FInputBuffer := Input;
	FInputSize := InputSize;
      end;
    end;
    // workaround for verifying encrypted PGP/MIME & S/MIME signatures like PGP does internally with encrypted clear signed data
    if FRecursivelyDecode and not FRecurring and (FPassOption <> AnalyseOnly)
    and (FAnalysis in [ENCRYPTED, SMIME_ENCRYPTED]) and (FSigPropsRec.sStatus = SIGNED_NOT) then begin
      if IsFile then
	FX509Mode := X509Off
      else if FIsPGPMime then FX509Mode := X509IsMime;
      OutputBuffer := FAllocatedOutputBuffer;
      BufferSize := FActualOutputSize;
      FPassOption := AnalysePass;
      FRecurring := TRUE;
      try
	Result := DoRecursion(Output, OutputSize);
      finally
	if FEyesOnly and (OutputBuffer <> NIL) then FillChar(OutputBuffer^, BufferSize, 0);
	PGPFreeData(OutputBuffer);
	// reset to start value
	if FAnalysis in [SMIME_UNRECOGNIZED, SMIME_SIGNED] then
	  FAnalysis := SMIME_ENCRYPTED
	else FAnalysis := ENCRYPTED;
	FRecurring := FALSE;
      end;
    end
    else if FFileToFile then begin
      if FRecurring then begin
	if FAnalysis = SIGNED then
	  DeleteFile(FInputFileName)
	else if not RenameFile(String(FInputFileName), String(FOutputFileName)) then begin
	  Result := kPGPError_FileOpFailed;
	  DeleteFile(FInputFileName);
	end;
      end;
    end
    else begin
      if FRecurring then begin
	if FSigPropsRec.sStatus = SIGNED_NOT then begin
	  Output := Input;
	  OutputSize := InputSize;
	end
	else if FAnalysis = SMIME_UNRECOGNIZED then begin
	  Output := PAnsiChar(MimeHeaders + MimeBody);
	  OutputSize := Length(MimeHeaders) + Length(MimeBody);
	end;
      end;
      if FEyesOnly then
	SecureView(Output, OutputSize)
      else if not FFileOutput then begin
	try
	  SetString(FOutputBuffer, NIL, OutputSize);
	  Move(Output^, Pointer(FOutputBuffer)^, OutputSize);
	except
	  on EOutOfMemory do Result := kPGPError_OutOfMemory;
	end;
      end
      else WriteOutputFile(Output, OutputSize);
    end;
  finally
    if Buffered then PGPFreeData(Input);
    try
      if FEyesOnly and (FAllocatedOutputBuffer <> NIL) then FillChar(FAllocatedOutputBuffer^, FActualOutputSize, 0);
    finally
      try
	PGPFreeData(FAllocatedOutputBuffer);
	FAllocatedOutputBuffer := NIL;
      finally
	FinitDecode;
      end;
    end;
  end;
end;

function TPGPDecodeCustom.AnalyseBuffer(const DataBuffer: RawByteString): Longint;
begin
  FPassOption := AnalyseOnly;
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := NIL;
  try
    Result := Decode(FALSE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.AnalyseFile(const FileName: RawByteString): Longint;
begin
  FPassOption := AnalyseOnly;
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := NIL;
  try
    Result := Decode(TRUE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.AnalyseMimeBuffer(const DataBuffer: RawByteString): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalyseOnly;
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := NIL;
  try
    Result := Decode(FALSE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.AnalyseMimeFile(const FileName: RawByteString): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalyseOnly;
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := NIL;
  try
    Result := Decode(TRUE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.DecodeBuffer(const DataBuffer: RawByteString): Longint;
begin
  FPassOption := AnalysePass;
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := NIL;
  try
    Result := Decode(FALSE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.DecodeFile(const FileName: RawByteString): Longint;
begin
  FPassOption := AnalysePass;
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := NIL;
  try
    Result := Decode(TRUE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.DecodeMimeBuffer(const DataBuffer: RawByteString): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalysePass;
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInputFileName := E;
  FOutputOptions := NIL;
  try
    Result := Decode(FALSE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.DecodeMimeFile(const FileName: RawByteString): Longint;
begin
  FX509Mode := X509IsMime;
  FPassOption := AnalysePass;
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  FOutputOptions := NIL;
  try
    Result := Decode(TRUE);
  finally
    PGPFreeOptionList(FOutputOptions);
  end;
end;

function TPGPDecodeCustom.AnalyzeHandler(Event: pPGPEvent): PGPError;
var
  SkipSection	: Longbool;
begin
  Result := 0;
  SkipSection := FALSE;
  with Event^.EData.AnalyzeData do begin
    case FPassOption of
      AnalyseOnly: begin
	FRecognized := TRUE;
	if Assigned(FOnAnalysis) then FOnAnalysis(FSectionNumber, TAnalysis(SectionType), SkipSection);
	if SectionType <> kPGPAnalyze_Key then Result := kPGPError_SkipSection; // avoid key data bug
      end;
      AnalysePass: begin
	FRecognized := (SectionType <> kPGPAnalyze_Unknown);
	case SectionType of
	  kPGPAnalyze_Signed: FClearSigned := TRUE;
	  kPGPAnalyze_DetachedSignature: FDetachedSigVerify := TRUE;
	end;
	if FRecognized then FAnalysis := TAnalysis(SectionType);
	if FDetachedSigVerify or FClearSigned or (FX509Mode >= X509IsMime) then
	  Result := kPGPError_UserAbort
	else if SectionType <> kPGPAnalyze_Key then Result := kPGPError_SkipSection; // avoid key data bug
      end;
      DecodePass: begin
	if (SectionType <> kPGPAnalyze_Unknown) and (FX509Mode < X509IsMime) then FAnalysis := TAnalysis(SectionType);
	if Assigned(FOnAnalysis) then begin
	  if FX509Mode >= X509IsMime then
	    FOnAnalysis(FSectionNumber, FAnalysis, SkipSection)
	  else FOnAnalysis(FSectionNumber, TAnalysis(SectionType), SkipSection)
	end;
	Result := Ord(SkipSection = TRUE) * kPGPError_SkipSection;
      end;
    end;
  end;
end;

function TPGPDecodeCustom.RecipientsHandler(Event: pPGPEvent): PGPError;
var
  KeyIDArraySize: PGPSize;
begin
  Result := 0;
  if FAnalysis = SMIME_UNRECOGNIZED then FAnalysis := SMIME_ENCRYPTED;
  with Event^.EData.RecipientsData do begin
    PGPIncKeySetRefCount(RecipientSet);
    FRecipientsData.RecipientSet := RecipientSet;
    FRecipientsData.KeyCount := KeyCount;
    if SDKVersion >= PGP7XX then
      KeyIDArraySize := SizeOf(TPGPKeyID7) * KeyCount
    else KeyIDArraySize := SizeOf(TPGPKeyID6) * KeyCount;
    FRecipientsData.KeyIDArray := PGPNewData(PGPGetDefaultMemoryMgr, KeyIDArraySize, kPGPMemoryMgrFlags_None);
    if FRecipientsData.KeyIDArray <> NIL then Move(KeyIDArray^, FRecipientsData.KeyIDArray^, KeyIDArraySize)
  end;
end;

function TPGPDecodeCustom.PassphraseHandler(Event: pPGPEvent): PGPError;
var
  Cancel	: Longbool;
  KeyPropsList	: TKeyPropsList;
  BadPassphrase	: Longbool;
begin
  Result := 0;
  Cancel := FALSE;
  with Event^.EData.PassphraseData do begin
    KeyPropsList := TKeyPropsList.Create(spgpKeyPropFlag_IDComplete);
    try
      if (FPassphrase = NIL) or (FSectionNumber = FPassSection) then begin
	FPassSection := FSectionNumber;
	BadPassphrase := FALSE;
	if FPassphrase <> NIL then begin
	  PGPFreeData(FPassphrase);
	  FPassphrase := NIL;
	  BadPassphrase := TRUE;
	end;
	if Boolean(IsConventional) or (FPassphrase <> NIL) or (FPassCacheSeconds = 0)
	or not GetCachedPassphrase(FContext, FDecryptionKey, KeySet, FPassphrase) then begin
	  if Assigned(FOnEnterPassphrase) then begin
	    if not Boolean(IsConventional) then begin
	      Result := GetKeySetProps(FContext, KeySet, KeyPropsList, spgpKeyPropFlag_IDComplete,
				       KeyFilterFlag_CanDecrypt, UserID_Ordering);
	      if Result = 0 then Result := kPGPError_SecretKeyNotFound;
	    end;
	    if Result >= 0 then begin
	      Result := kPGPError_UserAbort;
	      FPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
	      				      kPGPMemoryMgrFlags_Clear);
	      if FPassphrase <> NIL then begin
		FOnEnterPassphrase(FPassphrase, KeyPropsList, Boolean(IsConventional), BadPassphrase, Cancel);
		if not Cancel then begin
		  if SDKVersion >= PGP8XX then SecureAnsiToUtf8PChar(FPassphrase, FPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
		  if KeyPropsList.Count = 1 then
		    GetKeyFromKeySet(KeySet, FDecryptionKey)
		  else FDecryptionKey := GetMatchingKey(FContext, KeySet, FPassphrase);
		end
		else Exit;
	      end
	      else begin
		Result := kPGPError_OutOfMemory;
		Exit;
	      end;
	    end
	    else Exit;
	  end
	  else begin
	    if Boolean(IsConventional) then
	      Result := ConvDecPassphraseDialog(FContext, FPassphrase, FPassDlgPrompt, FParentHandle)
	    else with FRecipientsData do begin
	      Result := DecryptionPassphraseDialog(FContext, RecipientSet, KeyCount, KeyIDArray,
						   FDecryptionKey, FPassphrase, TRUE, FPassDlgPrompt, FParentHandle);
	    end;
	    if Result <> 0 then Exit;
	  end;
	end;
      end
      else FPassSection := FSectionNumber;
    finally
      KeyPropsList.Free;
    end;
  end;
  Result := PGPAddJobOptions(Event^.Job, PGPOPassphrase(FContext, FPassphrase), PGPOLastOption(FContext));
end;

function TPGPDecodeCustom.KeyFoundHandler(Event: pPGPEvent): PGPError;
var
  IsSMime	: Longbool;
  KeySetSelected: pPGPKeySet;
  CertsToImport	: pPGPKeySet;
begin
  Result := 0; // don't cancel decoding whatever happens here
  with Event^.EData.KeyFoundData do begin
    ClearSignCertProps;
    IsSMime := (FX509Mode in [X509IsMime..X509Signed]);
    if IsSMime then FValidCertChain := VerifyX509CertificateChainDB(KeysFound, FSignCertProps);
    if FQueryAddKeys then begin
      if SDKVersion >= PGP7XX then begin
	KeySetSelected := PGPPeekKeyDBRootKeySet(KeysFound);
	FKeyAddResult := GetExclusiveKeySet(KeySetSelected, FKeySetMain, FContext, FIgnoreKnownFlag);
	if FKeyAddResult <> 0 then Exit;
	if FKeyDlgPrompt <> E then begin
	  if not IsSMime then
	    CertsToImport := PGPPeekKeyDBRootKeySet(KeySetSelected)
	  else begin
	    FKeyAddResult := GetCertKeySetFromKeySet(PGPPeekKeyDBRootKeySet(KeysFound), FALSE, CertsToImport);
	    if FKeyAddResult <> 0 then Exit;
	  end;
	  try
	    FKeyAddResult := SelectAddKeysToKeyRing(FContext, CertsToImport, FKeySetMain,
						    FKeyPropsList, PAnsiChar(FKeyDlgPrompt),
						    FALSE, spgpKeyPropFlag_IDFlags,
						    FParentHandle);
	  finally
	    if IsSMime then PGPFreeKeySet(CertsToImport);
	  end;
	end
	else FKeyAddResult := AddKeysToKeyRing(FContext, FKeySetMain, KeySetSelected, FKeyPropsList, spgpKeyPropFlag_IDFlags);
      end
      else begin
	FKeyAddResult := GetExclusiveKeySet(KeysFound, FKeySetMain, FContext, FIgnoreKnownFlag);
	if FKeyAddResult <> 0 then Exit;
	if FKeyDlgPrompt <> E then begin
	  FKeyAddResult := SelectAddKeysToKeyRing(FContext, KeysFound, FKeySetMain,
						  FKeyPropsList, PAnsiChar(FKeyDlgPrompt),
						  FALSE, spgpKeyPropFlag_IDFlags,
						  FParentHandle);
	end
	else FKeyAddResult := AddKeysToKeyRing(FContext, FKeySetMain, KeysFound, FKeyPropsList, spgpKeyPropFlag_IDFlags);
      end;
      if FKeyAddResult > 0 then FKeyAddResult := 0;
    end;
  end;
end;

// causes trouble when used with PassThrough_ClearSigned on clear signed input
// causes trouble with PassThrough_Unrecognized and local output options
function TPGPDecodeCustom.OutputHandler(Event: pPGPEvent): PGPError;
var
  FileSpec	: pPGPFileSpec;
begin
  Result := 0;
  with Event^.EData.OutputData do begin
    if (ForYourEyesOnly = PGPTrue) and not FFileToFile then FEyesOnly := TRUE;
    if FOutputExists then begin
      if not FFileToFile then Result := SecureMove;
    end
    else if FFileOutput and not FEyesOnly then begin
      if (FOutputFileName = E) and (SuggestedName <> E) then begin
	FOutputFileName := ExtractFilePath(FInputFileName) + SuggestedName;
      end;
      if GetOutputFileName = E then begin
	Result := kPGPError_CantOpenFile;
	Exit;
      end;
      if FFileToFile then begin
	FileSpec := NIL;
	Result := PGPNewFileSpecFromFullPath(FContext, PAnsiChar(FOutputFileName), FileSpec);
	try
	  if Result = 0 then begin
	    Result := PGPBuildOptionList(FContext, FOutputOptions,
	      [
		PGPOOutputFile(FContext, FileSpec),
		PGPOAppendOutput(FContext, PGPTrue)
	      ]);
	  end;
	finally
	  PGPFreeFileSpec(FileSpec);
	end;
      end
      else begin
	Result := PGPBuildOptionList(FContext, FOutputOptions,
	  [
	    PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
	  ]);
      end;
    end
    else begin
      FFileToFile := FALSE;
      Result := PGPBuildOptionList(FContext, FOutputOptions,
	[
	  PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
	]);
    end;
  end;
  if Result = 0 then begin
    Result := PGPAddJobOptions(Event^.Job, FOutputOptions, PGPOLastOption(FContext));
    FOutputExists := TRUE;
  end;
end;

function TPGPDecodeCustom.SignatureHandler(Event: pPGPEvent): PGPError;
var
  SignKeyType	: PGPInt32;
  SignKeyOnRing	: pPGPKeyDBObj;
  PGPSignKeyID	: TPGPKeyID7;
  SignKeyID	: TKeyID;
  KeysFound	: Pointer;
begin
  Result := 0; // don't cancel decoding whatever happens here
  ClearSigPropsRec;
  if FAnalysis = SMIME_UNRECOGNIZED then FAnalysis := SMIME_SIGNED;
  if SDKVersion >= PGP7XX then with Event^.EData.SignatureData.PGPEventSignatureData7, FSigPropsRec do begin
    if (SDKVersion >= PGP7XX)
    and (PGPGetKeyDBObjNumericProperty(SigningKey, kPGPKeyDBObjProperty_ObjectType, SignKeyType) = 0) then begin
      if SignKeyType = kPGPKeyDBObjType_SubKey then begin
	sSubKeyHexID := GetKeyPropKeyID(SigningKey);
	SigningKey := PGPPeekKeyDBObjKey(SigningKey);
      end;
    end;
    if Boolean(Checked) then begin
      sHexID := GetKeyPropKeyID(SigningKey);
      sUserID := GetKeyPropUserID(SigningKey);
      sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
      sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
      sChecked := Boolean(Checked);
      sVerified := Boolean(Verified);
      // workaround for S/MIME signatures
      if PGPKeySetIsMember(SigningKey, FKeySetMain) = PGPFalse then begin
	GetKeyFromNewSet(SigningKey, FKeySetMain, SignKeyOnRing);
	sValidity := GetKeyPropValidity(SignKeyOnRing);
      end
      else sValidity := TValidityLevel(KeyValidity);
      if FAnalysis = SMIME_SIGNED then begin
	sRevoked := Boolean(KeyRevoked) or FSignCertProps.x509IsRevoked;
	sDisabled := Boolean(KeyDisabled) or FSignCertProps.x509IsCorrupt;
	sExpired := Boolean(KeyExpired) or FSignCertProps.x509IsExpired;
      end
      else begin
	sRevoked := Boolean(KeyRevoked);
	sDisabled := Boolean(KeyDisabled);
	sExpired := Boolean(KeyExpired);
      end;
      sAxiomatic := GetKeyPropIsAxiomatic(SigningKey);
      if (SDKVersion >= PGP8XX) and (ExpirationPeriod <> 0) then begin
	sExpTimeNum := sCreaTimeNum + ExpirationPeriod;
	sExpTimeStr := UnixTimeToLocalTimeStr(sExpTimeNum);
      end;
      if SDKVersion >= PGP9XX then sHashAlgorithm := GetTHashAlgorithm(HashAlgorithm);
      if GetKeyPropIsSigningKey(SigningKey) then begin
	sStatus := TSigStatus(Checked + Verified);
	if (SDKVersion >= PGP8XX) and (sStatus = SIGNED_GOOD) and (ExpirationPeriod <> 0)
	and (sExpTimeNum < SystemTimeToUnixTimeNum) then begin
	  sStatus := SIGNED_SIG_EXPIRED;
	end;
      end
      else sStatus := SIGNED_UNKNOWN_ALG;
      sValidCertChain := FValidCertChain;
    end
    else begin
      Move(SigningKeyID, PGPSignKeyID, SizeOf(TPGPKeyID7));
      Result := PGPGetKeyIDString(PGPSignKeyID, kPGPKeyIDString_Full, SignKeyID);
      if Result = 0 then begin
	sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
	sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
	sHexID := SignKeyID;
      end;
      sStatus := SIGNED_NO_KEY;
      if (Result = 0) and FGetKeyFromServer then begin
	KeysFound := NIL;
	Result := KeyServerDialog(FContext, FKeySetMain, SignKeyID, pPGPKeyDB(KeysFound), FParentHandle);
	if (Result = 0) and (KeysFound <> NIL) then begin
	  try
	    if FKeyDlgPrompt <> E then begin
	      Result := SelectAddKeysToKeyRing(FContext, PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysFound)), FKeySetMain,
					       FKeyPropsList, PAnsiChar(FKeyDlgPrompt),
					       FALSE, spgpKeyPropFlag_IDFlags,
					       FParentHandle);
	    end;
	    if Result = 0 then begin
	      PGPAddJobOptions(Event^.Job,
			       PGPOKeySetRef(FContext, PGPPeekKeyDBRootKeySet(pPGPKeyDB(KeysFound))),
			       PGPOLastOption(FContext));
	    end;
	  finally
	    PGPFreeKeyDB(pPGPKeyDB(KeysFound));
	  end;
	end;
      end;
      Result := 0;
    end;
    sDetached := FDetachedSigVerify;
    // GetPSMimeType() does this on buffered data for all versions
    // if PGP95 then FCharset := Charset;
  end
  else with Event^.EData.SignatureData.PGPEventSignatureData6, FSigPropsRec do begin
    if Boolean(Checked) then begin
      sHexID := GetKeyPropKeyID(SigningKey);
      sUserID := GetKeyPropUserID(SigningKey);
      sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
      sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
      sChecked := Boolean(Checked);
      sVerified := Boolean(Verified);
      sValidity := TValidityLevel(KeyValidity);
      sRevoked := Boolean(KeyRevoked);
      sDisabled := Boolean(KeyDisabled);
      sExpired := Boolean(KeyExpired);
      sAxiomatic := GetKeyPropIsAxiomatic(SigningKey);
      if GetKeyPropIsSigningKey(SigningKey) then
	sStatus := TSigStatus(Checked + Verified)
      else sStatus := SIGNED_UNKNOWN_ALG;
    end
    else begin
      Move(SigningKeyID, PGPSignKeyID, SizeOf(TPGPKeyID6));
      Result := PGPGetKeyIDString(PGPSignKeyID, kPGPKeyIDString_Full, SignKeyID);
      if Result = 0 then begin
	sCreaTimeNum := PGPGetStdTimeFromPGPTime(CreationTime);
	sCreaTimeStr := UnixTimeToLocalTimeStr(sCreaTimeNum);
	sHexID := SignKeyID;
      end;
      sStatus := SIGNED_NO_KEY;
      if (Result = 0) and FGetKeyFromServer then begin
	KeysFound := NIL;
	Result := KeyServerDialog(FContext, FKeySetMain, SignKeyID, pPGPKeySet(KeysFound), FParentHandle);
	if (Result = 0) and (KeysFound <> NIL) then begin
	  try
	    if FKeyDlgPrompt <> E then begin
	      SelectAddKeysToKeyRing(FContext, pPGPKeySet(KeysFound), FKeySetMain,
				     FKeyPropsList, PAnsiChar(FKeyDlgPrompt),
				     FALSE, spgpKeyPropFlag_IDFlags,
				     FParentHandle);
	    end;
	    PGPAddJobOptions(Event^.Job,
			     PGPOKeySetRef(FContext, pPGPKeySet(KeysFound)),
			     PGPOLastOption(FContext));
	  finally
	    PGPFreeKeySet(pPGPKeySet(KeysFound));
	  end;
	end;
      end;
      Result := 0;
    end;
    sDetached := FDetachedSigVerify;
  end;
  if FShowSigLog then ShowSigInfo;
  if Assigned(FOnGetSignature) then FOnGetSignature(SigPropsRec);
end;

function TPGPDecodeCustom.EndLexHandler(Event: pPGPEvent): PGPError;
begin
  Result := 0;
  with FRecipientsData do begin
    PGPFreeData(KeyIDArray);
    PGPFreeKeySet(RecipientSet);
  end;
  FillChar(FRecipientsData, SizeOf(TPGPEventRecipientsData), 0);
end;

function TPGPDecodeCustom.DecryptionHandler(Event: pPGPEvent): PGPError;
begin
  Result := 0;
  // GetPSMimeType() does this on buffered data for all versions
  // if PGP95 then FCharset := Event^.EData.DecryptionData.Charset;
end;


end.

