{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit PGPEncode;

interface

uses
  Windows, Classes, SysUtils, UTF8, PGPStrTools, PGPStringList, PSMimeTools, KeyPropTypes, X509Types,
  PrefFuncs, KeyFuncs, X509Funcs, PGPDialogs, pgpBase, pgpErrors, pgpPubTypes, pgpUtilities,
  pgpOptionList, pgpRandomPool, pgpMemoryMgr, pgpEvents, pgpKeys, pgpTLS, pgpUI, pgpSC;

type
  TSignAlgorithm = (
    HashAlgorithm_Default,
    HashAlgorithm_MD5,
    HashAlgorithm_SHA,
    HashAlgorithm_RIPEMD160,
    HashAlgorithm_SHA256,
    HashAlgorithm_SHA384,
    HashAlgorithm_SHA512,
    HashAlgorithm_SHA224
  );
  TConventionalAlgorithm = (
    CipherAlgorithm_IDEA,
    CipherAlgorithm_3DES,
    CipherAlgorithm_CAST5,
    CipherAlgorithm_AES128,
    CipherAlgorithm_AES192,
    CipherAlgorithm_AES256,
    CipherAlgorithm_Twofish256
  );
  TFormatOption = (
    Format_Armor,
    Format_MIME,
    Format_Textmode
  );
  TFormatOptions = Set of TFormatOption;
  TOnGetInputFileName = procedure(var SuggestedName: AnsiString) of Object;
  TOnGetOutputFileName = procedure(var SuggestedName: AnsiString) of Object;
  TOnEnterPassphrase = procedure(const Passphrase: PAnsiChar;
				 const SigningKeyList: TKeyPropsList;
				 var SelectedKey: Longint;
				 BadPassphrase: Longbool;
				 var Cancel: Longbool) of Object;
  TOnShowProgress = procedure(BytesProcessed, BytesTotal: Longint) of Object;
  TOnWipePassphrase = procedure(const Passphrase: PAnsiChar) of Object;
  TPGPEncodeCustom = class(TComponent)
  private
    // internal
    FContext: pPGPContext;
    FKeySetMain: pPGPKeySet;
    FPreferences: TPreferenceRec;
    FAllocatedOutputBuffer: PAnsiChar;
    FActualOutputSize: PGPSize;
    FInputSize: PGPSize;
    FInputBuffer: PAnsiChar;
    FInputFileName: AnsiString;
    FPassphrase: PAnsiChar;
    FEncryptKeySet: pPGPKeySet;
    FSigningKey: pPGPKey;
    FClear: Longbool;
    FCompress: Longbool;
    FConventional: Longbool;
    FDetachedSign: Longbool;
    FEncrypt: Longbool;
    FFileToFile: Longbool;
    FInsertCharset: Longbool;
    FSMime: Longbool;
    FSign: Longbool;
    // protected properties
    FOutputBuffer: RawByteString;
    FFileOutput: Longbool;
    FAlwaysQueryOutputFileName: Longbool;
    // public properties
    FMimeBodyOffset: Longint;
    FMimeSeparator: AnsiString;
    FParentHandle: THandle;
    // published properties
    FCharset: AnsiString;
    FComment: AnsiString;
    FConventionalAlgorithm: TConventionalAlgorithm;
    FEncryptToSelf: Longbool;
    FEyesOnly: Longbool;
    FFormatOptions: TFormatOptions;
    FOmitMimeVersion: Longbool;
    FParanoiaSwitch: Longbool;
    FPassCacheSeconds: TPassCacheSeconds;
    FPassCacheShare : Longbool;
    FProgressInterval: Cardinal;
    FSignAlgorithm: TSignAlgorithm;
    FEncryptKeyIDs: TRawByteStrings;
    FSignKeyID: AnsiString;
    FSignCertUserName: AnsiString;
    FKeyDlgPrompt: AnsiString;
    FPassDlgPrompt: AnsiString;
    FOutputFileName: AnsiString;
    // events
    FOnGetInputFileName: TOnGetInputFileName;
    FOnGetOutputFileName: TOnGetOutputFileName;
    FOnEnterPassphrase: TOnEnterPassphrase;
    FOnShowProgress: TOnShowProgress;
    FOnWipePassphrase: TOnWipePassphrase;
    procedure SetEncryptKeyIDs(const Value: TRawByteStrings);
    procedure SetClear(Value: Longbool);
    procedure SetConventional(Value: Longbool);
    procedure SetEncrypt(Value: Longbool);
    procedure SetSign(Value: Longbool);
    function  InitEncode: PGPError;
    procedure FinitEncode;
    function  AddSignOptions(const OptionList: pPGPOptionList): PGPError;
    function  GetEncryptKeyIDs(const EncryptKeyIDs: TRawByteStrings; SignKeyID: AnsiString): PGPError;
    function  GetOptionList(var OptionList: pPGPOptionList): PGPError;
    function  ReadInputFile: PGPError;
    function  SetInputOption(IsFile: Longbool; var OptionList: pPGPOptionList): PGPError;
    function  SetOutputOption(var OptionList: pPGPOptionList): PGPError;
    function  WriteOutputFile: PGPError;
    function  Encode(IsFile: Longbool): Longint;
  protected
    property OutputBuffer: RawByteString
      read FOutputBuffer;
    property FileOutput: Longbool
      write FFileOutput;
    property AlwaysQueryOutputFileName: Longbool
      write FAlwaysQueryOutputFileName;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function KeyEncryptBuffer(const DataBuffer: RawByteString; Sign, SMime: Longbool): Longint; virtual;
    function KeyEncryptFile(const FileName: RawByteString; Sign, SMime: Longbool): Longint; virtual;
    function ConventionalEncryptBuffer(const DataBuffer: RawByteString): Longint; virtual;
    function ConventionalEncryptFile(const FileName: RawByteString): Longint; virtual;
    function ClearSignBuffer(const DataBuffer: RawByteString; SMime: Longbool): Longint; virtual;
    function ClearSignFile(const FileName: RawByteString; SMime: Longbool): Longint; virtual;
    function ArmorBuffer(const DataBuffer: RawByteString): Longint; virtual;
    function ArmorFile(const FileName: RawByteString): Longint; virtual;
    function DetachedSignBuffer(const DataBuffer: RawByteString): Longint; virtual;
    function DetachedSignFile(const FileName: RawByteString): Longint; virtual;
    property MimeBodyOffset: Longint
      read FMimeBodyOffset;
    property MimeSeparator: AnsiString
      read FMimeSeparator;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property Charset: AnsiString
      read FCharset
      write FCharset;
    property Comment: AnsiString
      read FComment
      write FComment;
    property ConventionalAlgorithm: TConventionalAlgorithm
      read FConventionalAlgorithm
      write FConventionalAlgorithm;
    property EncryptToSelf: Longbool
      read FEncryptToSelf
      write FEncryptToSelf;
    property EyesOnly: Longbool
      read FEyesOnly
      write FEyesOnly;
    property FormatOptions: TFormatOptions
      read FFormatOptions
      write FFormatOptions;
    property OmitMimeVersion: Longbool
      read FOmitMimeVersion
      write FOmitMimeVersion;
    property ParanoiaSwitch: Longbool
      read FParanoiaSwitch
      write FParanoiaSwitch;
    property PassCacheSeconds: TPassCacheSeconds
      read FPassCacheSeconds
      write FPassCacheSeconds;
    property PassCacheShare : Longbool
      read FPassCacheShare
      write FPassCacheShare;
    property ProgressInterval: Cardinal
      read FProgressInterval
      write FProgressInterval;
    property SignAlgorithm: TSignAlgorithm
      read FSignAlgorithm
      write FSignAlgorithm;
    property EncryptKeyIDs: TRawByteStrings
      read FEncryptKeyIDs
      write SetEncryptKeyIDs;
    property SignKeyID: AnsiString
      read FSignKeyID
      write FSignKeyID;
    property SignCertUserName: AnsiString
      read FSignCertUserName
      write FSignCertUserName;
    property KeyDlgPrompt: AnsiString
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property PassDlgPrompt: AnsiString
      read FPassDlgPrompt
      write FPassDlgPrompt;
    property OutputFileName: AnsiString
      read FOutputFileName
      write FOutputFileName;
    property OnGetInputFileName: TOnGetInputFileName
      read FOnGetInputFileName
      write FOnGetInputFileName;
    property OnGetOutputFileName: TOnGetOutputFileName
      read FOnGetOutputFileName
      write FOnGetOutputFileName;
    property OnEnterPassphrase: TOnEnterPassphrase
      read FOnEnterPassphrase
      write FOnEnterPassphrase;
    property OnShowProgress: TOnShowProgress
      read FOnShowProgress
      write FOnShowProgress;
    property OnWipePassphrase: TOnWipePassphrase
      read FOnWipePassphrase
      write FOnWipePassphrase;
  end;


implementation

{$IFDEF UNICODE}
uses
  AnsiStrings;
{$ENDIF}

function EventHandler(Context: pPGPContext; Event: pPGPEvent; UserValue: PGPUserValue): PGPError; cdecl;
begin
  Result := 0;
  with TPGPEncodeCustom(UserValue) do begin
    case Event^.EType of
      kPGPEvent_NullEvent:	if not FClear and (TMethod(FOnShowProgress).Code <> NIL) then begin
				    // BytesTotal always stays 0 => use FInputSize (at least in 6.5.X)
				  with Event^.EData.NullData do FOnShowProgress(BytesWritten, FInputSize);
				  ProcessMessages;
				end;
      kPGPEvent_InitialEvent:	;
      kPGPEvent_FinalEvent:	;
      kPGPEvent_ErrorEvent:	Result := Event^.EData.ErrorData.Error;
      kPGPEvent_WarningEvent:	;
    end;
  end;
end;

constructor TPGPEncodeCustom.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCharset := 'Windows-' + IntToStr(GetACP);
  FEncryptKeyIDs := TRawByteStrings.Create;
  FProgressInterval := 1000;
end;

destructor TPGPEncodeCustom.Destroy;
begin
  EncryptKeyIDs.Free;
  inherited Destroy;
end;

procedure TPGPEncodeCustom.SetEncryptKeyIDs(const Value: TRawByteStrings);
begin
  FEncryptKeyIDs.Assign(Value);
end;

procedure TPGPEncodeCustom.SetClear(Value: Longbool);
begin
  if Value <> FClear then FClear := Value;
  if FClear then begin
    FSign := TRUE;
    FEncrypt := FALSE;
    FCompress := FALSE;
    FConventional := FALSE;
    FDetachedSign := FALSE;
    Include(FFormatOptions, Format_Armor);
    Include(FFormatOptions, Format_Textmode);
  end;
end;

procedure TPGPEncodeCustom.SetConventional(Value: Longbool);
begin
  if Value <> FConventional then FConventional := Value;
  if FConventional then begin
    FDetachedSign := FALSE;
    FEncrypt := FALSE;
    FClear := FALSE;
    FSign := FALSE;
  end;
end;

procedure TPGPEncodeCustom.SetEncrypt(Value: Longbool);
begin
  if Value <> FEncrypt then FEncrypt := Value;
  if FEncrypt then begin
    FDetachedSign := FALSE;
    FConventional := FALSE;
    FClear := FALSE;
  end;
end;

procedure TPGPEncodeCustom.SetSign(Value: Longbool);
begin
  if Value <> FSign then FSign := Value;
  if FSign then FConventional := FALSE;
end;

function TPGPEncodeCustom.InitEncode: PGPError;
begin
  FSigningKey := NIL;
  FPassphrase := NIL;
  FOutputBuffer := E;
  FActualOutputSize := 0;
  FAllocatedOutputBuffer := NIL;
  FMimeBodyOffset := 0;
  SetString(FMimeSeparator, NIL, SizeOf(TMimeSeparator));
  PAnsiChar(Pointer(FMimeSeparator))^ := #0;
  Result := KeyRings.InitKeyRings(FContext, FKeySetMain);
  GetPreferences(FPreferences, PrefsFlag_GroupsFile or PrefsFlag_DefaultKeyID);
end;

procedure TPGPEncodeCustom.FinitEncode;
begin
  KeyRings.FreeKeyRings;
end;

function SignToHashAlgorithm(SignAlgorithm: TSignAlgorithm): THashAlgorithm;
begin
  Result := THashAlgorithm(Ord(SignAlgorithm) + Ord(SignAlgorithm > HashAlgorithm_Default));
end;

function GetPGPHashAlgorithm(Algorithm: TSignAlgorithm): PGPHashAlgorithm;
begin
  Result := PGPHashAlgorithm(
    Ord(Algorithm) +
    Ord(Algorithm > HashAlgorithm_RIPEMD160) * Ord(pred(kPGPHashAlgorithm_SHA256 - kPGPHashAlgorithm_RIPEMD160))
  );
end;

function TPGPEncodeCustom.AddSignOptions(const OptionList: pPGPOptionList): PGPError;
var
  SignKeySet	: pPGPKeySet;
  SigningCert	: pPGPKeyDBObj;
  KeyPropsList	: TKeyPropsList;
  Armor		: PGPUInt32;
  MIME		: PGPUInt32;
  Textmode	: PGPUInt32;
  SignKeyHexID	: RawByteString;
  SelectedKey	: Longint;
  Cancel	: Longbool;
  BadPassphrase	: Longbool;
  RipeMD	: PGPUInt32;
  SignOption	: TSignOption;
  SMimeOptions	: pPGPOptionList;
begin
  if FSMime then begin
    SignKeySet := NIL;
    Result := GetX509CertifierSet(FKeySetMain, SignKeySet, FALSE, TRUE, TRUE);
    if Result <> 0 then Exit;
    SigningCert := NIL;
  end
  else SignKeySet := FKeySetMain;
  KeyPropsList := NIL;
  try
    Armor := Ord(Format_Armor in FFormatOptions);
    MIME := Ord(Format_MIME in FFormatOptions);
    Textmode := Ord(Format_Textmode in FFormatOptions);
    if Assigned(FOnEnterPassphrase) then begin
      if FindKeyProps(E, KeyPropsList, spgpKeyPropFlag_IDComplete,
		      KeyFilterFlag_CanSign, UserID_Ordering) > 0 then begin
	Result := GetHexIDByAnyID(FContext, SignKeySet, FSignKeyID, SignKeyHexID);
	if Result <> 1 then with FPreferences do begin
	  if DefaultKeyHexID <> E then begin
	    FSignKeyID := DefaultKeyHexID;
	    Result := 1;
	  end;
	end
	else FSignKeyID := SignKeyHexID;
	if Result = 1 then with KeyPropsList do begin
	  SelectedKey := IndexOf(FSignKeyID);
	  if SelectedKey > 0 then Move(SelectedKey, 0);
	end;
      end
      else begin
	Result := kPGPError_SecretKeyNotFound;
	Exit;
      end;
      if (GetKeyByHexID(SignKeySet, PAnsiChar(FSignKeyID), FSigningKey) <> 0) or (FPassCacheSeconds = 0)
      or not GetCachedPassphrase(FContext, FSigningKey, NIL, FPassphrase) then begin
	FPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
					kPGPMemoryMgrFlags_Clear);
	if FPassphrase <> NIL then begin
	  SelectedKey := 0;
	  Cancel := FALSE;
	  BadPassphrase := FALSE;
	  Result := kPGPError_UserAbort;
	  repeat
	    if FPassphrase[0] <> #0 then FillChar(FPassphrase^, succ(MaxPassphraseLength) * UTF8Factor, 0);
	    FOnEnterPassphrase(FPassphrase, KeyPropsList, SelectedKey, BadPassphrase, Cancel);
	    if not Cancel then begin
	      if SDKVersion >= PGP8XX then SecureAnsiToUtf8PChar(FPassphrase, FPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
	      if (SelectedKey >= 0) and (SelectedKey < KeyPropsList.Count) then begin
		FSignKeyID := KeyPropsList[SelectedKey];
		BadPassphrase := not PassphraseIsValid(FContext, SignKeySet, PAnsiChar(FSignKeyID), FPassphrase);
	      end
	      else begin
		Result := kPGPError_SecretKeyNotFound;
		Exit;
	      end;
	    end
	    else Exit;
	  until Cancel or not BadPassphrase;
	end
	else begin
	  Result := kPGPError_OutOfMemory;
	  Exit;
	end;
	Result := GetKeyByHexID(SignKeySet, PAnsiChar(FSignKeyID), FSigningKey);
	if Result <> 0 then Exit;
      end;
    end
    else begin
      if (FSignKeyID = E)
      or (FindKeyProps(FSignKeyID, KeyPropsList,
		       spgpKeyPropFlag_KeyID,
		       KeyFilterFlag_CanSign,
		       Creation_Ordering) <> 1) then begin
	FSignKeyID := FPreferences.DefaultKeyHexID;
	if (FSignKeyID = E) and (KeyPropsList.Count > 0) then FSignKeyID := KeyPropsList[0];
      end
      else FSignKeyID := KeyPropsList[0];
      if (GetKeyByHexID(SignKeySet, PAnsiChar(FSignKeyID), FSigningKey) <> 0) or (FPassCacheSeconds = 0)
      or not GetCachedPassphrase(FContext, FSigningKey, NIL, FPassphrase) then begin
	if FSMime then
	  SignOption := soNo
	else SignOption := TSignOption(Ord(FClear = TRUE) * Ord(soClear) + Ord(FDetachedSign = TRUE) * Ord(soDetached));
	RipeMD := Ord(FSignAlgorithm = HashAlgorithm_RIPEMD160);
	if FSMime then begin
	  Result := X509GetCertAndPassphraseDialog(FContext, SignKeySet, FSigningKey, SigningCert, FPassphrase,
						   FSignCertUserName, FSignKeyID, FPassDlgPrompt, FParentHandle);
	end
	else begin
	  Result := SigningPassphraseDialog(FContext, SignKeySet, FSigningKey, FPassphrase, FSignKeyID, TRUE,
					    SignOption, Armor, MIME, RipeMD, FPassDlgPrompt, FParentHandle);
	end;
	if Result <> 0 then Exit;
	if not FEncrypt then begin
	  FFormatOptions := [];
	  if Armor <> 0 then Include(FFormatOptions, Format_Armor);
	  if MIME <> 0 then Include(FFormatOptions, Format_MIME);
	  if Textmode <> 0 then Include(FFormatOptions, Format_Textmode);
	end;
	if FSMime and not (FSignAlgorithm in [HashAlgorithm_MD5, HashAlgorithm_SHA]) then
	  FSignAlgorithm := HashAlgorithm_Default
	else if RipeMD <> 0 then FSignAlgorithm := HashAlgorithm_RIPEMD160;
      end;
    end;
    if FSignAlgorithm = HashAlgorithm_Default then begin
      case GetKeyPropKeyAlg(FSigningKey) of
	KeyAlgorithm_RSA..KeyAlgorithm_RSASignOnly: if GetKeyPropLegacy(FSigningKey) then
	  FSignAlgorithm := HashAlgorithm_MD5
	else FSignAlgorithm := HashAlgorithm_SHA;
	KeyAlgorithm_DH..KeyAlgorithm_DHDSS: FSignAlgorithm := HashAlgorithm_SHA;
      else
	Result := kPGPError_UnknownPublicKeyAlgorithm;
	Exit;
      end;
    end;
    if FSMime then begin
      if SigningCert = NIL then SigningCert := GetValidX509CertFromKey(FSignCertUserName, FSigningKey, NIL, SignKeySet);
      Result := PGPBuildOptionList(FContext, SMimeOptions,
	[
	  PGPOPassphrase(FContext, FPassphrase),
	  PGPOSMIMESigner(FContext, SigningCert)
	]);
      if Result <> 0 then Exit;
      try
	Result := PGPAppendOptionList(OptionList,
	  [
	    PGPOSignWithKey(FContext, FSigningKey, SMimeOptions, PGPOLastOption(FContext)),
	    PGPOHashAlgorithm(FContext, GetPGPHashAlgorithm(FSignAlgorithm)),
	    PGPODetachedSig(FContext, PGPOLastOption(FContext), NIL)
	  ]);
      finally
	PGPFreeOptionList(SMimeOptions);
      end;
    end
    else begin
      Result := PGPAppendOptionList(OptionList,
	[
	  PGPOSignWithKey(FContext, FSigningKey, PGPOPassphrase(FContext, FPassphrase), PGPOLastOption(FContext)),
	  PGPOHashAlgorithm(FContext, GetPGPHashAlgorithm(FSignAlgorithm))
	]);
    end;
  finally
    KeyPropsList.Free;
    if FSMime then PGPFreeKeySet(SignKeySet);
  end;
end;

function TPGPEncodeCustom.GetEncryptKeyIDs(const EncryptKeyIDs: TRawByteStrings; SignKeyID: AnsiString): PGPError;
var
  RecipientKeys	: Pointer;
  KeyPropsList	: TKeyPropsList;
  RecipientDB	: pPGPKeyDB;
  Armor		: PGPUInt32;
  MIME		: PGPUInt32;
  Textmode	: PGPUInt32;
  KeyCount	: PGPUInt32;
begin
  RecipientKeys := NIL;
  with KeyRings do if GroupsFile = E then GroupsFile := FPreferences.GroupsFile;
  if FEncryptToSelf then begin
    if not FSign then begin
      KeyPropsList := NIL;
      try
	if FindKeyProps(SignKeyID, KeyPropsList,
			spgpKeyPropFlag_KeyID or
			spgpKeyPropFlag_CanEncrypt,
			KeyFilterFlag_CanSign,
			Creation_Ordering) <> 1 then begin
	  SignKeyID := FPreferences.DefaultKeyHexID;
	  if (SignKeyID = E) and (KeyPropsList.Count > 0) then SignKeyID := KeyPropsList[0];
	end;
      finally
	KeyPropsList.Free;
      end;
      if SignKeyID = E then begin
	Result := kPGPError_PublicKeyNotFound;
	Exit;
      end;
    end;
  end
  else SignKeyID := E;
  Result := GetKeySetByAnyIDs(FContext, FKeySetMain, EncryptKeyIDs.CommaText, pPGPKeySet(RecipientKeys));
  if Result <> 0 then begin
    if SDKVersion >= PGP7XX then begin
      RecipientDB := NIL;
      try
	EncryptKeyIDs.Clear;
	Armor := Ord(Format_Armor in FFormatOptions);
	MIME := Ord(Format_MIME in FFormatOptions);
	Textmode := Ord(Format_Textmode in FFormatOptions);
	if FSMime then begin
	  Result := GetX509EncryptCertDB(FContext, FKeySetMain, RecipientDB);
	  if Result <> 0 then Exit;
	end
	else RecipientDB := PGPPeekKeySetKeyDB(FKeySetMain);
	Result := RecipientsDialog(FContext, PGPPeekKeyDBRootKeySet(RecipientDB), TRUE, FSMime, Armor, Textmode,
				   PGPUInt32(FEyesOnly), MIME, SignKeyID, pPGPKeyDB(RecipientKeys),
				   FKeyDlgPrompt, FParentHandle);
	if Result = 0 then begin
	  FFormatOptions := [];
	  if Armor <> 0 then Include(FFormatOptions, Format_Armor);
	  if MIME <> 0 then Include(FFormatOptions, Format_MIME);
	  if Textmode <> 0 then Include(FFormatOptions, Format_Textmode);
	  KeyPropsList := NIL;
	  try
	    KeyCount := GetKeySetProps(FContext, PGPPeekKeyDBRootKeySet(pPGPKeyDB(RecipientKeys)),
				       KeyPropsList, spgpKeyPropFlag_KeyID,
				       KeyFilterFlag_CanEncrypt,
				       UserID_Ordering);
	    if KeyCount > 0 then EncryptKeyIDs.Text := {$IFDEF UNICODE}AnsiStrings.{$ENDIF}TrimRight(KeyPropsList.Text);
	  finally
	    KeyPropsList.Free;
	  end;
	end;
      finally
	PGPFreeKeyDB(pPGPKeyDB(RecipientKeys));
	if FSMime then PGPFreeKeyDB(RecipientDB);
      end;
    end
    else begin
      try
	EncryptKeyIDs.Clear;
	Armor := Ord(Format_Armor in FFormatOptions);
	MIME := Ord(Format_MIME in FFormatOptions);
	Textmode := Ord(Format_Textmode in FFormatOptions);
	Result := RecipientsDialog(FContext, FKeySetMain, TRUE, FSMime, Armor, PGPUInt32(FEyesOnly), MIME, Textmode,
				   SignKeyID, pPGPKeySet(RecipientKeys), FKeyDlgPrompt, FParentHandle);
	if Result = 0 then begin
	  FFormatOptions := [];
	  if Armor <> 0 then Include(FFormatOptions, Format_Armor);
	  if MIME <> 0 then Include(FFormatOptions, Format_MIME);
	  if Textmode <> 0 then Include(FFormatOptions, Format_Textmode);
	  KeyPropsList := NIL;
	  try
	    KeyCount := GetKeySetProps(FContext, pPGPKeySet(RecipientKeys),
				       KeyPropsList, spgpKeyPropFlag_KeyID,
				       KeyFilterFlag_CanEncrypt,
				       UserID_Ordering);
	    if KeyCount > 0 then EncryptKeyIDs.Text := {$IFDEF UNICODE}AnsiStrings.{$ENDIF}TrimRight(KeyPropsList.Text);
	  finally
	    KeyPropsList.Free;
	  end;
	end;
      finally
	PGPFreeKeySet(pPGPKeySet(RecipientKeys));
      end;
    end;
  end
  else if FEncryptToSelf and (EncryptKeyIDs.Count <> 0) then EncryptKeyIDs.Add(SignKeyID);
end;

function TPGPEncodeCustom.GetOptionList(var OptionList: pPGPOptionList): PGPError;
var
  SelectedKey	: Longint;
  Cancel	: Longbool;
  BadPassphrase	: Longbool;
  KeyPropsList	: TKeyPropsList;
  AlgorithmList	: TPGPCipherAlgorithms;
  AlgCount	: Longint;
begin
  if FSMime then begin
    if FEncrypt then begin
      Result := PGPBuildOptionList(FContext, OptionList,
	[
	  PGPOSendNullEvents(FContext, FProgressInterval),
	  PGPOEventHandler(FContext, EventHandler, Self),
	  PGPODataIsASCII(FContext, PGPTrue),
	  PGPOOutputFormat(FContext, kPGPOutputFormat_SMIMEBodyEncryptedData)
	]);
    end
    else begin
      Result := PGPBuildOptionList(FContext, OptionList,
	[
	  PGPOEventHandler(FContext, EventHandler, Self),
	  PGPODataIsASCII(FContext, PGPTrue),
	  PGPOOutputFormat(FContext, kPGPOutputFormat_SMIMEBodySignedData)
	]);
    end;
  end
  else if FDetachedSign then begin
    Result := PGPBuildOptionList(FContext, OptionList,
      [
	PGPOEventHandler(FContext, EventHandler, Self),
	PGPOArmorOutput(FContext, PGPBoolean(Format_Armor in FFormatOptions)),
	PGPODetachedSig(FContext, PGPOLastOption(FContext), NIL),
	PGPOCommentString(FContext, PAnsiChar(FComment)),
	PGPOVersionString(FContext, MyVersion)
      ]);
  end
  else begin
    Result := PGPBuildOptionList(FContext, OptionList,
      [
	PGPOSendNullEvents(FContext, FProgressInterval),
	PGPOEventHandler(FContext, EventHandler, Self),
	PGPOClearSign(FContext, PGPBoolean(FClear = TRUE)),
	PGPOCompression(FContext, PGPBoolean(FCompress = TRUE)),
	PGPOCommentString(FContext, PAnsiChar(FComment)),
	PGPOVersionString(FContext, MyVersion)
      ]);
  end;
  if Result <> 0 then Exit;
  if FConventional then begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOCipherAlgorithm(FContext, PGPCipherAlgorithm(succ(FConventionalAlgorithm)))
      ]);
    if Result <> 0 then Exit;
    if Assigned(FOnEnterPassphrase) then begin
      FPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
				      kPGPMemoryMgrFlags_Clear);
      if FPassphrase <> NIL then begin
	SelectedKey := -1;
	Cancel := FALSE;
	BadPassphrase := FALSE;
	Result := kPGPError_UserAbort;
	KeyPropsList := TKeyPropsList.Create(spgpKeyPropFlag_IDComplete);
	try
	  FOnEnterPassphrase(FPassphrase, KeyPropsList, SelectedKey, BadPassphrase, Cancel);
	finally
	  KeyPropsList.Free;
	end;
	if not Cancel then begin
	  if SDKVersion >= PGP8XX then SecureAnsiToUtf8PChar(FPassphrase, FPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
	end
	else Exit;
      end
      else begin
	Result := kPGPError_OutOfMemory;
	Exit;
      end;
    end
    else begin
      Result := ConvEncPassphraseDialog(FContext, FPassphrase, FPassDlgPrompt, FParentHandle);
      if Result <> 0 then Exit;
    end;
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOConventionalEncrypt(FContext, PGPOPassphrase(FContext, FPassphrase), PGPOLastOption(FContext))
      ]);
    if Result <> 0 then Exit;
  end
  else begin
    if FSign or FDetachedSign then begin
      Result := AddSignOptions(OptionList);
      if Result <> 0 then Exit;
    end;
    if FEncrypt and not (FSMime and FSign) then begin
      if FEncryptToSelf and not GetKeyPropCanEncrypt(FSigningKey) then
	Result := GetEncryptKeyIDs(FEncryptKeyIDs, E)
      else Result := GetEncryptKeyIDs(FEncryptKeyIDs, GetKeyPropKeyID(FSigningKey));
      if Result <> 0 then Exit;
      Result := GetKeySetByAnyIDs(FContext, FKeySetMain, FEncryptKeyIDs.CommaText, FEncryptKeySet);
      if Result <> 0 then Exit;
      if FSMime then begin
	Result := CheckValidX509CertifierSet(FEncryptKeySet);
	if Result <> 0 then Exit;
      end;
      Result := PGPAppendOptionList(OptionList, [PGPOEncryptToKeySet(FContext, FEncryptKeySet)]);
      if Result <> 0 then Exit;
      if GetAllowedCipherAlgorithms(AlgorithmList, AlgCount) = 0 then begin
	PGPAppendOptionList(OptionList, [PGPOPreferredAlgorithms(FContext, AlgorithmList, AlgCount)]);
      end;
    end;
  end;
  if not (FDetachedSign or FSMime) then begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOForYourEyesOnly(FContext, PGPBoolean(FEyesOnly = TRUE)),
	PGPOArmorOutput(FContext, PGPBoolean(Format_Armor in FFormatOptions)),
	PGPODataIsASCII(FContext, PGPBoolean(Format_Textmode in FFormatOptions))
      ]);
    if Result <> 0 then Exit;
    if Format_Mime in FFormatOptions then begin
      Result := PGPAppendOptionList(OptionList,
	[
	  PGPOPGPMIMEEncoding(FContext, PGPTrue, PGPSize(FMimeBodyOffset), Pointer(FMimeSeparator)),
	  PGPOOmitMIMEVersion(fContext, PGPBoolean(FOmitMimeVersion = TRUE))
	]);
      if Result <> 0 then Exit;
    end;
  end;
end;

function TPGPEncodeCustom.ReadInputFile: PGPError;
var
  InputFile	: THandle;
  BytesRead	: DWord;
begin
  FInputBuffer := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(FInputSize), kPGPMemoryMgrFlags_None);
  if FInputBuffer <> NIL then begin
    FInputBuffer[FInputSize] := #0;
    InputFile := CreateFileA(PAnsiChar(FInputFileName), GENERIC_READ, FILE_SHARE_READ, NIL, OPEN_EXISTING,
			     FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if InputFile <> INVALID_HANDLE_VALUE then begin
      try
	if not ReadFile(InputFile, FInputBuffer[0], FInputSize, BytesRead, NIL) or (BytesRead <> FInputSize) then
	  Result := kPGPError_ReadFailed
	else Result := kPGPError_NoErr;
      finally
	CloseHandle(InputFile);
      end;
    end
    else Result := kPGPError_CantOpenFile;
  end
  else begin
    Result := kPGPError_OutOfMemory;
    Exit;
  end;
end;

function TPGPEncodeCustom.SetInputOption(IsFile: Longbool; var OptionList: pPGPOptionList): PGPError;
var
  FileHandle	: THandle;
  FileSizeHigh	: DWord;
  InFileSpec	: pPGPFileSpec;
begin
  Result := 0;
  if IsFile then begin
    FileHandle := FileOpen(String(FInputFileName), fmOpenRead or fmShareDenyNone);
    if FileHandle <> INVALID_HANDLE_VALUE then begin
      FInputSize := GetFileSize(FileHandle, @FileSizeHigh);
      FileClose(FileHandle);
    end
    else FInputSize := $FFFFFFFF;
    if (not FParanoiaSwitch and (FInputSize = 0)) or (FInputSize = $FFFFFFFF) or (FileSizeHigh <> 0) then begin
      Result := kPGPError_ReadFailed;
      Exit;
    end;
    if not FSMime then begin
      InFileSpec := NIL;
      Result := PGPNewFileSpecFromFullPath(FContext, PAnsiChar(FInputFileName), InFileSpec);
      if Result <> 0 then Exit;
      try
	Result := PGPAppendOptionList(OptionList, [PGPOInputFile(FContext, InFileSpec)]);
      finally
	PGPFreeFileSpec(InFileSpec);
      end;
    end
    else Result := ReadInputFile;
    if Result <> 0 then Exit;
  end;
  if not IsFile or FSMime then begin
    Result := PGPAppendOptionList(OptionList, [PGPOInputBuffer(FContext, FInputBuffer, FInputSize)]);
  end;
end;

function TPGPEncodeCustom.SetOutputOption(var OptionList: pPGPOptionList): PGPError;
var
  FileSpec	: pPGPFileSpec;
begin
  if FFileOutput then begin
    if FAlwaysQueryOutputFileName or (FOutputFileName = E) or FileExists(FOutputFileName) then begin
      if (FOutputFileName = E) or (CompareText(FOutputFileName, FInputFileName) = 0) then begin
	if FDetachedSign then
	  FOutputFileName := FInputFileName + '.sig'
	else begin
	  if FSMime or (Format_Armor in FFormatOptions) then
	    FOutputFileName := FInputFileName + '.asc'
	  else FOutputFileName := FInputFileName + '.pgp';
	end;
      end;
      if Assigned(FOnGetOutputFileName) then begin
	FOnGetOutputFileName(FOutputFileName);
	DeleteFile(FOutputFileName);
      end
      else FOutputFileName := E;
    end;
    if FOutputFileName = E then begin
      Result := kPGPError_CantOpenFile;
      Exit;
    end;
  end;
  if FFileToFile then begin
    FileSpec := NIL;
    Result := PGPNewFileSpecFromFullPath(FContext, PAnsiChar(FOutputFileName), FileSpec);
    try
      if Result = 0 then Result := PGPAppendOptionList(OptionList, [PGPOOutputFile(FContext, FileSpec)]);
    finally
      PGPFreeFileSpec(FileSpec);
    end;
  end
  else begin
    Result := PGPAppendOptionList(OptionList,
      [
	PGPOAllocatedOutputBuffer(FContext, FAllocatedOutputBuffer, $FFFFFFFF, FActualOutputSize)
      ]);
  end;
end;

function TPGPEncodeCustom.WriteOutputFile: PGPError;
var
  OutputFile	: THandle;
  BytesToWrite	: DWord;
  BytesWritten	: DWord;
begin
  OutputFile := CreateFileA(PAnsiChar(FOutputFileName), GENERIC_WRITE, 0, NIL, CREATE_ALWAYS,
			    FILE_ATTRIBUTE_ARCHIVE or FILE_FLAG_WRITE_THROUGH, 0);
  if OutputFile <> INVALID_HANDLE_VALUE then begin
    try
      BytesToWrite := Length(FOutputBuffer);
      if not WriteFile(OutputFile, Pointer(FOutputBuffer)^, BytesToWrite, BytesWritten, NIL)
      or (BytesWritten <> BytesToWrite) then
	Result := kPGPError_WriteFailed
      else Result := kPGPError_NoErr;
    finally
      CloseHandle(OutputFile);
    end;
  end
  else Result := kPGPError_CantOpenFile;
end;

function TPGPEncodeCustom.Encode(IsFile: Longbool): Longint;
var
  OptionList	: pPGPOptionList;
begin
  Result := 0;
  if IsFile then begin
    FInputFileName := FInputBuffer;
    if not FileExists(FInputFileName) then begin
      if Assigned(FOnGetInputFileName) then begin
	FOnGetInputFileName(FInputFileName);
	if FInputFileName = E then Result := kPGPError_FileNotFound;
      end
      else Result := kPGPError_FileNotFound;
    end;
    if (SDKVersion >= PGP7XX) and not (FSMime or (Format_Mime in FFormatOptions)) then
      FFileToFile := FFileOutput
    else FFileToFile := FALSE;
  end
  else if not FParanoiaSwitch and (FInputSize = 0) then Result := kPGPError_ItemNotFound;
  if Result = 0 then begin
    OptionList := NIL;
    FEncryptKeySet := NIL;
    Result := InitEncode;
    if Result <> 0 then Exit;
    try
      Result := GetOptionList(OptionList);
      try
	if Result <> 0 then Exit;
	Result := SetInputOption(IsFile, OptionList);
	if Result <> 0 then Exit;
	Result := SetOutputOption(OptionList);
	if Result <> 0 then Exit;
	FInsertCharset := FInsertCharset and
			  (FFormatOptions * [Format_Armor, Format_Textmode] = [Format_Armor, Format_Textmode]);
	Result := pgpEvents.PGPEncode(FContext, OptionList, PGPOLastOption(FContext));
	if not FFileToFile and (Result = 0) and (FActualOutputSize <> 0) then begin
	  try
	    SetString(FOutputBuffer, NIL, FActualOutputSize);
	    Move(FAllocatedOutputBuffer^, Pointer(FOutputBuffer)^, FActualOutputSize);
	  except
	    on EOutOfMemory do Result := kPGPError_OutOfMemory;
	  end;
	end;
      finally
	try
	  Delete(FMimeSeparator, succ(StrLen(PAnsiChar(FMimeSeparator))), MAXINT);
	  PGPFreeOptionList(OptionList);
	  PGPFreeKeySet(FEncryptKeySet);
	  if Assigned(FOnWipePassphrase) then begin
	    if SDKVersion >= PGP8XX then SecureUtf8ToAnsiPChar(FPassphrase, FPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
	    FOnWipePassphrase(FPassphrase);
	  end;
	  if FPassCacheSeconds <> 0 then CachePassphrase(FContext, FSigningKey, FPassCacheSeconds, FPassCacheShare, FPassphrase);
	finally
	  PGPFreeData(FAllocatedOutputBuffer);
	  FAllocatedOutputBuffer := NIL;
	  PGPFreeData(FPassphrase);
	  FPassphrase := NIL;
	end;
      end;
    finally
      FinitEncode;
    end;
  end;
end;

function TPGPEncodeCustom.KeyEncryptBuffer(const DataBuffer: RawByteString; Sign, SMime: Longbool): Longint;
var
  SMimeSigned: RawByteString;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetSign(Sign);
  SetEncrypt(TRUE);
  FCompress := TRUE;
  FSMime := SMime;
  if SMime and Sign then FEncrypt := FALSE;
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInsertCharset := (FCharset <> E);
  Result := Encode(FALSE);
  if Result = 0 then begin
    if FSMime then begin
      if not FEncrypt then begin
	SMimeSigned := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), FALSE, TRUE);
	FInputBuffer := PAnsiChar(SMimeSigned);
	FInputSize := Length(SMimeSigned);
	FEncrypt := TRUE;
	FSign := FALSE;
	Result := Encode(FALSE);
	if Result <> 0 then Exit;
      end;
      FOutputBuffer := SMimeBuild(FOutputBuffer, E, FMimeSeparator, FMimeBodyOffset,
				  HashAlgorithm_Invalid, TRUE, FOmitMimeVersion);
    end
    else if Format_Mime in FFormatOptions then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    HashAlgorithm_Invalid, TRUE, FOmitMimeVersion);
    end
    else if FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, FALSE)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => RawByteString based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncodeCustom.KeyEncryptFile(const FileName: RawByteString; Sign, SMime: Longbool): Longint;
var
  SMimeSigned: RawByteString;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetSign(Sign);
  SetEncrypt(TRUE);
  FCompress := TRUE;
  FSMime := SMime;
  if SMime and Sign then FEncrypt := FALSE;
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  FInsertCharset := (FCharset <> E);
  Result := Encode(TRUE);
  if Result = 0 then begin
    if FSMime then begin
      if not FEncrypt then begin
	SMimeSigned := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), FALSE, TRUE);
	FInputBuffer := PAnsiChar(SMimeSigned);
	FInputSize := Length(SMimeSigned);
	FEncrypt := TRUE;
	FSign := FALSE;
	Result := Encode(FALSE);
	if Result <> 0 then Exit;
      end;
      FOutputBuffer := SMimeBuild(FOutputBuffer, E, FMimeSeparator, FMimeBodyOffset,
				  HashAlgorithm_Invalid, TRUE, FOmitMimeVersion);
    end
    else if Format_Mime in FFormatOptions then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    HashAlgorithm_Invalid, TRUE, FOmitMimeVersion);
    end
    else if not FFileToFile and FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, FALSE)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => RawByteString based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncodeCustom.ConventionalEncryptBuffer(const DataBuffer: RawByteString): Longint;
begin
  FCompress := TRUE;
  SetConventional(TRUE);
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  Result := Encode(FALSE);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncodeCustom.ConventionalEncryptFile(const FileName: RawByteString): Longint;
begin
  FCompress := TRUE;
  SetConventional(TRUE);
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  Result := Encode(TRUE);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncodeCustom.ClearSignBuffer(const DataBuffer: RawByteString; SMime: Longbool): Longint;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetClear(TRUE);
  FSMime := SMime;
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  FInsertCharset := (FCharset <> E);
  Result := Encode(FALSE);
  if Result = 0 then begin
    if FSMime then begin
      FOutputBuffer := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), FALSE, FOmitMimeVersion);
    end
    else if Format_Mime in FFormatOptions then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    SignToHashAlgorithm(FSignAlgorithm), FALSE, FOmitMimeVersion);
    end
    else if FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, TRUE)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => RawByteString based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncodeCustom.ClearSignFile(const FileName: RawByteString; SMime: Longbool): Longint;
begin
  if SMime and (SDKVersion < PGP81X) then begin
    Result := kPGPError_FeatureNotAvailable;
    Exit;
  end;
  SetClear(TRUE);
  FSMime := SMime;
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  FInsertCharset := (FCharset <> E);
  Result := Encode(TRUE);
  if Result = 0 then begin
    if FSMime then begin
      FOutputBuffer := SMimeBuild(FInputBuffer, FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				  SignToHashAlgorithm(FSignAlgorithm), FALSE, FOmitMimeVersion);
    end
    else if Format_Mime in FFormatOptions then begin
      FOutputBuffer := PGPMimeBuild(FOutputBuffer, FMimeSeparator, FMimeBodyOffset,
				    SignToHashAlgorithm(FSignAlgorithm), FALSE, FOmitMimeVersion);
    end
    else if not FFileToFile and FInsertCharset then
      FOutputBuffer := InsertCharset(FCharset, FOutputBuffer, TRUE)
    else Exit;
    if FFileOutput then Result := WriteOutputFile;
  end;
  // zero based FMimeBodyOffset => RawByteString based
  if FMimeBodyOffset > 0 then inc(FMimeBodyOffset);
end;

function TPGPEncodeCustom.ArmorBuffer(const DataBuffer: RawByteString): Longint;
begin
  FClear := FALSE;
  FCompress := TRUE;
  FDetachedSign := FALSE;
  FEncrypt := FALSE;
  FSign := FALSE;
  Include(FFormatOptions, Format_Armor);
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  Result := Encode(FALSE);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncodeCustom.ArmorFile(const FileName: RawByteString): Longint;
begin
  FClear := FALSE;
  FCompress := TRUE;
  FDetachedSign := FALSE;
  FEncrypt := FALSE;
  FSign := FALSE;
  Include(FFormatOptions, Format_Armor);
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  Result := Encode(TRUE);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncodeCustom.DetachedSignBuffer(const DataBuffer: RawByteString): Longint;
begin
  FClear := FALSE;
  FConventional := FALSE;
  FDetachedSign := TRUE;
  FEncrypt := FALSE;
  FSign := FALSE;
  FInputBuffer := PAnsiChar(DataBuffer);
  FInputSize := Length(DataBuffer);
  Result := Encode(FALSE);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;

function TPGPEncodeCustom.DetachedSignFile(const FileName: RawByteString): Longint;
begin
  FClear := FALSE;
  FConventional := FALSE;
  FDetachedSign := TRUE;
  FEncrypt := FALSE;
  FSign := FALSE;
  FInputBuffer := PAnsiChar(FileName);
  FInputSize := Length(FileName);
  Result := Encode(TRUE);
  if FFileOutput and not FFileToFile then Result := WriteOutputFile;
end;


end.

