{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2001-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit PGPKeyGenerate;

interface

uses
  Windows, Classes, SysUtils, UTF8, PGPStrTools, PSMimeTools, KeyPropTypes, X509Types, PrefFuncs, KeyFuncs, X509Funcs, PGPDialogs,
  pgpBase, pgpErrors, pgpPubTypes, pgpUtilities, pgpOptionList, pgpRandomPool,
  pgpMemoryMgr, pgpEvents, pgpKeys, pgpTLS, pgpUI, pgpSC;

type
  TKeySize = 512..4096;
  TSubKeySize = 512..4096;
  TMinPassLen = 0..255;
  TMinPassQual = 0..100;
  TCipherCount = 0..Ord(High(TCipherAlgorithm));

  TOnEnterPassphrase = procedure(const Passphrase: PAnsiChar; const MasterKeyProps: TKeyPropsRec;
				 MinPassLen: TMinPassLen; MinPassQual: TMinPassQual;
				 var Cancel: Longbool) of Object;
  TOnGetUserNameAddress = procedure(var UserName, EmailAddress: AnsiString) of Object;
  TOnKeyGeneration = procedure(const NewHexID: AnsiString; const MasterKeyProps: TKeyPropsRec; Aborted: Longbool) of Object;
  TOnShowState = procedure(State: Char; var Cancel: Longbool) of Object;
  TPGPKeysGenerateCustom = Class(TComponent)
  private
    FAborted: Longbool;
    FContext: pPGPContext;
    FKeySetMain: pPGPKeySet;
    FPubKeyAlgorithm: TKeyAlgorithm;
    FCipherAlgorithm: TCipherAlgorithm;
    FCipherAlgorithmList: Array[TCipherCount] of TCipherAlgorithm;
    FExpires: Longint;
    FFastGenerate: Longbool;
    FFailNoEntropy: Longbool;
    FLegacyKey: Longbool;
    FMinPassLen: TMinPassLen;
    FMinPassQual: TMinPassQual;
    FPassphrase: PAnsiChar;
    FUserID: AnsiString;
    FNewHexID: AnsiString;
    FUserName: AnsiString;
    FEmailAddress: AnsiString;
    FMasterKeyHexID: AnsiString;
    FNewKeyHexID: AnsiString;
    FNewSubKeyHexID: AnsiString;
    FKeySize: TKeySize;
    FSubKeySize: TSubKeySize;
    FParentHandle: THandle;
    FKeyDlgPrompt: AnsiString;
    FPassDlgPrompt: AnsiString;
    FOnEnterPassphrase: TOnEnterPassphrase;
    FOnGetUserNameAddress: TOnGetUserNameAddress;
    FOnKeyGeneration: TOnKeyGeneration;
    FOnShowState: TOnShowState;
    function InitKeyGen: PGPError;
    procedure FinitKeyGen(var Result: PGPError);
    function GetUserID: AnsiString;
    function GetMasterKeyHexID: Longint;
    function GetUserNameAddress: Longint;
    function GetPassphrase(OfMasterKey: Longbool): PGPError;
    function GetEntropy(IncludeSubKey: Longbool): PGPError;
    function KeyGenerate(IncludeSubKey: Longbool): PGPError;
    function SubKeyGenerate(ForOldMasterKey: Longbool): PGPError;
  protected
    procedure SetUserName(const Name: AnsiString);
    procedure SetEmailAddress(const Address: AnsiString);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetCipherAlgPrefs(const CipherAlgList: Array of TCipherAlgorithm);
    function DHDSSKeyGenerate: Longint; virtual;
    function DHSubKeyGenerate: Longint; virtual;
    function DSAKeyGenerate: Longint; virtual;
    function RSASubKeyGenerate: Longint; virtual;
    function RSAKeyGenerate(Legacy: Longbool): Longint; virtual;
    property ParentHandle: THandle
      read FParentHandle
      write FParentHandle;
  published
    property CipherAlgorithm: TCipherAlgorithm
      read FCipherAlgorithm
      write FCipherAlgorithm;
    property Expires: Longint
      read FExpires
      write FExpires;
    property FastGenerate: Longbool
      read FFastGenerate
      write FFastGenerate;
    property FailNoEntropy: Longbool
      read FFailNoEntropy
      write FFailNoEntropy;
    property MinPassLen: TMinPassLen
      read FMinPassLen
      write FMinPassLen;
    property MinPassQual: TMinPassQual
      read FMinPassQual
      write FMinPassQual;
    property UserName: AnsiString
      read FUserName
      write SetUserName;
    property EmailAddress: AnsiString
      read FEmailAddress
      write SetEmailAddress;
    property MasterKeyHexID: AnsiString
      read FMasterKeyHexID
      write FMasterKeyHexID;
    property KeySize: TKeySize
      read FKeySize
      write FKeySize;
    property SubKeySize: TSubKeySize
      read FSubKeySize
      write FSubKeySize;
    property KeyDlgPrompt: AnsiString
      read FKeyDlgPrompt
      write FKeyDlgPrompt;
    property PassDlgPrompt: AnsiString
      read FPassDlgPrompt
      write FPassDlgPrompt;
    property OnEnterPassphrase: TOnEnterPassphrase
      read FOnEnterPassphrase
      write FOnEnterPassphrase;
    property OnGetUserNameAddress: TOnGetUserNameAddress
      read FOnGetUserNameAddress
      write FOnGetUserNameAddress;
    property OnKeyGeneration: TOnKeyGeneration
      read FOnKeyGeneration
      write FOnKeyGeneration;
    property OnShowState: TOnShowState
      read FOnShowState
      write FOnShowState;
  end;


implementation

{$IFDEF UNICODE}
uses
  AnsiStrings;
{$ENDIF}

function EventHandler(Context: pPGPContext; Event: pPGPEvent; UserValue: PGPUserValue): PGPError; cdecl;
var
  Cancel	: Longbool;
begin
  Result := 0;
  Cancel := FALSE;
  with TPGPKeysGenerateCustom(UserValue) do begin
    case Event^.EType of
      kPGPEvent_KeyGenEvent: begin
	if TMethod(FOnShowState).Code <> NIL then with Event^.EData.KeyGenData do begin
	  FOnShowState(Char(State), Cancel);
	  if Cancel then begin
	    Result := kPGPError_UserAbort;
	    FAborted := TRUE;
	  end;
	end;
	ProcessMessages;
      end;
    end;
  end;
end;

constructor TPGPKeysGenerateCustom.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFastGenerate := TRUE;
  FMinPassLen := 10;
  FMinPassQual := 100;
  FKeySize := 1024;
  FSubKeySize := 1024;
end;

destructor TPGPKeysGenerateCustom.Destroy;
begin
  inherited Destroy;
end;

function TPGPKeysGenerateCustom.InitKeyGen: PGPError;
begin
  FUserID := E;
  FNewHexID := E;
  FAborted := FALSE;
  FLegacyKey := TRUE;
  FPassphrase := NIL;
  FNewKeyHexID := E;
  FNewSubKeyHexID := E;
  Result := KeyRings.InitKeyRings(FContext, FKeySetMain);
end;

procedure TPGPKeysGenerateCustom.FinitKeyGen(var Result: PGPError);
var
  KeyPropsList	: TKeyPropsList;
begin
  PGPFreeData(FPassphrase);
  KeyRings.FreeKeyRings;
  try
    if not FAborted then begin
      KeyPropsList := NIL;
      FindKeyProps(FMasterKeyHexID, KeyPropsList,
		   spgpKeyPropFlag_IDComplete,
		   KeyFilterFlag_CanSign,
		   Any_Ordering);
    end
    else Result := kPGPError_UserAbort;
    if Assigned(FOnKeyGeneration) then FOnKeyGeneration(FNewHexID, pKeyPropsRec(KeyPropsList.Objects[0])^, Result <> 0);
  finally
    KeyPropsList.Free;
  end;
end;

procedure TPGPKeysGenerateCustom.SetUserName(const Name: AnsiString);
var
  ErrorStr	: RawByteString;
begin
  if Name <> FUserName then begin
    if FirstCharPos(AT, Name) = 0 then
      FUserName := Trim(Name)
    else begin
      SetString(ErrorStr, NIL, 255);
      PGPGetErrorString(kPGPError_BadParams, 255, PAnsiChar(ErrorStr));
      Delete(ErrorStr, succ(StrLen(PAnsiChar(ErrorStr))), MAXINT);
      MessageBoxA(FParentHandle, PAnsiChar(ErrorStr + ': ' + Name), 'PGP', MB_ICONERROR);
    end;
  end;
end;

procedure TPGPKeysGenerateCustom.SetEmailAddress(const Address: AnsiString);
var
  ErrorStr	: RawByteString;
begin
  if Address <> FEmailAddress then begin
    if (Address = E) or (FirstCharPos(AnsiChar(AT), Address) <> 0) then
      FEmailAddress := Trim(Address)
    else begin
      SetString(ErrorStr, NIL, 255);
      PGPGetErrorString(kPGPError_BadParams, 255, PAnsiChar(ErrorStr));
      Delete(ErrorStr, StrLen(PAnsiChar(ErrorStr)), MAXINT);
      MessageBoxA(FParentHandle, PAnsiChar(ErrorStr + ': ' + Address), 'PGP', MB_ICONERROR);
    end;
  end;
end;

function TPGPKeysGenerateCustom.GetUserID: AnsiString;
begin
  if FEmailAddress <> E then begin
    Result := LT + FEmailAddress + GT;
    if FUserName <> E then Result := FUserName + SP + Result;
  end
  else Result := FUserName;
end;

function TPGPKeysGenerateCustom.GetMasterKeyHexID: Longint;
var
  KeyPropsList	: TKeyPropsList;
begin
  KeyPropsList := NIL;
  try
    Result := FindKeyProps(FMasterKeyHexID,
			   KeyPropsList,
			   spgpKeyPropFlag_KeyID,
			   KeyFilterFlag_CanSign or KeyFilterFlag_V4,
			   Any_Ordering);
    if Result = 0 then begin
      Result := kPGPError_SecretKeyNotFound;
      Exit;
    end
    else if Result <> 1 then begin
      if KeyPropsList <> NIL then KeyPropsList.Clear;
      Result := SelectKeysDialog(FKeyDlgPrompt, KeyPropsList, TRUE, spgpKeyPropFlag_KeyID,
			       KeyFilterFlag_CanSign or KeyFilterFlag_V4, FParentHandle);
      if Result = 0 then FMasterKeyHexID := KeyPropsList.Strings[0];
    end
    else begin
      FMasterKeyHexID := KeyPropsList.Strings[0];
      Result := 0;
    end;
  finally
    KeyPropsList.Free;
  end;
end;

function TPGPKeysGenerateCustom.GetUserNameAddress: Longint;
begin
  Result := 0;
  if (FUserName = E) or (FEmailAddress = E) then begin
    if Assigned(FOnGetUserNameAddress) then FOnGetUserNameAddress(FUserName, FEmailAddress);
  end;
  if FUserName = E then Result := kPGPError_BadParams;
end;

function TPGPKeysGenerateCustom.GetPassphrase(OfMasterKey: Longbool): PGPError;
var
  Cancel	: Longbool;
  KeySetFound	: pPGPKeySet;
  PassLength	: Longint;
  KeyPropsRec	: TKeyPropsRec;
  KeyPropsList	: TKeyPropsList;
begin
  Result := 0;
  Cancel := FALSE;
  KeySetFound := NIL;
  if Assigned(FOnEnterPassphrase) then begin
    FillChar(KeyPropsRec, SizeOf(TKeyPropsRec), 0);
    KeyPropsList := NIL;
    try
      if OfMasterKey then begin
	if FindKeyProps(FMasterKeyHexID, KeyPropsList, spgpKeyPropFlag_IDComplete,
			KeyFilterFlag_CanSign or KeyFilterFlag_V4, Any_Ordering) = 1 then begin
	  KeyPropsRec := pKeyPropsRec(KeyPropsList.Strings[0])^;
	end
	else begin
	  Result := kPGPError_SecretKeyNotFound;
	  Exit;
	end;
      end;
      FPassphrase := PGPNewSecureData(PGPGetDefaultMemoryMgr, succ(MaxPassphraseLength) * UTF8Factor,
				      kPGPMemoryMgrFlags_Clear);
      if FPassphrase <> NIL then begin
	repeat
	  if SDKVersion >= PGP8XX then SecureUtf8ToAnsiPChar(FPassphrase, FPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
	  FOnEnterPassphrase(FPassphrase, KeyPropsRec, FMinPassLen, FMinPassQual, Cancel);
	  PassLength := StrLen(FPassphrase);
	  if SDKVersion >= PGP8XX then SecureAnsiToUtf8PChar(FPassphrase, FPassphrase, succ(MaxPassphraseLength) * UTF8Factor);
	until Cancel or ((PassLength >= FMinPassLen) and (PGPEstimatePassphraseQuality(FPassphrase) >= FMinPassQual));
	if Cancel then Result := kPGPError_UserAbort;
      end
      else Result := kPGPError_OutOfMemory;
    finally
      KeyPropsList.Free;
    end;
  end
  else begin
    if OfMasterKey then begin
      Result := GetKeySetByAnyIDs(FContext, FKeySetMain, FMasterKeyHexID, KeySetFound);
      if Result <> 0 then Exit;
      try
	Result := KeyPassphraseDialog(FContext, KeySetFound, FPassphrase, FPassDlgPrompt, FParentHandle);
      finally
	PGPFreeKeySet(KeySetFound);
      end;
    end
    else begin
      Result := ConfirmationPassphraseDialog(FContext, FPassphrase, FMinPassLen, FMinPassQual, TRUE,
					   FPassDlgPrompt, FParentHandle);
    end;
  end;
end;

function TPGPKeysGenerateCustom.GetEntropy(IncludeSubKey: Longbool): PGPError;
var
  OptionList	: pPGPOptionList;
  EntropyNeeded	: PGPUInt32;
  ParamOption	: pPGPoptionList;
begin
  OptionList := NIL;
  Result := PGPBuildOptionList(FContext, OptionList,
    [
      PGPOKeyGenParams(FContext, PGPPublicKeyAlgorithm(FPubKeyAlgorithm), FKeySize),
      PGPOKeyGenFast(FContext, PGPBoolean(FFastGenerate = TRUE))
    ]);
  if Result <> 0 then Exit;
  try
    EntropyNeeded := PGPGetKeyEntropyNeeded(FContext, OptionList, PGPOLastOption(FContext));
  finally
    PGPFreeOptionList(OptionList);
  end;
  if IncludeSubKey then begin
    OptionList := NIL;
    if FPubKeyAlgorithm <> KeyAlgorithm_RSA then
      ParamOption := PGPOKeyGenParams(FContext, PGPPublicKeyAlgorithm(KeyAlgorithm_DH), FSubKeySize)
    else ParamOption := PGPOKeyGenParams(FContext, PGPPublicKeyAlgorithm(KeyAlgorithm_RSA), FSubKeySize);
    Result := PGPBuildOptionList(FContext, OptionList,
      [
	ParamOption, PGPOKeyGenFast(FContext, PGPBoolean(FFastGenerate = TRUE))
      ]);
    if Result <> 0 then Exit;
    try
      EntropyNeeded := EntropyNeeded + PGPGetKeyEntropyNeeded(FContext, OptionList, PGPOLastOption(FContext));
    finally
      PGPFreeOptionList(OptionList);
    end;
  end;
  if FFailNoEntropy and (EntropyNeeded > PGPGlobalRandomPoolGetEntropy) then begin
    Result := kPGPError_OutOfEntropy;
    Exit;
  end;
  // PGP 8 returns kPGPError_OutOfEntropy if we don't call CollectRandomDataDialog() in any case
  Result := CollectRandomDataDialog(FContext, EntropyNeeded, FParentHandle);
end;

function GetPGPCipherAlgorithm(CipherAlgorithm: TCipherAlgorithm): PGPCipherAlgorithm;
begin
  Result := PGPCipherAlgorithm(
    Ord(CipherAlgorithm) +
    Ord(CipherAlgorithm > CipherAlgorithm_CAST5) * Ord(pred(kPGPCipherAlgorithm_AES128 - kPGPCipherAlgorithm_CAST5))
  );
end;

function TPGPKeysGenerateCustom.KeyGenerate(IncludeSubKey: Longbool): PGPError;
var
  OptionList	: pPGPOptionList;
  CipherAlgs	: TPGPCipherAlgorithms;
  CipherIndex	: Longint;
  UsageOption	: pPGPOptionList;
  UserIDUtf8	: AnsiString;
  NewKey	: pPGPKey;
begin
  OptionList := NIL;
  if FCipherAlgorithmList[0] = CipherAlgorithm_None then begin
    case FCipherAlgorithm of
      CipherAlgorithm_None: begin
	case FPubKeyAlgorithm of
	  KeyAlgorithm_RSA: FCipherAlgorithm := CipherAlgorithm_IDEA;
	  KeyAlgorithm_DSS: FCipherAlgorithm := CipherAlgorithm_CAST5;
	end;
      end;
    end;
    CipherAlgs[0] := GetPGPCipherAlgorithm(FCipherAlgorithm);
    CipherIndex := 1;
  end
  else begin
    CipherIndex := 0;
    while FCipherAlgorithmList[CipherIndex] <> CipherAlgorithm_None do begin
      CipherAlgs[CipherIndex] := GetPGPCipherAlgorithm(FCipherAlgorithmList[CipherIndex]);
      inc(CipherIndex);
    end;
  end;
  if (SDKVersion >= PGP7XX) and not FLegacyKey then
    UsageOption := PGPOKeyFlags(FContext, kPGPKeyPropertyFlags_UsageSign)
  else UsageOption := PGPONullOption(FContext);
  if SDKVersion >= PGP8XX then
    UserIDUtf8 := AnsiToUtf8(FUserID)
  else UserIDUtf8 := FUserID;
  if Length(UserIDUtf8) > kPGPMaxUserIDSize then begin
    Result := kPGPError_BadParams;
    Exit;
  end;
  Result := PGPBuildOptionList(FContext, OptionList,
    [
      PGPOKeySetRef(FContext, FKeySetMain),
      PGPOPassphrase(FContext, FPassphrase),
      PGPOKeyGenName(FContext, PAnsiChar(UserIDUtf8), Length(UserIDUtf8)),
      PGPOKeyGenParams(FContext, PGPPublicKeyAlgorithm(FPubKeyAlgorithm), FKeySize),
      PGPOPreferredAlgorithms(FContext, CipherAlgs, CipherIndex),
      PGPOKeyGenFast(FContext, PGPBoolean(FFastGenerate = TRUE)),
      PGPOExpiration(FContext, FExpires),
      UsageOption,
      PGPOEventHandler(FContext, EventHandler, Self)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPGenerateKey(FContext, NewKey, OptionList, PGPOLastOption(FContext));
    if Result <> 0 then Exit;
    if not IncludeSubKey then begin
      Result := KeyRings.UpdateKeyRings;
      if Result <> 0 then Exit;
    end;
    FNewKeyHexID := GetKeyPropKeyID(NewKey);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

function TPGPKeysGenerateCustom.SubKeyGenerate(ForOldMasterKey: Longbool): PGPError;
var
  OptionList	: pPGPOptionList;
  MasterKey	: pPGPKey;
  UsageOption	: pPGPOptionList;
  NewSubKey	: pPGPKey;
begin
  OptionList := NIL;
  Result := GetKeyByHexID(FKeySetMain, FMasterKeyHexID, MasterKey);
  if Result <> 0 then Exit;
  if (SDKVersion >= PGP7XX) and not FLegacyKey then
    UsageOption := PGPOKeyFlags(FContext, kPGPKeyPropertyFlags_UsageEncrypt)
  else UsageOption := PGPONullOption(FContext);
  Result := PGPBuildOptionList(FContext, OptionList,
    [
      PGPOKeyGenMasterKey(FContext, MasterKey),
      PGPOPassphrase(FContext, FPassphrase),
      PGPOKeyGenParams(FContext, PGPPublicKeyAlgorithm(FPubKeyAlgorithm), FSubKeySize),
      PGPOKeyGenFast(FContext, PGPBoolean(FFastGenerate = TRUE)),
      PGPOExpiration(FContext, FExpires),
      UsageOption,
      PGPOEventHandler(FContext, EventHandler, Self)
    ]);
  if Result <> 0 then Exit;
  try
    Result := PGPGenerateSubKey(FContext, NewSubKey, OptionList, PGPOLastOption(FContext));
    if Result <> 0 then Exit;
    Result := KeyRings.UpdateKeyRings;
    if Result <> 0 then Exit;
    if ForOldMasterKey then FUserID := GetKeyPropUserID(MasterKey);
    FNewSubKeyHexID := GetSubKeyPropKeyID(NewSubKey);
  finally
    PGPFreeOptionList(OptionList);
  end;
end;

procedure TPGPKeysGenerateCustom.SetCipherAlgPrefs(const CipherAlgList: Array of TCipherAlgorithm);
var
  CipherCount	: Longint;
  CipherIndex	: Longint;
begin
  CipherCount := succ(High(CipherAlgList));
  if CipherCount > High(TCipherCount) then CipherCount := High(TCipherCount);
  for CipherIndex := 0 to pred(CipherCount) do FCipherAlgorithmList[CipherIndex] := CipherAlgList[CipherIndex];
  FCipherAlgorithmList[CipherCount] := CipherAlgorithm_None;
end;

function TPGPKeysGenerateCustom.DHDSSKeyGenerate: Longint;
begin
  Result := 0;
  Result := InitKeyGen;
  if Result <> 0 then Exit;
  try
    Result := GetUserNameAddress;
    if Result <> 0 then Exit;
    if (FKeySize > 1024) and (SDKVersion < PGP95X) then FKeySize := 1024;
    FPubKeyAlgorithm := KeyAlgorithm_DSS;
    FUserID := GetUserID;
    Result := GetPassphrase(FALSE);
    if Result <> 0 then Exit;
    Result := GetEntropy(TRUE);
    if Result <> 0 then Exit;
    Result := KeyGenerate(TRUE);
    if Result <> 0 then Exit;
    FPubKeyAlgorithm := KeyAlgorithm_DH;
    FMasterKeyHexID := FNewKeyHexID;
    Result := SubKeyGenerate(FALSE);
    if Result = 0 then FNewHexID := FNewKeyHexID;
  finally
    FinitKeyGen(Result);
  end;
end;

function TPGPKeysGenerateCustom.DHSubKeyGenerate: Longint;
begin
  Result := 0;
  Result := InitKeyGen;
  if Result <> 0 then Exit;
  try
    Result := GetMasterKeyHexID;
    if Result <> 0 then Exit;
    FPubKeyAlgorithm := KeyAlgorithm_DH;
    Result := GetPassphrase(TRUE);
    if Result <> 0 then Exit;
    Result := GetEntropy(FALSE);
    if Result <> 0 then Exit;
    Result := SubKeyGenerate(TRUE);
    if Result = 0 then FNewHexID := FNewSubKeyHexID;
  finally
    FinitKeyGen(Result);
  end;
end;

function TPGPKeysGenerateCustom.DSAKeyGenerate: Longint;
begin
  Result := 0;
  Result := InitKeyGen;
  if Result <> 0 then Exit;
  try
    Result := GetUserNameAddress;
    if Result <> 0 then Exit;
    if (FKeySize > 1024) and (SDKVersion < PGP95X) then FKeySize := 1024;
    FPubKeyAlgorithm := KeyAlgorithm_DSS;
    FUserID := GetUserID;
    Result := GetPassphrase(FALSE);
    if Result <> 0 then Exit;
    Result := GetEntropy(FALSE);
    if Result <> 0 then Exit;
    Result := KeyGenerate(FALSE);
    if Result = 0 then FNewHexID := FNewKeyHexID;
  finally
    FinitKeyGen(Result);
  end;
end;

function TPGPKeysGenerateCustom.RSASubKeyGenerate: Longint;
begin
  Result := 0;
  Result := InitKeyGen;
  if Result <> 0 then Exit;
  try
    Result := GetMasterKeyHexID;
    if Result <> 0 then Exit;
    FPubKeyAlgorithm := KeyAlgorithm_RSA;
    Result := GetPassphrase(TRUE);
    if Result <> 0 then Exit;
    Result := GetEntropy(FALSE);
    if Result <> 0 then Exit;
    Result := SubKeyGenerate(TRUE);
    if Result = 0 then FNewHexID := FNewSubKeyHexID;
  finally
    FinitKeyGen(Result);
  end;
end;

function TPGPKeysGenerateCustom.RSAKeyGenerate(Legacy: Longbool): Longint;
begin
  if Legacy or (SDKVersion >= PGP7XX) then begin
    Result := 0;
    Result := InitKeyGen;
    if Result <> 0 then Exit;
    try
      Result := GetUserNameAddress;
      if Result <> 0 then Exit;
      FPubKeyAlgorithm := KeyAlgorithm_RSA;
      FUserID := GetUserID;
      FLegacyKey := Legacy;
      Result := GetPassphrase(FALSE);
      if Result <> 0 then Exit;
      Result := GetEntropy(not Legacy);
      if Result <> 0 then Exit;
      Result := KeyGenerate(not Legacy);
      if Result <> 0 then Exit;
      if not Legacy then begin
	FPubKeyAlgorithm := KeyAlgorithm_RSA;
	FMasterKeyHexID := FNewKeyHexID;
	Result := SubKeyGenerate(FALSE);
      end;
      if Result = 0 then FNewHexID := FNewKeyHexID;
    finally
      FinitKeyGen(Result);
    end;
  end
  else Result := kPGPError_UnknownKeyVersion;
end;


end.

