{------------------------------------------------------------------------------}
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2005-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,R-,Z4}
unit PSMimeTools;

interface

uses
  pgpBase, Isaac, Base64, PGPStrTools, KeyPropTypes, X509Types;

type
  TPSMime = (psNone, psPGP, psSigned, psEncrypted, psClearSigned);

function GetPSMimeType(const MimeMessage: RawByteString; var Headers, Body, Signature: RawByteString;
		       var Charset: AnsiString; AnalyseOnly: Longbool): TPSMime;
function SMimeBuild(const MimeMessage, Signature: RawByteString; var Separator: AnsiString; var BodyOffset: Longint;
		    HashAlgorithm: THashAlgorithm; Encrypted, OmitMimeVersion: Longbool): RawByteString;
function PGPMimeBuild(const MimeMessage: RawByteString; var Separator: AnsiString; var BodyOffset: Longint;
		      HashAlgorithm: THashAlgorithm; Encrypted, OmitMimeVersion: Longbool): RawByteString;
function InsertCharset(const Charset: AnsiString; const MessageData: RawByteString; ClearSigned: Longbool): RawByteString;


implementation

uses
  Windows, SysUtils
  {$IFDEF UNICODE},
  AnsiStrings
  {$ENDIF};

const
  MIME_VERSION	= 'Mime-Version: 1.0';
  BOUNDARY	= '-=-=-Boundary-=-=-';
  PGPMIME	= 'PGPMime';
  SMIME		= 'SMime';

  APPDATA	= 'application/';
  MULTIPART	= 'multipart/';
  PROTOCOL	= 'protocol=';

  B64		= 'base64';
  BOUNDS	= 'boundary';

  CD_HEADER	= 'Content-Disposition:';
  CT_HEADER	= 'Content-Type:';
  CTE_HEADER	= 'Content-Transfer-Encoding:';

  PGP		= 'pgp-';
  PM_ENC	= 'encrypted';
  PM_SIG	= 'signature';
  PM_KEYS	= 'keys';

  MIME		= 'mime';
  PKCS7		= 'pkcs7-';
  SM_TYPE	= 'smime-type';
  SM_ENC	= 'enveloped';
  SM_SIG	= 'signed';
  SM_CERT	= 'certs';
  DATA		= '-data';

  PGP_BEG	= '-----BEGIN PGP';
  PGP_END	= '-----END PGP';
  PGP_MSG	= ' MESSAGE';
  PGP_SIGNED	= ' SIGNED';
  PGP_SIGNATURE	= ' SIGNATURE';
  DASHES	= '-----';

  DASH		= '-';
  DD		= DASH + DASH;

  CHAR_SET	= 'Charset:';

  MAX_HEAD_LEN	= 76;

var
  RNG: TIsaac;

function AdjustLineBreaks(const Text: RawByteString): RawByteString;
type
  TCRLF = Array[0..1] of AnsiChar;
  PCRLF = ^TCRLF;
var
  iBeg, iEnd, iLen: Integer;
  pResult: PAnsiChar;
begin
  if Text <> E then begin
    SetString(Result, NIL, Length(Text) shl 1);
    pResult := Pointer(Result);
    iBeg := 1;
    iEnd := iBeg;
    repeat
      while not (Text[iEnd] in [NUL, LF, CR]) do inc(iEnd);
      if (Text[iEnd] = CR) and (Text[succ(iEnd)] = LF) then begin
	inc(iEnd, Length(CRLF));
	Continue;
      end
      else begin
	iLen := iEnd - iBeg;
	Move(Text[iBeg], pResult^, iLen);
	inc(pResult, iLen);
	if Text[iEnd] <> NUL then begin
	  PCRLF(pResult)^ := CRLF;
	  inc(pResult, Length(CRLF));
	end
	else begin
	  Delete(Result, succ(pResult - Pointer(Result)), MAXINT);
	  Exit;
	end;
	inc(iEnd);
	iBeg := iEnd;
      end;
    until Text = E;
  end;
  Result := E;
end;

function ExtractMimeHeaders(var MimeMessage: RawByteString; CopyOnly: Longbool): RawByteString;
var
  iLen, iPos: Integer;
begin
  iPos := FirstStrPos(PARA, MimeMessage);
  if iPos > 0 then begin
    Result := Copy(MimeMessage, 1, succ(iPos));
    if not CopyOnly then begin
      iLen := Length(MimeMessage);
      while (iPos <= iLen) and (MimeMessage[iPos] <= SP) do inc(iPos);
      Delete(MimeMessage, 1, pred(iPos));
    end;
  end
  else begin
    if (MimeMessage <> E) and not StrAtPos(CRLF, Result, pred(Length(MimeMessage))) then
      Result := MimeMessage + CRLF
    else Result := MimeMessage;
    MimeMessage := E;
  end;
end;

function TrimCopy(const Text: RawByteString; BegPos, EndPos: Integer): RawByteString;
var
  iLen: Integer;
begin
  iLen := Length(Text);
  if BegPos < 1 then BegPos := 1;
  if EndPos > iLen then EndPos := iLen;
  while (BegPos <= EndPos) and (Text[BegPos] <= SP) do inc(BegPos);
  while (EndPos >= BegPos) and (Text[EndPos] <= SP) do dec(EndPos);
  iLen := succ(EndPos - BegPos);
  if iLen > 0 then begin
    SetString(Result, NIL, iLen);
    StrLCopy(PAnsiChar(Result), @Text[BegPos], iLen);
  end
  else Result := E;
end;

function GetHeaderData(const HeaderName, Headers: RawByteString): RawByteString;
var
  iFullLen, iHdrLen, iBegPos, iEndPos: Integer;
begin
  iFullLen := Length(Headers);
  iHdrLen := Length(HeaderName);
  iBegPos := 0;
  Result := E;
  repeat
    iBegPos := NextLoStrPos(HeaderName, Headers, succ(iBegPos));
    if (iBegPos = 1) or ((iBegPos > 1) and (Headers[pred(iBegPos)] = LF)) then begin
      iEndPos := NextLoStrPos(LF, Headers, iBegPos + iHdrLen);
      while (iEndPos < iFullLen) and (Headers[succ(iEndPos)] in [SP, TAB]) do begin
	iEndPos := NextCharPos(LF, Headers, succ(iEndPos));
      end;
      if iEndPos = 0 then iEndPos := iFullLen;
      if iEndPos > iBegPos then begin
	inc(iEndPos);
	inc(iBegPos, iHdrLen);
	Result := TrimCopy(Headers, iBegPos, iEndPos);
      end;
    end;
  until (Result <> E) or (iBegPos = 0);
end;

function GetApplicationType(const HeaderData: RawByteString; var AppTypePos: Integer): RawByteString;
var
  iBegPos, iEndPos: Integer;
begin
  Result := E;
  iBegPos := FirstLoStrPos(APPDATA, HeaderData);
  if iBegPos > 0 then begin
    inc(iBegPos, Length(APPDATA));
    iEndPos := NextCharPos(SC, HeaderData, iBegPos);
    if iEndPos = 0 then iEndPos := Length(HeaderData);
    Result := TrimCopy(HeaderData, iBegPos, iEndPos);
    iEndPos := FirstCharPos(QU, Result);
    if iEndPos > 0 then begin
      dec(iEndPos);
      while (iEndPos > 0) and (Result[iEndPos] <= SP) do dec(iEndPos);
      if iEndPos > 0 then
	Delete(Result, succ(iEndPos), MAXINT)
      else Result := E;
    end;
  end;
  AppTypePos := iBegPos;
end;

function GetBoundary(const HeaderData: RawByteString): RawByteString;
var
  iBegPos, iEndPos, iQUPos: Integer;
begin
  Result := E;
  iBegPos := NextLoStrPos(BOUNDS, HeaderData, succ(Length(MULTIPART)));
  if iBegPos > 0 then begin
    iBegPos := NextCharPos(EQ, HeaderData, iBegPos + Length(BOUNDS));
    if iBegPos > 0 then begin
      inc(iBegPos);
      iEndPos := NextCharPos(SC, HeaderData, succ(iBegPos));
      if iEndPos <> 0 then
	dec(iEndPos)
      else iEndPos := Length(HeaderData);
      if iEndPos > iBegPos then begin
	iQUPos := iBegPos;
	while (iQUPos < iEndPos) and (HeaderData[iQUPos] <> QU) do inc(iQUPos);
	if iQUPos < iEndPos then begin
	  while (iEndPos > iQUPos) and (HeaderData[iEndPos] <> QU) do dec(iEndPos);
	  if iQUPos < iEndPos then begin
	    iBegPos := succ(iQUPos);
	    dec(iEndPos);
	  end;
	end;
	while (iBegPos < iEndPos) and (HeaderData[iBegPos] <= SP) do inc(iBegPos);
	while (iEndPos > iBegPos) and (HeaderData[iEndPos] <= SP) do dec(iEndPos);
	if iEndPos > iBegPos then Result := '--' + Copy(HeaderData, iBegPos, succ(iEndPos - iBegPos));
      end;
    end;
  end;
end;

function DeleteHeaders(const HeaderNames, Headers: RawByteString): RawByteString;
var
  sNameBuffer, sName : RawByteString;
  iBufferLen, iCommaPos, iFullLen, iNameLen, iBegPos, iEndPos: Integer;
begin
  sNameBuffer := HeaderNames;
  Result := Headers;
  repeat
    iBufferLen := Length(sNameBuffer);
    iCommaPos := PrevCharPos(CM, HeaderNames, iBufferLen);
    sName := TrimCopy(sNameBuffer, succ(iCommaPos), iBufferLen);
    Delete(sNameBuffer, iCommaPos, MAXINT);
    iFullLen := Length(Result);
    iNameLen := Length(sName);
    iBegPos := 0;
    repeat
      iBegPos := NextLoStrPos(sName, Result, succ(iBegPos));
      if (iBegPos = 1) or ((iBegPos > 1) and (Result[pred(iBegPos)] = LF)) then begin
	iEndPos := NextLoStrPos(LF, Result, iBegPos + iNameLen);
	while (iEndPos < iFullLen) and (Result[succ(iEndPos)] in [SP, TAB]) do begin
	  iEndPos := NextCharPos(LF, Result, succ(iEndPos));
	end;
	if iEndPos = 0 then iEndPos := iFullLen;
	if iEndPos > iBegPos then begin
	  Delete(Result, iBegPos, succ(iEndPos - iBegPos));
	  Break;
	end;
      end;
    until iBegPos = 0;
  until iCommaPos = 0;
end;

function FindValidBoundary(const Boundary, Data: RawByteString; StartPos: Integer): Integer;
begin
  Result := NextStrPos(Boundary, Data, StartPos);
  while (Result > 1) and (Data[pred(Result)] <> LF) do begin
    Result := NextStrPos(Boundary, Data, Result + Length(Boundary));
  end;
end;

function GetMimePart(const Boundary: RawByteString; var MimeBody: RawByteString): RawByteString;
var
  iBegPos, iEndPos: Integer;
begin
  Result := E;
  iBegPos := FindValidBoundary(Boundary, MimeBody, 1);
  if iBegPos > 0 then begin
    iEndPos := FindValidBoundary(Boundary, MimeBody, iBegPos + Length(Boundary));
    if iEndPos > iBegPos then begin
      dec(iEndPos, Length(CRLF));
      inc(iBegPos, Length(Boundary) + Length(CRLF));
      Result := Copy(MimeBody, iBegPos, iEndPos - iBegPos);
      Delete(MimeBody, 1, iEndPos + pred(Length(CRLF)));
    end;
  end;
end;

function DecodeMimePart(const MimePart: RawByteString): RawByteString;
var
  sHeaders: RawByteString;
begin
  Result := MimePart;
  sHeaders := ExtractMimeHeaders(Result, FALSE);
  if LowerCase(GetHeaderData(CTE_HEADER, sHeaders)) = B64 then begin
    Result := MimeDecodeString(Trim(Result));
  end;
end;

function WrapHeaderLine(const Header: RawByteString): RawByteString;
var
  iLen, iResPos, iBegPos, iEndPos, iCopyLen: Integer;
  pResult: PAnsiChar;
begin
  iLen := Length(Header);
  if (iLen > MAX_HEAD_LEN) and (FirstStrPos(CRLF, Header) = 0) then begin
    SetString(Result, NIL, iLen + (GetCharCount(SC, Header) * (Length(CRLF) + Length(SP))));
    pResult := Pointer(Result);
    iResPos := 0;
    iBegPos := 1;
    repeat
      iEndPos := NextCharPos(SC, Header, iBegPos);
      iCopyLen := succ(iEndPos - iBegPos);
      if (iCopyLen > 0) and (iCopyLen + (iLen - iEndPos) > MAX_HEAD_LEN) then begin
	Move(Header[iBegPos], pResult[iResPos], iCopyLen);
	inc(iResPos, iCopyLen);
	pResult[iResPos] := CR;
	inc(iResPos);
	pResult[iResPos] := LF;
	inc(iResPos);
	iBegPos := succ(iEndPos);
	if (iBegPos <= iLen) and not (Header[iBegPos] in [TAB, SP]) then begin
	  pResult[iResPos] := SP;
	  inc(iResPos);
	end;
	Continue;
      end
      else begin
	iCopyLen := succ(iLen - iBegPos);
	if iCopyLen > 0 then begin
	  Move(Header[iBegPos], pResult[iResPos], iCopyLen);
	  inc(iResPos, iCopyLen);
	end;
	Break;
      end;
    until iBegPos > iLen;
    Delete(Result, succ(iResPos), MAXINT);
  end
  else Result := Header;
end;

function GetPGPData(const BodyPart: RawByteString): RawByteString;
var
  iBegPos, iEndPos: Integer;
begin
  Result := E;
  if StrAtPos(PGP_BEG, BodyPart, 1) then
    iBegPos := 1
  else begin
    iBegPos := FirstStrPos(LF + PGP_BEG, BodyPart);
    if iBegPos > 0 then inc(iBegPos);
  end;
  if iBegPos > 0 then begin
    iEndPos := NextStrPos(LF + PGP_END, BodyPart, iBegPos + Length(PGP_BEG));
    if iEndPos > iBegPos then begin
      iEndPos := NextStrPos(DASHES, BodyPart, iEndPos + Length(PGP_END));
      if iEndPos > 0 then begin
	inc(iEndPos, Length(DASHES));
	while BodyPart[iEndPos] in [NUL, LF, CR] do inc(iEndPos);
      end;
      Result := Copy(BodyPart, iBegPos, iEndPos - iBegPos);
    end;
  end;
end;

function GetCharset(const MessageData: RawByteString): RawByteString;
var
  iLen, iBegPos, iEndPos: Integer;
begin
  iLen := Length(MessageData);
  if StrAtPos(PGP_BEG, MessageData, 1) then
    iBegPos := 1
  else iBegPos := FirstStrPos(LF + PGP_BEG, MessageData);
  if iBegPos > 0 then begin
    if iBegPos > 1 then
      inc(iBegPos, succ(Length(PGP_BEG)))
    else inc(iBegPos, Length(PGP_BEG));
    if StrAtPos(PGP_SIGNED, MessageData, iBegPos) then begin
      inc(iBegPos, Length(PGP_SIGNED) + Length(PGP_MSG) + Length(DASHES) + Length(CRLF));
      iBegPos := NextLoStrPos(LF + PGP_BEG, MessageData, iBegPos);
      if (iBegPos > 0) and StrAtPos(PGP_SIGNATURE, MessageData, iBegPos + succ(Length(PGP_BEG))) then begin
	inc(iBegPos, succ(Length(PGP_BEG) + Length(PGP_SIGNATURE) + Length(DASHES) + LENGTH(CRLF)));
      end;
    end
    else if StrAtPos(PGP_MSG, MessageData, iBegPos) then inc(iBegPos, Length(PGP_MSG) + Length(DASHES) + LENGTH(CRLF));
    if iBegPos > 0 then begin
      repeat
	if (iBegPos < iLen) and (MessageData[iBegPos] <> CR) then begin
	  if LoStrAtPos(CHAR_SET, MessageData, iBegPos) then begin
	    inc(iBegPos, Length(CHAR_SET));
	    while (MessageData[iBegPos] <> NUL) and (MessageData[iBegPos] = SP) do inc(iBegPos);
	    iEndPos := iBegPos;
	    while (MessageData[iEndPos] <> NUL) and (MessageData[iEndPos] >= SP) do inc(iEndPos);
	    Result := UpperCase(Copy(MessageData, iBegPos, iEndPos - iBegPos));
	    Exit;
	  end
	  else begin
	    while not (MessageData[iBegPos] in [NUL, LF]) do inc(iBegPos);
	    if MessageData[iBegPos] = LF then
	      inc(iBegPos)
	    else Break;
	  end;
	end
	else Break;
      until MessageData[iBegPos] = NUL;
    end;
  end;
  Result := E;
end;

function GetPSMimeType(const MimeMessage: RawByteString; var Headers, Body, Signature: RawByteString;
		       var Charset: AnsiString; AnalyseOnly: Longbool): TPSMime;
var
  sMimeParts, sCTHeader, sAppType, sBoundary: RawByteString;
  iAppTypePos, iDataPos: Integer;
begin
  Result := psNone;
  Signature := E;
  Body := AdjustLineBreaks(MimeMessage);
  Headers := ExtractMimeHeaders(Body, AnalyseOnly);
  sCTHeader := GetHeaderData(CT_HEADER, Headers);
  sAppType := GetApplicationType(sCTHeader, iAppTypePos);
  if FirstLoStrPos(PKCS7, sAppType) > 0 then begin
    if FirstLoStrPos(MULTIPART + SM_SIG, sCTHeader) > 0 then begin
      if not AnalyseOnly then begin
	sBoundary := GetBoundary(sCTHeader);
	if sBoundary <> E then begin
	  sMimeParts := Body;
	  Headers := DeleteHeaders(CT_HEADER + CM + CTE_HEADER + CM + CD_HEADER, Headers);
	  Body := GetMimePart(sBoundary, sMimeParts);
	  Signature := DecodeMimePart(GetMimePart(sBoundary, sMimeParts));
	end;
      end;
      Result := psClearSigned;
    end
    else begin
      iDataPos := NextLoStrPos(SM_TYPE, sCTHeader, iAppTypePos + succ(Length(PKCS7 + MIME)));
      if iDataPos > 0 then begin
	inc(iDataPos, succ(Length(SM_TYPE)));
	if (NextLoStrPos(SM_ENC, sCTHeader, iDataPos) > 0) then begin
	  if not AnalyseOnly then begin
	    Headers := DeleteHeaders(CT_HEADER + CM + CTE_HEADER + CM + CD_HEADER, Headers);
	    Body := DecodeMimePart(MimeMessage);
	  end;
	  Result := psEncrypted;
	end
	else if (NextLoStrPos(SM_SIG, sCTHeader, iDataPos) > 0)
	or (NextLoStrPos(SM_CERT, sCTHeader, iDataPos) > 0) then begin
	  if not AnalyseOnly then begin
	    Headers := DeleteHeaders(CT_HEADER + CM + CTE_HEADER + CM + CD_HEADER, Headers);
	    Body := DecodeMimePart(MimeMessage);
	  end;
	  Result := psSigned;
	end;
      end;
    end;
  end
  else if FirstLoStrPos(PGP, sAppType) > 0 then begin
    if not AnalyseOnly then begin
      if FirstLoStrPos(PM_SIG, sAppType) > 0 then begin
	sBoundary := GetBoundary(sCTHeader);
	if sBoundary <> E then begin
	  Headers := DeleteHeaders(CT_HEADER + CM + CTE_HEADER + CM + CD_HEADER, Headers);
	  Body := WrapHeaderLine(CT_HEADER + SP + sCTHeader) + PARA + Body;
	end;
      end
      else if (FirstLoStrPos(PM_ENC, sAppType) > 0) or (FirstLoStrPos(PM_KEYS, sAppType) > 0) then begin
	Headers := DeleteHeaders(CT_HEADER + CM + CTE_HEADER + CM + CD_HEADER, Headers);
	Body := GetPGPData(Body);
      end;
    end;
    Result := psPGP;
  end
  else Charset := GetCharset(MimeMessage);
end;

function GetHashAlgorithmString(HashAlgorithm: THashAlgorithm): RawByteString;
const
  MICALG = 'micalg=';
begin
  case HashAlgorithm of
    HashAlgorithm_MD5: Result := MICALG + 'md5';
    HashAlgorithm_SHA: Result := MICALG + 'sha1';
    HashAlgorithm_RIPEMD160: Result := MICALG + 'ripemd160';
    HashAlgorithm_SHA256: Result := MICALG + 'sha256';
    HashAlgorithm_SHA384: Result := MICALG + 'sha384';
    HashAlgorithm_SHA512: Result := MICALG + 'sha512';
  else
    Result := E;
  end;
end;

function SMimeBuild(const MimeMessage, Signature: RawByteString; var Separator: AnsiString; var BodyOffset: Longint;
		    HashAlgorithm: THashAlgorithm; Encrypted, OmitMimeVersion: Longbool): RawByteString;
const
  ATTACHMENT_FILE = 'attachment; file';
  NAME_SMIME = 'name=SMime';
  ENC_EXT = '.p7m';
  SIG_EXT = '.p7s';
  X_ = 'x-';
var
  sMimeVersion, sBoundary, sMainHeaders, sAlgorithm, sSigHeaders: RawByteString;
begin
  if OmitMimeVersion then
    sMimeVersion := E
  else sMimeVersion := MIME_VERSION + CRLF;
  if Encrypted then begin
    sBoundary := E;
    sMainHeaders := sMimeVersion +
		    CT_HEADER + SP + APPDATA + PKCS7 + MIME + SC + CRLF +
		    SP + SM_TYPE + EQ + SM_ENC + DATA + SC + CRLF +
		    SP + NAME_SMIME + ENC_EXT + CRLF +
		    CTE_HEADER + SP + B64 + CRLF +
		    CD_HEADER + SP + ATTACHMENT_FILE + NAME_SMIME + ENC_EXT;
    Result := sMainHeaders + PARA + MimeEncodeString(MimeMessage) + CRLF;
  end
  else begin
    sBoundary := SMIME + BOUNDARY + PGPStrTools.IntToHex(RNG.Val, 8);
    sAlgorithm := GetHashAlgorithmString(HashAlgorithm) + SC + SP;
    sMainHeaders:= sMimeVersion +
		   CT_HEADER + SP + MULTIPART + SM_SIG + SC + SP +
		   PROTOCOL + QU + APPDATA + X_ + PKCS7 + PM_SIG + QU + SC + CRLF +
		   SP + sAlgorithm + BOUNDS + EQ + QU + sBoundary + QU;
    sSigHeaders := CT_HEADER + SP + APPDATA + X_ + PKCS7 + PM_SIG + SC + SP + NAME_SMIME + SIG_EXT + CRLF +
		   CTE_HEADER + SP + B64 + CRLF + CD_HEADER + SP + ATTACHMENT_FILE + NAME_SMIME + SIG_EXT;
    Result := sMainHeaders + PARA + DD + sBoundary + CRLF +
	      MimeMessage + CRLF + DD + sBoundary + CRLF +
	      sSigHeaders + PARA + MimeEncodeString(Signature) + CRLF + DD + sBoundary + DD + CRLF;
  end;
  BodyOffset := Length(sMainHeaders) + pred(Length(PARA));
  Separator := sBoundary;
end;

function PGPMimeBuild(const MimeMessage: RawByteString; var Separator: AnsiString; var BodyOffset: Longint;
		      HashAlgorithm: THashAlgorithm; Encrypted, OmitMimeVersion: Longbool): RawByteString;
var
  sMimeVersion, sBoundary, sMainHeaders, sAlgorithm: RawByteString;
  iSepLen, iMsgLen, iBoundsLen, iHeaderLen, iCopyLen: Integer;
  iDestPos, iSourcePos, iBoundsPos: Integer;
  pResult: PAnsiChar;
begin
  if OmitMimeVersion then
    sMimeVersion := E
  else sMimeVersion := MIME_VERSION + CRLF;
  sBoundary := PGPMIME + BOUNDARY + PGPStrTools.IntToHex(RNG.Val, 8);
  iSepLen := Length(Separator);
  iMsgLen := Length(MimeMessage);
  iBoundsLen := Length(sBoundary);
  if Encrypted then begin
    sMainHeaders := sMimeVersion +
		    CT_HEADER + SP + MULTIPART + PM_ENC + SC + SP +
		    PROTOCOL + QU + APPDATA + PGP + PM_ENC + QU + SC + CRLF +
		    SP + BOUNDS + EQ + QU + sBoundary + QU + PARA;
  end
  else begin
    sAlgorithm := GetHashAlgorithmString(HashAlgorithm) + SC + SP;
    sMainHeaders := sMimeVersion +
		    CT_HEADER + SP + MULTIPART + SM_SIG + SC + SP +
		    PROTOCOL + QU + APPDATA + PGP + PM_SIG + QU + SC + CRLF +
		    SP + sAlgorithm + BOUNDS + EQ + QU + sBoundary + QU + PARA;
  end;
  iHeaderLen := Length(sMainHeaders);
  SetString(Result, NIL, iMsgLen + iHeaderLen + (iBoundsLen shl 2));
  pResult := Pointer(Result);
  iDestPos := 0;
  Move(Pointer(sMainHeaders)^, pResult[iDestPos], iHeaderLen);
  inc(iDestPos, iHeaderLen);
  iSourcePos := succ(BodyOffset);
  BodyOffset := 0;
  repeat
    iBoundsPos := NextStrPos(Separator, MimeMessage, iSourcePos);
    if iBoundsPos > 0 then begin
      iCopyLen := iBoundsPos - iSourcePos;
      Move(MimeMessage[iSourcePos], pResult[iDestPos], iCopyLen);
      if BodyOffset = 0 then BodyOffset := iDestPos;
      inc(iDestPos, iCopyLen);
      Move(Pointer(sBoundary)^, pResult[iDestPos], iBoundsLen);
      inc(iDestPos, iBoundsLen);
      inc(iSourcePos, iCopyLen + iSepLen);
    end;
  until iBoundsPos = 0;
  pResult[iDestPos] := DASH;
  inc(iDestPos);
  pResult[iDestPos] := DASH;
  inc(iDestPos);
  pResult[iDestPos] := CR;
  inc(iDestPos);
  pResult[iDestPos] := LF;
  inc(iDestPos);
  Delete(Result, succ(iDestPos), MAXINT);
  Separator := sBoundary;
end;

function InsertCharset(const Charset: AnsiString; const MessageData: RawByteString; ClearSigned: Longbool): RawByteString;
var
  iPos: Integer;
begin
  Result := MessageData;
  if MessageData <> E then begin
    if ClearSigned then begin
      iPos := LastStrPos(LF + PGP_BEG + PGP_SIGNATURE, Result);
      iPos := NextStrPos(PARA, Result, iPos + Length(PGP_BEG) + Length(PGP_SIGNATURE) + Length(DASHES) + Length(CRLF));
      Insert(CRLF + CHAR_SET + SP + Charset, Result, iPos);
    end
    else begin
      iPos := FirstStrPos(PGP_BEG + PGP_MSG, Result);
      iPos := NextStrPos(PARA, Result, iPos + Length(PGP_BEG) + Length(PGP_MSG) + Length(DASHES) + Length(CRLF));
      Insert(CRLF + CHAR_SET + SP + Charset, Result, iPos);
    end;
  end;
end;

procedure RNGInit;
var
  stLocal: TSystemTime;
  ptCursor: TPoint;
begin
  GetLocalTime(stLocal);
  GetCursorPos(ptCursor);
  with stLocal, ptCursor do begin
    RNG := TIsaac.Create;
    RNG.Seed([MakeLong(wMonth, wSecond), MakeLong(wHour, wDay),
	      MakeLong(wMilliseconds, wYear), MakeLong(wDayOfWeek, wMinute),
	      MakeLong(X, Y), GetMessageTime, GetMessagePos, GetTickCount], TRUE);
  end;
end;

procedure RNGFinit;
begin
  RNG.Free;
end;


initialization

  RNGInit;


finalization

  RNGFinit;


end.

