{------------------------------------------------------------------------------}
{                                                                              }
{          UTF8 utilities - especially for versions prior to Delphi 6          }
{                                                                              }
{                                 This code is                                 }
{              Copyright (C) 2003-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$IFDEF CONDITIONALEXPRESSIONS} {$WARN SYMBOL_PLATFORM OFF} {$ENDIF}
{$J+,Z4}
unit UTF8;

// These functions are only required when using PGP 8.x or later
// UTF8Sec.dll provides them for using Delphi versions prior to 6.x
// Secure memory management prevents passphrases from being swapped to disk

interface

uses
  pgpBase, pgpMemoryMgr;

function SecureAnsiToUtf8PChar(const Ansi, Utf8: PAnsiChar; Len: Cardinal): Longint;
function SecureUtf8ToAnsiPChar(const Utf8, Ansi: PAnsiChar; Len: Cardinal): Longint;

function AnsiToUtf8(const S: RawByteString): RawByteString;
function Utf8ToAnsi(const S: RawByteString): RawByteString;

function Utf8OrAnsi(const S: RawByteString): RawByteString;

function StrToWide(const Str: RawByteString): PWideChar;

const
  UTF8Factor = 4;

const // from WinNls.h
  CP_UTF7		= 65000;	// UTF-7 translation
  CP_UTF8		= 65001;	// UTF-8 translation

  MB_PRECOMPOSED	= $00000001;	// use precomposed chars
  MB_COMPOSITE		= $00000002;	// use composite chars
  MB_USEGLYPHCHARS	= $00000004;	// use glyph chars, not ctrl chars
  MB_ERR_INVALID_CHARS	= $00000008;	// error for invalid chars

  WC_COMPOSITECHECK	= $00000200;	// convert composite to precomposed
  WC_DISCARDNS		= $00000010;	// discard non-spacing chars
  WC_SEPCHARS		= $00000020;	// generate separate chars
  WC_DEFAULTCHAR	= $00000040;	// replace w/ default char


implementation

uses
  Windows, SysUtils;

function SecureAnsiToUtf8PChar(const Ansi, Utf8: PAnsiChar; Len: Cardinal): Longint;
var
  Wide: PWideChar;
  WideLen: Cardinal;
begin
  Result := 0;
  if (Ansi <> NIL) and (Utf8 <> NIL) and (Len > 0) then begin
    WideLen := succ(Len div Utf8Factor);
    Wide := PGPNewSecureData(PGPGetDefaultMemoryMgr, WideLen shl 1, kPGPMemoryMgrFlags_Clear);
    try
      WideLen := MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED or MB_ERR_INVALID_CHARS, Ansi, -1, Wide, WideLen);
      if WideLen > 0 then Result := WideCharToMultiByte(CP_UTF8, 0, Wide, -1, Utf8, Len, NIL, NIL);
    finally
      PGPFreeData(Wide);
    end;
  end;
end;

function SecureUtf8ToAnsiPChar(const Utf8, Ansi: PAnsiChar; Len: Cardinal): Longint;
var
  Wide: PWideChar;
  WideLen: Cardinal;
begin
  Result := 0;
  if (Utf8 <> NIL) and (Ansi <> NIL) and (Len > 0) then begin
    WideLen := succ(Len);
    Wide := PGPNewSecureData(PGPGetDefaultMemoryMgr, WideLen shl 1, kPGPMemoryMgrFlags_Clear);
    try
      WideLen := MultiByteToWideChar(CP_UTF8, 0, Utf8, -1, Wide, WideLen);
      if WideLen > 0 then Result := WideCharToMultiByte(CP_ACP, 0, Wide, -1, Ansi, Len, NIL, NIL);
    finally
      PGPFreeData(Wide);
    end;
  end;
end;

function WideToUtf8String(const Wide: RawByteString): RawByteString; register; assembler;
asm	// EAX = @Wide
  OR	EAX,EAX
  JE	@EXIT
  PUSH	EAX
  PUSH	EDX
  MOV	EAX,[EAX - 04h]
  MOV	EDX,Utf8Factor / 2
  MUL	EDX
  CALL	SYSTEM.@NEWANSISTRING
  POP	EDX
  MOV	[EDX],EAX
  MOV	EDX,EAX
  POP	EAX
  PUSH	EBX
  PUSH	EDX
  XOR	EBX,EBX
  @STR_LOOP:
  MOV	BX,[EAX]
  OR	EBX,EBX
  JE	@DONE
  ADD	EDX,Utf8Factor - 1
  XOR	ECX,ECX
  DEC	EDX
  CMP	EBX,00000800h
  JNB	@3BYTES
  DEC	EDX
  CMP	EBX,00000080h
  JNB	@2BYTES
  DEC	EDX
  JMP	@ASCII
  @3BYTES:
  INC	ECX
  MOV	BYTE PTR [EDX],BL
  AND	BYTE PTR [EDX],03Fh
  OR	BYTE PTR [EDX],080h
  SHR	EBX,6
  DEC	EDX
  @2BYTES:
  INC	ECX
  MOV	BYTE PTR [EDX],BL
  AND	BYTE PTR [EDX],03Fh
  OR	BYTE PTR [EDX],080h
  SHR	EBX,6
  DEC	EDX
  INC	ECX
  MOV	[EDX],BL
  MOV	BX,00FFh
  ROR	BX,CL
  OR	[EDX],BH
  ADD	EDX,ECX
  ADD	EAX,2
  JMP	@STR_LOOP
  @ASCII:
  MOV	[EDX],BL
  ADD	EAX,2
  INC	EDX
  JMP	@STR_LOOP
  @DONE:
  MOV	BYTE PTR [EDX],0
  POP	EAX
  SUB	EDX,EAX
  MOV	DWORD PTR [EAX - 04h],EDX
  POP	EBX
  RET
  @EXIT:
  MOV	EAX,EDX
  CALL	System.@LStrClr
end;

function Utf8StringToWide(const Utf8: RawByteString): RawByteString; register; assembler;
asm	// EAX = @Utf8
  OR	EAX,EAX
  JE	@EXIT
  PUSH	ESI
  PUSH	EDI
  PUSH	EBX
  MOV	ESI,EAX
  MOV	EDI,EDX
  MOV	EAX,[EAX - 04h]
  SHL	EAX,1
  ADD	EAX,2
  CALL	SYSTEM.@NEWANSISTRING
  MOV	[EDI],EAX
  MOV	EBX,EDI
  MOV	EDI,EAX
  @STR_LOOP:
  XOR	EAX,EAX
  MOV	AL,[ESI]
  CMP	AL,0FDh
  JA	@ZERO
  CMP	AL,080h
  JB	@ASCII
  XOR	ECX,ECX
  @COUNT:
  INC	CL
  SHL	AL,1
  JC	@COUNT
  CMP	CL,3
  JB	@ZERO
  SHR	AL,CL
  DEC	ECX
  ADD	ESI,ECX
  NEG	ECX
  INC	ECX
  @LOOP:
  MOV	DL,[ESI + ECX]
  CMP	DL,080h
  JB	@ZERO
  AND	DL,07Fh
  CMP	DL,040h
  JNB	@ZERO
  SHL	EAX,6
  OR	AL,DL
  INC	ECX
  JNE	@LOOP
  MOV	[EDI],AX
  ADD	EDI,2
  JMP	@STR_LOOP
  @ASCII:
  MOV	[EDI],AX
  ADD	EDI,2
  INC	ESI
  OR	EAX,EAX
  JNE	@STR_LOOP
  @DONE:
  MOV	EBX,[EBX]
  SUB	EDI,EBX
  MOV	DWORD PTR [EBX - 04h],EDI
  POP	EBX
  POP	EDI
  POP	ESI
  RET
  @ZERO:
  MOV	EAX,EBX
  CALL	System.@LStrClr
  POP	EBX
  POP	EDI
  POP	ESI
  RET
  @EXIT:
  MOV	EAX,EDX
  CALL	System.@LStrClr
end;

function AnsiToUtf8(const S: RawByteString): RawByteString;
var
  WideLen: Cardinal;
  Wide: RawByteString;
begin
  if S <> '' then begin
    WideLen := succ(Length(S));
    SetString(Wide, NIL, WideLen shl 1);
    WideLen := MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED or MB_ERR_INVALID_CHARS, Pointer(S), -1, Pointer(Wide), WideLen);
    if WideLen > 0 then begin
      Result := WideToUtf8String(Wide);
      Exit;
    end;
  end;
  Result := '';
end;

function Utf8ToAnsi(const S: RawByteString): RawByteString;
var
  Wide: RawByteString;
  WideLen: Cardinal;
begin
  if S <> '' then begin
    Wide := Utf8StringToWide(S);
    WideLen := Length(Wide) shr 1;
    if WideLen > 1 then begin
      SetString(Result, NIL, pred(WideLen));
      WideLen := WideCharToMultiByte(CP_ACP, 0, Pointer(Wide), -1, Pointer(Result), WideLen, NIL, NIL);
      if WideLen > 0 then Exit;
    end;
  end;
  Result := '';
end;

function Utf8OrAnsi(const S: RawByteString): RawByteString;
begin
  Result := Utf8ToAnsi(S);
  if Result = '' then Result := S;
end;

function StrToWide(const Str: RawByteString): PWideChar;
var
  dwSize: DWord;
begin
  dwSize := succ(Length(Str)) shl 1;
  Result := PWideChar(AllocMem(dwSize));
  if Result <> NIL then StringToWideChar(String(Str), Result, dwSize);
end;


end.

