{------------------------------------------------------------------------------}
{                                                                              }
{                 This unit is partly based on Steve Heller's                  }
{         SPGP sources available from http://www.oz.net/~srheller/spgp/        }
{                                                                              }
{                Portions created by Michael in der Wiesche are                }
{              Copyright (C) 2004-2010 by Michael in der Wiesche               }
{                                                                              }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit X509Funcs; // PGP 7.X and later only

interface

uses
  Windows, Classes, SysUtils, KeyPropTypes, X509Types, KeyFuncs, PGPStrTools, PGPStringList, UTF8,
  pgpBase, pgpErrors, pgpPubTypes, pgpUtilities, pgpOptionList, pgpMemoryMgr, pgpKeys, PGPDialogs;

// returns the last valid certificate issued by SignKey (if specified)
function GetValidX509CertFromKey(const CertUserData: RawByteString; CertKey, SignKey: pPGPKeyDBObj;
				 KeySetToSearch: pPGPKeySet): pPGPKeyDBObj;

// create a key DB with valid encrypting X509 certified keys from the specified keyset
function GetX509EncryptCertDB(Context: pPGPContext; KeySetMain: pPGPKeySet; var KeyDBEncryptCerts: pPGPKeyDB): PGPError;

// check whether all keys in the specified keyset have valid certificates
function CheckValidX509CertifierSet(const KeySetToCheck: pPGPKeySet): PGPError;

// filter the specified keyset for keys with valid/root/signing certificates
function GetX509CertifierSet(const KeySetToFilter: pPGPKeySet; var KeySetValidCertifiers: pPGPKeySet;
			     RootOnly, SignOnly, ValidOnly: Longbool): PGPError;

// dialog for retrieving a signing key and it's passphrase
function X509GetCertAndPassphraseDialog(Context: pPGPContext; KeySetMain: pPGPKeySet;
					var SignKey, SignCert: pPGPKeyDBObj; var Passphrase: PAnsiChar;
					const SignCertUser, SignCertIASN, SignKeyAndPassPrompt: RawByteString;
					WinHandle: Integer): Integer;
{ specifying ...Prompt strings will present the respective PGP dialogs regardless of specifying any ...ID(s) }

// if CertAttributes' CommonName or Email are undefined they are taken from the CertKey's primary user ID
function CreateX509CertificateRequest(const CertKeyAndPassPrompt, CertKeyHexID: RawByteString; ReqKeyPassphrase: PAnsiChar;
				      var CertAttributes: TX509CertAttributes; Armor: Longbool;
				      var Request: RawByteString; var ReqKeyProps: TKeyPropsRec;
				      KeyPropFlags, WinHandle: Integer): Integer;
// specifying a unique CertAttributes' SerialNumber is mandatory, CommonName and Email are taken from the request
// if the SignCertIASN RawByteString contains a SignKeyHexID the key's last valid certificate will be used for certifying
function CreateX509CertificateFromRequest(const SignKeyAndPassPrompt, SignCertIASN: RawByteString; IssuerPassphrase: PAnsiChar;
					  var CertAttributes: TX509CertAttributes; ValidDays: Integer;
					  Armor: Longbool; const Request: RawByteString; var Certificate: RawByteString;
					  var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
// if the RevokeCertIASN RawByteString contains valid CertKeyHexID(s) all certificates on the respective key(s) will be revoked
// if no RevokeCertIASN RawByteString is specified all (non-self) certificates issued by the respective SignKey will be revoked
// if the SignCertIASN RawByteString contains a SignKeyHexID it's last valid certificate will be used for revoking certificates
// if SignCertIASN isn't provided and RevokeCertIASN contains a unique IASN the certifying key can be found automatically
function CreateX509CertificateRevocationList(const CertKeyPrompt, SignKeyAndPassPrompt, RevokeCertIASN, SignCertIASN: RawByteString;
					     RevokerPassphrase: PAnsiChar; ValidDays: Integer;
					     var CertPropsList: TX509CertPropsList;
					     WinHandle: Integer): Integer;
// specifying a unique CertAttributes.SerialNumber is mandatory
// splitting a user ID into CommonName and Email will certify it
// self-signed certificates cannot be created for an existing user ID
function CreateX509Certificate(const CertKeyPrompt, SignKeyAndPassPrompt, CertKeyHexID, SignCertIASN: RawByteString;
			       IssuerPassphrase: PAnsiChar; var CertAttributes: TX509CertAttributes; ValidDays: Integer;
			       SelfSigned: Longbool; var CertPropsList: TX509CertPropsList; WinHandle: Longint): Integer;
// self-signed certificates can only be removed including the user ID they are attached to
// user IDs of certificates can only be removed if they don't have any other signatures assigned
// if there's only a single user ID left on a key the key will be removed from the keyring all together
function RemoveX509Certificate(const CertKeyPrompt, RemoveCertIASN: RawByteString; IncludeUserID: Longbool;
			       var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
// import binary or armored standard X509 certificates or PGP keys to keyring(s)
function ImportX509Certificate(const CertKeyPrompt, CertPassPrompt, CertData: RawByteString;
			       PKCS12Passphrase, CertPassphrase: PAnsiChar; FromFile: Longbool;
			       var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
// either export a single certificate or a certificate chain with parent certificate(s) included
// chains suitable for VerifyX509CertificateChain can be created with Armor disabled
function ExportX509Certificate(const CertKeyOrKeyAndPassPrompt, ExportCertIASN: RawByteString; SecKeyPassphrase: PAnsiChar;
			       Armor, Chain, Secret, ToFile: Longbool; var CertData: RawByteString;
			       var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
// the required certificate chain can be created using ExportX509Certificate() with Chain enabled and Armor disabled
function VerifyX509CertificateChain(const CertChain: RawByteString; FromFile: Longbool): Integer;
// verify certificate chain in given key DB
function VerifyX509CertificateChainDB(CertKeyDB: pPGPKeyDB; var MostRecentCertProps: TX509CertPropsRec): TX509ChainValidity;
// select certificate and return it's properties
function SelectX509CertificateDialog(const CertKeyPrompt: RawByteString;
				     var CertPropsList: TX509CertPropsList;
				     WinHandle: Integer): Integer;
// returns number of certificates found or error
function GetX509CertificateProperties(const CertData: RawByteString; FromFile: Longbool;
				      var CertPropsList: TX509CertPropsList): Integer;
// returns number of keyring certificates or error
// KeyFilterFlags will be applied to any kind of search
// SignKeyID can narrow the search with CertKey...IDs specified
// IncludeSigner is only valid without any CertKey...IDs specified
// in this case you will get the following results depending on SignKeyHexID and IncludeSigner:
// (SignKey = NIL) and (IncludeSigner = FALSE)  => all certified keys on keyring
// (SignKey = NIL) and (IncludeSigner = TRUE)   => all certifier keys on keyring (signing keys only)
// (SignKey <> NIL) and (IncludeSigner = FALSE) => all SignKey-certified keys on keyring without certifier
// (SignKey <> NIL) and (IncludeSigner = TRUE)  => all SignKey-certified keys on keyring including certifier
function FindX509CertProps(const CertKeyUserID, CertKeyHexID, SignKeyHexID: RawByteString; KeyFilterFlags: DWord;
			   IncludeSigner: Longbool; var CertPropsList: TX509CertPropsList): Integer;


implementation

function IsSerialNumberUnique(Context: pPGPContext; KeySetToCheck: pPGPKeySet; SignKey: pPGPKeyDBObj;
			      const SerialNumber: RawByteString; var Unique: Longbool): Integer; forward;
function GetX509CertPropsList(Certificate: pPGPKeyDBObj; var CertPropsList: TX509CertPropsList): Integer; forward;
function GetX509CertSetProps(CertSet, KeySetToSearch: pPGPKeySet; SignKey: pPGPKeyDBObj;
			     const UserID: RawByteString; Filter: DWord; Order: PGPKeyOrdering;
			     var CertPropsList: TX509CertPropsList): Integer; forward;

// convert X509 errors to PGP errors with corresponding messages
function TranslateX509ErrorToPGPError(X509Error: PGPError): PGPError;
begin
  case X509Error of
    k509Error_NeededCertNotAvailable: Result := kPGPError_X509NeededCertNotAvailable;
    k509Error_SelfSignedCert: Result := kPGPError_X509SelfSignedCert;
    k509Error_InvalidCertificateSignature: Result := kPGPError_X509InvalidCertificateSignature;
    k509Error_InvalidCertificateFormat: Result := kPGPError_X509InvalidCertificateFormat;
  else
    Result := X509Error;
  end;
end;

// init main keyset and return filtered set of keys with X509 certificates
function InitX509CertKeySet(var Context: pPGPContext; var KeySetMain, KeySetX509: pPGPKeySet): PGPError;
begin
  Context := NIL;
  KeySetMain := NIL;
  KeySetX509 := NIL;
  Result := KeyRings.InitKeyRings(Context, KeySetMain);
  if Result <> 0 then Exit;
  try
    Result := GetCertKeySetFromKeySet(KeySetMain, FALSE, KeySetX509);
  finally
    if Result <> 0 then begin
      KeyRings.FreeKeyRings;
      KeySetMain := NIL;
      Context := NIL;
    end;
  end;
end;

// finit X509 keyset and free keyrings, convert X509 specific errors
procedure FinitX509CertKeySet(var KeySetX509: pPGPKeySet);
begin
  PGPFreeKeySet(KeySetX509);
  KeyRings.FreeKeyRings;
  KeySetX509 := NIL;
end;

// retrieve a key's primary user ID for using them as CommonName and Email on a certificate
function GetPrimaryUserIDAsAttributes(CertKey: pPGPKeyDBObj; var Name, Address: AnsiString): PGPError;
var
  StrPos, StrLen: Integer;
begin
  try
    Name := GetKeyPropUserID(CertKey);
    StrPos := LastCharPos(SP, Name);
    if StrPos <> 0 then Address := Copy(Name, succ(StrPos), MAXINT);
    Delete(Name, StrPos, MAXINT);
    StrLen := Length(Address);
    if StrLen > 2 then begin
      if Address[StrLen] = '>' then Delete(Address, StrLen, 1);
      if Address[1] = '<' then Delete(Address, 1, 1);
    end;
  finally
    Result := Ord((Name = E) and (Address = E)) * kPGPError_InvalidDistinguishedName;
  end;
end;

// convert Attributes to Utf8
procedure GetUtf8Attributes(var CertAttributes: TX509CertAttributes);
begin
  with CertAttributes do begin
    CommonName := AnsiToUtf8(CommonName);
    Email := AnsiToUtf8(Email);
    OrganizationName := AnsiToUtf8(OrganizationName);
    OrganizationalUnitName := AnsiToUtf8(OrganizationalUnitName);
    SurName := AnsiToUtf8(SurName);
    Country := AnsiToUtf8(Country);
    Locality := AnsiToUtf8(Locality);
    State := AnsiToUtf8(State);
    StreetAddress := AnsiToUtf8(StreetAddress);
    Title := AnsiToUtf8(Title);
    Description := AnsiToUtf8(Description);
    PostalCode := AnsiToUtf8(PostalCode);
    POBOX := AnsiToUtf8(POBOX);
    PhysicalDeliveryOfficeName := AnsiToUtf8(PhysicalDeliveryOfficeName);
    TelephoneNumber := AnsiToUtf8(TelephoneNumber);
    X121Address := AnsiToUtf8(X121Address);
    ISDN := AnsiToUtf8(ISDN);
    DestinationIndicator := AnsiToUtf8(DestinationIndicator);
    Name := AnsiToUtf8(Name);
    GivenName := AnsiToUtf8(GivenName);
    Initials := AnsiToUtf8(Initials);
    HouseIdentifier := AnsiToUtf8(HouseIdentifier);
    DirectoryManagementDomain := AnsiToUtf8(DirectoryManagementDomain);
    DomainComponent := AnsiToUtf8(DomainComponent);
    UnstructuredName := AnsiToUtf8(UnstructuredName);
    UnstructuredAddress := AnsiToUtf8(UnstructuredAddress);
    RFC822Name := AnsiToUtf8(RFC822Name);
    DNSName := AnsiToUtf8(DNSName);
    AnotherName := AnsiToUtf8(AnotherName);
    IPAddress := AnsiToUtf8(IPAddress);
  end;
end;

// convert attribute record to option list parameter using risky type cast ...
function GetAttributeListFromRecord(var CertAttributes: TX509CertAttributes;
				    var AttributeList: TX509CertAttrArray;
				    var AttributeCount: Longint): PGPError;
var
  AttributeAddr	: ^PAnsiChar;
  AttributeIndex: PGPUInt32;
begin
  Result := 0;
  AttributeCount := 0;
  try
    AttributeAddr := @CertAttributes;
    GetUtf8Attributes(CertAttributes);
    FillChar(AttributeList, SizeOf(TX509CertAttrArray), 0);
    for AttributeIndex := 0 to pred(SizeOf(CertAttributes) div SizeOf(Pointer)) do begin
      with AttributeList[AttributeCount] do begin
	if AttributeAddr^ <> E then begin
	  Attribute := PGPAVAttribute(AttributeIndex);
	  Value.PointerValue := AttributeAddr^;
	  Size := Length(Value.PointerValue);
	  inc(AttributeCount);
	end;
	inc(AttributeAddr);
      end;
    end;
  except
    on EAccessViolation do
      Result := kPGPError_BadMemAddress
    else Result := kPGPError_UnknownError;
  end;
end;

// search for certificates and their signing keys
function FindCertificate(KeyIter: pPGPKeyIter; KeySetToSearch: pPGPKeySet; var Certificate, SignKey: pPGPKeyDBObj): PGPError;
var
  X509Prop	: PGPBoolean;
begin
  SignKey := NIL;
  repeat
    Result := PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Any, Certificate);
    if (Result = 0) and (PGPGetKeyDBObjBooleanProperty(Certificate, kPGPSigProperty_IsX509, X509Prop) = 0)
    and (X509Prop = PGPTrue) then begin
      PGPGetSigCertifierKey(Certificate, KeySetToSearch, SignKey);
      Break;
    end;
  until Result <> 0;
end;

// check whether a user ID already has a certificate by the signing key(s)
function HasUserIDCertifierCert(CertKey: pPGPKeyDBObj; const CertKeyUserID: RawByteString;
				KeySetToSearch: pPGPKeySet; SignKey: pPGPKeyDBObj): Longbool;
var
  KeySet	: pPGPKeySet;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  UTF8ID	: RawByteString;
  UIDFound	: pPGPKeyDBObj;
  UserIDBuf	: TUserID;
  IDSize	: PGPSize;
  SigFound	: pPGPKeyDBObj;
  X509Prop	: PGPBoolean;
  KeyFound	: pPGPKeyDBObj;
begin
  Result := FALSE;
  if GetSingleKeyDBObjIter(CertKey, KeySet, KeyList, KeyIter) = 0 then begin
    try
      UTF8ID := AnsiToUtf8(CertKeyUserID);
      while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_UserID, UIDFound) = 0 do begin
	if (PGPGetKeyDBObjDataProperty(UIDFound, kPGPUserIDProperty_Name, @UserIDBuf, SizeOf(TUserID), IDSize) = 0)
	and ((PGPCompareUserIDStrings(UserIDBuf, PAnsiChar(CertKeyUserID)) = 0) or
	     (PGPCompareUserIDStrings(UserIDBuf, PAnsiChar(UTF8ID)) = 0)) then begin
	  while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Signature, SigFound) = 0 do begin
	    if (PGPGetKeyDBObjBooleanProperty(SigFound, kPGPSigProperty_IsX509, X509Prop) = 0)
	    and (X509Prop = PGPTrue) and (PGPGetSigCertifierKey(SigFound, KeySetToSearch, KeyFound) = 0)
	    and ((KeyFound = SignKey) or ((SignKey = NIL) and GetKeyPropHasCertificate(KeyFound))) then begin
	      Result := TRUE;
	      Exit;
	    end;
	  end;
	end;
      end;
    finally
      PGPFreeKeyIter(KeyIter);
      PGPFreeKeyIter(KeyList);
      PGPFreeKeySet(KeySet);
    end;
  end;
end;

// count valid (SignKey signed) certificates of a key
function ValidKeyCertCount(CertKey, SignKey: pPGPKeyDBObj): PGPUInt32;
var
  KeySet	: pPGPKeySet;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  RootKeySet	: pPGPKeySet;
  Certificate	: pPGPKeyDBObj;
  KeyFound	: pPGPKeyDBObj;
  X509Prop	: PGPBoolean;
begin
  Result := 0;
  if GetSingleKeyDBObjIter(CertKey, KeySet, KeyList, KeyIter) <> 0 then Exit;
  try
    RootKeySet := PGPPeekKeyDBRootKeySet(PGPPeekKeyDBObjKeyDB(CertKey));
    while (PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Any, Certificate) = 0) do begin
      if (PGPGetKeyDBObjBooleanProperty(Certificate, kPGPSigProperty_IsX509, X509Prop) = 0)
      and (X509Prop = PGPTrue) and GetCertPropIsValid(Certificate)
      and ((SignKey = NIL) or ((PGPGetSigCertifierKey(Certificate, RootKeySet, KeyFound) = 0) and (KeyFound = SignKey)))
      then begin
	inc(Result);
      end;
    end;
  finally
    PGPFreeKeyIter(KeyIter);
    PGPFreeKeyIter(KeyList);
    PGPFreeKeySet(KeySet);
  end;
end;

// returns the last valid certificate issued by SignKey (if specified)
function GetValidX509CertFromKey(const CertUserData: RawByteString; CertKey, SignKey: pPGPKeyDBObj;
				 KeySetToSearch: pPGPKeySet): pPGPKeyDBObj;
var
  KeySet	: pPGPKeySet;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
  KeyFound	: pPGPKeyDBObj;
begin
  Result := NIL;
  if GetSingleKeyDBObjIter(CertKey, KeySet, KeyList, KeyIter) = 0 then begin
    try
      while FindCertificate(KeyIter, KeySetToSearch, Certificate, KeyFound) = 0 do begin
	if GetCertPropIsValid(Certificate) and ((SignKey = NIL) or (KeyFound = SignKey))
	and ((CertUserData = E) or
	     (FirstStrPos(CertUserData, GetCertPropOwnerName(Certificate)) <> 0) or
	     (FirstStrPos(CertUserData, GetCertPropOwnerUserID(Certificate)) <> 0))
	then begin
	  Result := Certificate;
	end;
      end;
    finally
      PGPFreeKeyIter(KeyIter);
      PGPFreeKeyIter(KeyList);
      PGPFreeKeySet(KeySet);
    end;
  end;
end;

// create a key DB with valid encrypting X509 certified keys from the specified keyset
function GetX509EncryptCertDB(Context: pPGPContext; KeySetMain: pPGPKeySet; var KeyDBEncryptCerts: pPGPKeyDB): PGPError;
var
  KeySetEncCerts: pPGPKeySet;
  KeySetBuffer	: pPGPKeySet;
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
  SignKey	: pPGPKeyDBObj;
begin
  KeyDBEncryptCerts := NIL;
  KeySetBuffer := NIL;
  KeyIter := NIL;
  Result := 0;
  try
    Result := PGPNewEmptyKeySet(PGPPeekKeySetKeyDB(KeySetMain), KeySetEncCerts);
    if Result <> 0 then Exit;
    Result := GetCertKeySetFromKeySet(KeySetMain, TRUE, KeySetBuffer);
    if Result <> 0 then Exit;
    Result := PGPNewKeyIterFromKeySet(KeySetBuffer, KeyIter);
    if Result <> 0 then Exit;
    while FindCertificate(KeyIter, KeySetBuffer, Certificate, SignKey) = 0 do begin
      if GetCertPropCanEncrypt(Certificate, GetCertKeyUsage(Certificate)) then begin
	PGPAddKey(Certificate, KeySetEncCerts);
      end;
    end;
    PGPFreeKeySet(KeySetBuffer);
    KeySetBuffer := NIL;
    Result := PGPNewKeyDB(Context, KeyDBEncryptCerts);
    if Result = 0 then Result := PGPCopyKeys(KeySetEncCerts, KeyDBEncryptCerts, KeySetBuffer);
  finally
    PGPFreeKeyIter(KeyIter);
    PGPFreeKeySet(KeySetBuffer);
    PGPFreeKeySet(KeySetEncCerts);
    if Result <> 0 then PGPFreeKeyDB(KeyDBEncryptCerts);
  end;
end;

// check whether all keys in the specified keyset have valid certificates
function CheckValidX509CertifierSet(const KeySetToCheck: pPGPKeySet): PGPError;
var
  KeyCount	: PGPUInt32;
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
  SignKey	: pPGPKeyDBObj;
begin
  Result := PGPCountKeys(KeySetToCheck, KeyCount);
  if (Result = 0) and (KeyCount <> 0) then begin
    Result := PGPNewKeyIterFromKeySet(KeySetToCheck, KeyIter);
    if Result <> 0 then Exit;
    try
      while (FindCertificate(KeyIter, KeySetToCheck, Certificate, SignKey) = 0)
      and GetCertPropIsValid(Certificate) do begin
	dec(KeyCount);
      end;
      if KeyCount > 0 then Result := kPGPError_X509NeededCertNotAvailable;
    finally
      PGPFreeKeyIter(KeyIter);
    end;
  end;
end;

// filter the specified keyset for keys with valid/root/signing certificates
function GetX509CertifierSet(const KeySetToFilter: pPGPKeySet; var KeySetValidCertifiers: pPGPKeySet;
			     RootOnly, SignOnly, ValidOnly: Longbool): PGPError;
var
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
  SignKey	: pPGPKeyDBObj;
  CertKey	: pPGPKeyDBObj;
begin
  KeySetValidCertifiers := NIL;
  Result := PGPNewEmptyKeySet(PGPPeekKeySetKeyDB(KeySetToFilter), KeySetValidCertifiers);
  if Result <> 0 then Exit;
  Result := PGPNewKeyIterFromKeySet(KeySetToFilter, KeyIter);
  if Result <> 0 then begin
    PGPFreeKeySet(KeySetValidCertifiers);
    KeySetValidCertifiers := NIL;
    Exit;
  end;
  try
    while FindCertificate(KeyIter, KeySetToFilter, Certificate, SignKey) = 0 do begin
      CertKey := PGPPeekKeyDBObjKey(Certificate);
      if ValidOnly and not GetCertPropIsValid(Certificate) then Exit;
      if SignOnly and not GetKeyPropCanSign(CertKey) then Exit;
      if RootOnly and (SignKey <> CertKey) then Exit;
      PGPAddKey(CertKey, KeySetValidCertifiers);
    end;
  finally
    PGPFreeKeyIter(KeyIter);
  end;
end;

// retrieve keys with certificates - options:
// (SignKey = NIL) and (IncludeSigner = FALSE)  => all certified keys on keyring
// (SignKey = NIL) and (IncludeSigner = TRUE)   => all certifier keys on keyring (signing keys only)
// (SignKey <> NIL) and (IncludeSigner = FALSE) => all SignKey-certified keys on keyring without certifier
// (SignKey <> NIL) and (IncludeSigner = TRUE)  => all SignKey-certified keys on keyring including certifier
function GetCertKeySetBySignKey(Context: pPGPContext; KeySetToFilter: pPGPKeySet; SignKey: pPGPKeyDBObj;
				IncludeSigner, ValidKeysOnly: Longbool; var KeySetFiltered: pPGPKeySet): PGPError;
var
  SignKeyID	: TPGPKeyID7;
  SignKeyFilter	: pPGPFilter;
  ExcludeFilter	: pPGPFilter;
  CertKeyFilter	: pPGPFilter;
  KeyCount	: PGPUInt32;
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
  KeyFound	: pPGPKeyDBObj;
  KeySetChecked	: pPGPKeySet;
begin
  KeySetFiltered := NIL;
  if SignKey = NIL then begin
    if IncludeSigner then begin
      // signing keys with certificates only
      Result := GetX509CertifierSet(KeySetToFilter, KeySetFiltered, FALSE, TRUE, FALSE);
    end
    else begin
      // all certified keys in keyset (as is)
      Result := PGPIncKeySetRefCount(KeySetToFilter);
      if Result = 0 then KeySetFiltered := KeySetToFilter;
    end;
  end
  else begin
    SignKeyFilter := NIL;
    ExcludeFilter := NIL;
    CertKeyFilter := NIL;
    try
      // filter by signing key
      Result := PGPGetKeyIDFromKey(SignKey, SignKeyID);
      if Result <> 0 then Exit;
      Result := PGPNewKeyDBObjDataFilter(Context, kPGPSigProperty_KeyID, @SignKeyID,
					 SizeOf(TPGPKeyID7), kPGPMatchCriterion_Equal, SignKeyFilter);
      if Result <> 0 then Exit;
      // remove signing key?
      if not IncludeSigner then begin
	ExcludeFilter := GetKeyExcludeFilter(Context, SignKey);
	Result := PGPIntersectFilters(SignKeyFilter, ExcludeFilter, CertKeyFilter);
	SignKeyFilter := NIL;
	ExcludeFilter := NIL;
	if Result <> 0 then Exit;
      end
      else begin
	CertKeyFilter := SignKeyFilter;
	SignKeyFilter := NIL;
      end;
      Result := PGPFilterKeySet(KeySetToFilter, CertKeyFilter, KeySetFiltered);
    finally
      PGPFreeFilter(SignKeyFilter);
      PGPFreeFilter(ExcludeFilter);
      PGPFreeFilter(CertKeyFilter);
    end;
  end;
  if Result <> 0 then Exit;
  Result := PGPCountKeys(KeySetFiltered, KeyCount);
  if Result <> 0 then Exit;
  if KeyCount <> 0 then begin
    // remove keys without (valid) SignKey certificates
    if SignKey <> NIL then begin
      KeySetChecked := NIL;
      KeyIter := NIL;
      try
	Result := PGPNewEmptyKeySet(PGPPeekKeySetKeyDB(KeySetFiltered), KeySetChecked);
	if Result = 0 then begin
	  Result := PGPNewKeyIterFromKeySet(KeySetFiltered, KeyIter);
	  if Result = 0 then begin
	    while FindCertificate(KeyIter, KeySetFiltered, Certificate, KeyFound) = 0 do begin
	      if (KeyFound = SignKey) and (not ValidKeysOnly or GetCertPropIsValid(Certificate)) then begin
		PGPAddKey(PGPPeekKeyDBObjKey(Certificate), KeySetChecked);
	      end;
	    end;
	    if Result = 0 then begin
	      Result := PGPCountKeys(KeySetChecked, KeyCount);
	      if (Result = 0) and (KeyCount = 0) then Result := kPGPError_X509NeededCertNotAvailable;
	    end;
	  end;
	end;
      finally
	PGPFreeKeyIter(KeyIter);
	PGPFreeKeySet(KeySetFiltered);
	if Result = 0 then
	  KeySetFiltered := KeySetChecked
	else begin
	  PGPFreeKeySet(KeySetChecked);
	  KeySetFiltered := NIL;
	end;
      end;
    end;
  end
  else begin
    Result := kPGPError_X509NeededCertNotAvailable;
    PGPFreeKeySet(KeySetFiltered);
    KeySetFiltered := NIL;
  end;
end;

// search the specified keyset for a certificate with the specified CertificateIASN
function GetCertByCertIASN(KeySetToSearch: pPGPKeySet; const CertificateIASN: RawByteString;
			   var SignKey: pPGPKeyDBObj): pPGPKeyDBObj;
var
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
begin
  Result := NIL;
  SignKey := NIL;
  KeyIter := NIL;
  if PGPNewKeyIterFromKeySet(KeySetToSearch, KeyIter) <> 0 then Exit;
  try
    while FindCertificate(KeyIter, KeySetToSearch, Certificate, SignKey) = 0 do begin
      if CertificateIASN = GetCertPropIASN(Certificate) then begin
	Result := Certificate;
	Break;
      end;
    end;
  finally
    PGPFreeKeyIter(KeyIter);
  end;
end;

// retrieve a set containing a single certificate
function GetCertSetByCertIASN(Context: pPGPContext; KeySetToSearch: pPGPKeySet; const CertificateIASN: RawByteString;
			      var SingleCertSet: pPGPKeySet; var SignKey: pPGPKeyDBObj): Integer;
var
  Certificate	: pPGPKeyDBObj;
begin
  Certificate := GetCertByCertIASN(KeySetToSearch, CertificateIASN, SignKey);
  if Certificate <> NIL then
    Result := PGPNewSingletonKeySet(Certificate, SingleCertSet)
  else Result := kPGPError_X509NeededCertNotAvailable;
end;

// a simple file read function
function ReadDataFromFile(const CertFile: RawByteString; var CertData: RawByteString): Integer;
var
  CertStream	: TMemoryStream;
begin
  Result := 0;
  CertData := E;
  try
    CertStream := TMemoryStream.Create;
    if CertStream <> NIL then with CertStream do begin
      try
	LoadFromFile(String(CertFile));
	if Size <> 0 then begin
	  SetString(CertData, NIL, Size);
	  ReadBuffer(Pointer(CertData)^, Size);
	end;
      finally
	CertStream.Free;
      end;
    end;
  except
    Result := kPGPError_ReadFailed;
  end;
end;

// a simple file write function
function WriteDataToFile(const CertData, CertFile: RawByteString): Integer;
var
  CertStream	: TMemoryStream;
begin
  Result := 0;
  try
    CertStream := TMemoryStream.Create;
    if CertStream <> NIL then with CertStream do begin
      try
	WriteBuffer(Pointer(CertData)^, Length(CertData));
	SaveToFile(String(CertFile));
      finally
	CertStream.Free;
      end;
    end
    else Result := kPGPError_OutOfMemory;
  except
    Result := kPGPError_WriteFailed;
  end;
end;

// copy (binary) data to RawByteString
function CopyDataToString(const InData: PAnsiChar; var OutData: RawByteString; Size: DWord): Integer;
begin
  Result := 0;
  try
    SetString(OutData, NIL, Size);
    Move(InData^, Pointer(OutData)^, Size);
  except
    on EOutOfMemory do begin
      Result := kPGPError_OutOfMemory;
      OutData := E;
    end;
  end;
end;

// dialog for retrieving a single (to get) certified key
function X509GetKeyDialog(Context: pPGPContext; CertSetMain: pPGPKeySet; CertsOnly: Longbool;
			  var CertOrKey: pPGPKeyDBObj; const CertKeyPrompt: RawByteString;
			  WinHandle: Integer): Integer;
var
  CertSetSel	: pPGPKeySet;
begin
  Result := SelectX509CertifiedKeysDialog(Context, CertSetMain, CertSetSel, CertOrKey,
					  CertKeyPrompt, CertsOnly, TRUE, WinHandle);
  if Result <> 0 then Exit;
  try
    if not ((SDKVersion >= PGP81X) and CertsOnly) then Result := GetKeyFromKeySet(CertSetSel, CertOrKey);
  finally
    PGPFreeKeySet(CertSetSel);
  end;
end;

// dialog for retrieving a signing key and it's passphrase
function X509GetCertAndPassphraseDialog(Context: pPGPContext; KeySetMain: pPGPKeySet;
					var SignKey, SignCert: pPGPKeyDBObj; var Passphrase: PAnsiChar;
					const SignCertUser, SignCertIASN, SignKeyAndPassPrompt: RawByteString;
					WinHandle: Integer): Integer;
var
  KeySetFiltered: pPGPKeySet;
  UserEmail	: RawByteString;
  UserIDFilter	: pPGPFilter;
  KeyCount	: PGPUInt32;
  KeyDBTemp	: pPGPKeyDB;
  KeySetTemp	: pPGPKeySet;
  CertSetTemp	: pPGPKeySet;
  Dummy		: PGPUInt32;
  CertPrompt	: RawByteString;
begin
  SignKey := NIL;
  SignCert := NIL;
  if SignKeyAndPassPrompt <> E then begin // prompt for signing key and passphrase
    Passphrase := NIL;
    KeySetFiltered := NIL;
    if SignCertUser <> E then begin
      UserEmail := GetUserFromLongName(SignCertUser, TRUE);
      if UserEmail <> E then UserEmail := ExtractEmail(UserEmail);
      if UserEmail = E then UserEmail := ExtractEmail(SignCertUser);
      if UserEmail <> E then begin
	if PGPNewKeyDBObjDataFilter(Context, kPGPUserIDProperty_EmailAddress, PAnsiChar(UserEmail),
				    Length(UserEmail), kPGPMatchCriterion_Equal, UserIDFilter) = 0 then begin
	  try
	    PGPFilterKeySet(KeySetMain, UserIDFilter, KeySetFiltered);
	  finally
	    PGPFreeFilter(UserIDFilter);
	  end;
	end;
      end;
    end;
    if (KeySetFiltered = NIL) or (PGPCountKeys(KeySetFiltered, KeyCount) <> 0) or (KeyCount = 0) then begin
      KeySetFiltered := KeySetMain;
      UserEmail := E;
    end;
    try
      Result := PGPNewKeyDB(Context, KeyDBTemp);
      if Result <> 0 then Exit;
      try
	Result := PGPCopyKeys(KeySetFiltered, KeyDBTemp, KeySetTemp);
	if Result <> 0 then Exit;
	try
	  Result := GetCertKeySetFromKeySet(KeySetTemp, FALSE, CertSetTemp);
	  if Result <> 0 then Exit;
	  try
	    Result := SigningPassphraseDialog(Context, CertSetTemp, SignKey, Passphrase, SignCertIASN,
					      TRUE, soNo, Dummy, Dummy, Dummy, SignKeyAndPassPrompt, WinHandle);
	  finally
	    PGPFreeKeySet(CertSetTemp);
	  end;
	  if Result <> 0 then begin
	    if Result = kPGPError_SecretKeyNotFound then Result := kPGPError_X509NeededCertNotAvailable;
	    Exit;
	  end;
	  Result := GetKeyFromNewSet(SignKey, KeySetMain, SignKey);
	  if Result <> 0 then Exit;
	  if (SDKVersion >= PGP81X) and ((SignCertUser = E) or (UserEmail = E))
	  and (ValidKeyCertCount(SignKey, NIL) > 1) then begin
	    Result := PGPNewSingletonKeySet(SignKey, CertSetTemp);
	    if Result <> 0 then Exit;
	    try
	      CertPrompt := Copy(SignKeyAndPassPrompt, 1, FirstCharPos(SP, SignKeyAndPassPrompt));
	      if CertPrompt <> E then CertPrompt[Length(CertPrompt)] := ':';
	      Result := X509GetKeyDialog(Context, CertSetTemp, TRUE, SignCert, CertPrompt, WinHandle);
	    finally
	      PGPFreeKeySet(CertSetTemp);
	    end;
	  end
	  else SignCert := GetValidX509CertFromKey(SignCertUser, SignKey, NIL, KeySetMain);
	finally
	  PGPFreeKeySet(KeySetTemp);
	end;
      finally
	PGPFreeKeyDB(KeyDBTemp);
      end;
    finally
      if KeySetFiltered <> KeySetMain then PGPFreeKeySet(KeySetFiltered);
    end;
  end
  else begin // do we have a hex ID or an IASN?
    Result := GetKeyByHexID(KeySetMain, SignCertIASN, SignKey);
    if Result <> 0 then begin
      SignCert := GetCertByCertIASN(KeySetMain, SignCertIASN, SignKey);
      if SignCert <> NIL then Result := 0;
    end
    else begin
      if GetKeyPropCanSign(SignKey) then
	SignCert := GetValidX509CertFromKey(E, SignKey, NIL, KeySetMain)
      else Result := kPGPError_KeyUnusableForSignature;
    end;
    if Result <> 0 then Exit;
    if PGPPassphraseIsValid(SignKey, PGPOPassphrase(Context, Passphrase), PGPOLastOption(Context)) = PGPFalse then begin
      Result := kPGPError_BadPassphrase;
    end;
  end;
  if (SignCert = NIL) and (Result = 0) then Result := kPGPError_X509NeededCertNotAvailable;
end;

// if CertAttributes' CommonName or Email are undefined they are taken from the CertKey's primary user ID
function CreateX509CertificateRequest(const CertKeyAndPassPrompt, CertKeyHexID: RawByteString; ReqKeyPassphrase: PAnsiChar;
				      var CertAttributes: TX509CertAttributes; Armor: Longbool;
				      var Request: RawByteString; var ReqKeyProps: TKeyPropsRec;
				      KeyPropFlags, WinHandle: Integer): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeyToCertify	: pPGPKeyDBObj;
  Dummy		: pPGPKeyDBObj;
  AttributeCount: Longint;
  AttributeList	: TX509CertAttrArray;
  OptionList	: pPGPOptionList;
  CertOutput	: PAnsiChar;
  CertLength	: PGPSize;
begin
  if SDKVersion >= PGP7XX then begin
    Result := KeyRings.InitKeyRings(Context, KeySetMain);
    if Result <> 0 then Exit;
    try
      Result := X509GetCertAndPassphraseDialog(Context, KeySetMain, KeyToCertify, Dummy, ReqKeyPassphrase,
					       E, CertKeyHexID, CertKeyAndPassPrompt, WinHandle);
      if (Result <> 0) and (Result <> kPGPError_X509NeededCertNotAvailable) then Exit;
      Result := GetKeyProps(KeySetMain, KeyToCertify, KeyPropFlags, KeyFilterFlag_AllKeys, ReqKeyProps);
      with CertAttributes do begin
	if (CommonName = E) or (Email = E) then begin
	  Result := GetPrimaryUserIDAsAttributes(KeyToCertify, CommonName, Email);
	  if Result <> 0 then Exit;
	end;
      end;
      Result := GetAttributeListFromRecord(CertAttributes, AttributeList, AttributeCount);
      if Result <> 0 then Exit;
      OptionList := NIL;
      CertOutput := NIL;
      Result := PGPBuildOptionList(Context, OptionList,
	[
	  PGPOExportKey(Context, KeyToCertify),
	  PGPOPassphrase(Context, ReqKeyPassphrase),
	  PGPOArmorOutput(Context, PGPBoolean(Armor = TRUE)),
	  PGPOExportFormat(Context, kPGPExportFormat_X509CertReq),
	  PGPOAttributeValue(Context, @AttributeList, AttributeCount),
	  PGPOAllocatedOutputBuffer(Context, CertOutput, $FFFFFFFF, CertLength)
	]);
      if Result <> 0 then Exit;
      try
	Result := PGPExport(Context, OptionList, PGPOLastOption(Context));
	if (Result = 0) and (CertLength <> 0) then begin
	  Result := CopyDataToString(CertOutput, Request, CertLength);
	  if Result <> 0 then Exit;
	end
	else begin
	  Request := E;
	  Exit;
	end;
      finally
	PGPFreeData(CertOutput);
	PGPFreeOptionList(OptionList);
      end;
    finally
      if CertKeyAndPassPrompt <> E then PGPFreeData(ReqKeyPassphrase);
      KeyRings.FreeKeyRings;
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// specifying a unique CertAttributes' SerialNumber is mandatory, CommonName and Email are taken from the request
// if the SignCertIASN RawByteString contains a SignKeyHexID the key's last valid certificate will be used for certifying
function CreateX509CertificateFromRequest(const SignKeyAndPassPrompt, SignCertIASN: RawByteString; IssuerPassphrase: PAnsiChar;
					  var CertAttributes: TX509CertAttributes; ValidDays: Integer;
					  Armor: Longbool; const Request: RawByteString; var Certificate: RawByteString;
					  var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetX509	: pPGPKeySet;
  KeySetX509Sign: pPGPKeySet;
  SignKey	: pPGPKeyDBObj;
  SignCert	: pPGPKeyDBObj;
  SNUnique	: Longbool;
  AttributeCount: Longint;
  AttributeList	: TX509CertAttrArray;
  OptionList	: pPGPOptionList;
  KeyDBTemp	: pPGPKeyDB;
  KeySetTemp	: pPGPKeySet;
  RequestCert	: pPGPKeyDBObj;
  RequestKey	: pPGPKeyDBObj;
  UserIDFound	: pPGPKeyDBObj;
  KeyFound	: pPGPKeyDBObj;
  UserID	: TUserID;
  IDSize	: PGPSize;
  RequestCertSet: pPGPKeySet;
  CertOutput	: PAnsiChar;
  CertLength	: PGPSize;
begin
  if SDKVersion >= PGP7XX then begin
    Result := InitX509CertKeySet(Context, KeySetMain, KeySetX509);
    if Result <> 0 then Exit;
    try
      Result := GetX509CertifierSet(KeySetX509, KeySetX509Sign, FALSE, TRUE, TRUE);
      if Result <> 0 then Exit;
      try
	Result := X509GetCertAndPassphraseDialog(Context, KeySetX509Sign, SignKey, SignCert, IssuerPassphrase,
						 E, SignCertIASN, SignKeyAndPassPrompt, WinHandle);
      finally
	PGPFreeKeySet(KeySetX509Sign);
      end;
      if Result <> 0 then Exit;
      Result := IsSerialNumberUnique(Context, KeySetMain, SignKey, CertAttributes.SerialNumber, SNUnique);
      if Result <> 0 then Exit;
      if not SNUnique then begin
	Result := kPGPError_DuplicateCert;
	Exit;
      end;
      Result := GetAttributeListFromRecord(CertAttributes, AttributeList, AttributeCount);
      if Result <> 0 then Exit;
      OptionList := NIL;
      Result := PGPBuildOptionList(Context, OptionList,
	[
	  PGPOInputBuffer(Context, PAnsiChar(Request), Length(Request)),
	  PGPOAttributeValue(Context, @AttributeList, AttributeCount),
	  PGPOPassphrase(Context, IssuerPassphrase),
	  PGPOExpiration(Context, ValidDays)
	]);
      if Result <> 0 then Exit;
      try
	Result := PGPNewKeyDB(Context, KeyDBTemp);
	if Result <> 0 then Exit;
	try
	  Result := PGPCopyKeys(KeySetX509, KeyDBTemp, KeySetTemp);
	  if Result <> 0 then Exit;
	  try
	    Result := PGPCreateX509CertificateFromRequest(SignCert, RequestCert, OptionList, PGPOLastOption(Context));
	    if Result <> 0 then Exit;
	    Result := KeyRings.UpdateKeyRings;
	    if Result <> 0 then Exit;
	    Result := GetX509CertPropsList(RequestCert, CertPropsList);
	    if Result <> 0 then Exit;
	    RequestKey := PGPPeekKeyDBObjKey(RequestCert);
	    if RequestKey = SignKey then begin
	      Result := kPGPError_X509SelfSignedCert;
	      if GetKeyFromNewSet(SignKey, KeySetTemp, SignKey) = 0 then begin
		RemoveKeyFromKeySet(RequestKey, KeySetX509);
		AddKeyToKeySet(SignKey, KeySetX509);
	      end;
	    end
	    else begin
	      UserIDFound := PGPPeekKeyDBObjUserID(RequestCert);
	      if (GetKeyFromNewSet(SignKey, KeySetTemp, SignKey) = 0)
	      and (GetKeyFromNewSet(RequestKey, KeySetTemp, KeyFound) = 0)
	      and (PGPGetKeyDBObjDataProperty(UserIDFound, kPGPUserIDProperty_Name, @UserID, SizeOf(TUserID), IDSize) = 0)
	      and HasUserIDCertifierCert(KeyFound, UserID, KeySetTemp, SignKey) then begin
		Result := kPGPError_DuplicateUserID;
		RemoveKeyFromKeySet(RequestKey, KeySetX509);
		AddKeyToKeySet(KeyFound, KeySetX509);
	      end;
	    end;
	  finally
	    PGPFreeKeySet(KeySetTemp);
	  end;
	finally
	  PGPFreeKeyDB(KeyDBTemp);
	end;
      finally
	PGPFreeOptionList(OptionList);
      end;
      if Result <> 0 then Exit;
      Result := PGPNewSingletonKeySet(RequestCert, RequestCertSet);
      if Result <> 0 then Exit;
      try
	OptionList := NIL;
	Result := PGPBuildOptionList(Context, OptionList,
	  [
	    PGPOExportKeySet(Context, RequestCertSet),
	    PGPOExportFormat(Context, kPGPExportFormat_X509Cert),
	    PGPOArmorOutput(Context, PGPBoolean(Armor = TRUE)),
	    PGPOAllocatedOutputBuffer(Context, CertOutput, $FFFFFFFF, CertLength)
	  ]);
	if Result <> 0 then Exit;
	try
	  Result := PGPExport(Context, OptionList, PGPOLastOption(Context));
	  if Result <> 0 then Exit;
	  try
	    Result := CopyDataToString(CertOutput, Certificate, CertLength);
	  finally
	    PGPFreeData(CertOutput);
	  end;
	finally
	  PGPFreeOptionList(OptionList);
	end;
      finally
	PGPFreeKeySet(RequestCertSet);
      end;
    finally
      if SignKeyAndPassPrompt <> E then PGPFreeData(IssuerPassphrase);
      FinitX509CertKeySet(KeySetX509);
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// if the RevokeCertIASN RawByteString contains valid CertKeyHexID(s) all certificates on the respective key(s) will be revoked
// if no RevokeCertIASN RawByteString is specified all (non-self) certificates issued by the respective SignKey will be revoked
// if the SignCertIASN RawByteString contains a SignKeyHexID it's last valid certificate will be used for revoking certificates
// if SignCertIASN isn't provided and RevokeCertIASN contains a unique IASN the certifying key can be found automatically
function CreateX509CertificateRevocationList(const CertKeyPrompt, SignKeyAndPassPrompt, RevokeCertIASN, SignCertIASN: RawByteString;
					     RevokerPassphrase: PAnsiChar; ValidDays: Integer;
					     var CertPropsList: TX509CertPropsList;
					     WinHandle: Integer): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetX509	: pPGPKeySet;
  KeySetX509Sign: pPGPKeySet;
  SignKey	: pPGPKeyDBObj;
  SignCert	: pPGPKeyDBObj;
  SignKeyCertSet: pPGPKeySet;
  RevokeCertSet	: pPGPKeySet;
  RevokeCert	: pPGPKeyDBObj;
  SignKeyFound	: pPGPKeyDBObj;
  OptionList	: pPGPOptionList;
begin
  if SDKVersion >= PGP7XX then begin
    Result := InitX509CertKeySet(Context, KeySetMain, KeySetX509);
    if Result <> 0 then Exit;
    try
      Result := GetX509CertifierSet(KeySetX509, KeySetX509Sign, TRUE, TRUE, TRUE);
      if Result <> 0 then Exit;
      if (SignKeyAndPassPrompt <> E) or (SignCertIASN <> E) then begin
	try
	  Result := X509GetCertAndPassphraseDialog(Context, KeySetX509Sign, SignKey, SignCert, RevokerPassphrase,
						   E, SignCertIASN, SignKeyAndPassPrompt, WinHandle);
	finally
	  PGPFreeKeySet(KeySetX509Sign);
	end;
	if Result <> 0 then Exit;
	if not GetCertPropIsRoot(SignCert) then begin
	  Result := kPGPError_MissingX509Certificate;
	  Exit;
	end;
      end
      else SignCert := NIL;
      RevokeCertSet := NIL;
      try
	if CertKeyPrompt <> E then begin // prompt for certified key(s)
	  Result := GetCertKeySetBySignKey(Context, KeySetX509, SignKey, TRUE, TRUE, SignKeyCertSet);
	  if Result <> 0 then Exit;
	  try
	    Result := SelectX509CertifiedKeysDialog(Context, SignKeyCertSet, RevokeCertSet, RevokeCert,
						    CertKeyPrompt, TRUE, TRUE, WinHandle);
	  finally
	    PGPFreeKeySet(SignKeyCertSet);
	  end;
	end
	else begin
	  if RevokeCertIASN <> E then begin // do we have hex ID(s) or an IASN?
	    Result := GetKeySetByAnyIDs(Context, KeySetX509, RevokeCertIASN, RevokeCertSet);
	    if Result <> 0 then begin
	      Result := GetCertSetByCertIASN(Context, KeySetX509, RevokeCertIASN, RevokeCertSet, SignKeyFound);
	      if (Result = 0) and (SignCert = NIL) then begin
		SignCert := GetValidX509CertFromKey(E, SignKeyFound, SignKeyFound, KeySetX509);
		if SignCert <> NIL then SignKey := SignKeyFound;
	      end;
	    end;
	  end
	  else begin // get key(s) certified by signing key
	    Result := GetCertKeySetBySignKey(Context, KeySetX509, SignKey, FALSE, TRUE, RevokeCertSet);
	  end;
	end;
	if Result <> 0 then Exit;
	OptionList := NIL;
	Result := PGPBuildOptionList(Context, OptionList,
	  [
	    PGPOPassphrase(Context, RevokerPassphrase),
	    PGPOExpiration(Context, ValidDays)
	  ]);
	if Result <> 0 then Exit;
	try
	  Result := PGPCreateX509CRL(SignCert, RevokeCertSet, OptionList, PGPOLastOption(Context));
	  if Result <> 0 then Exit;
	  Result := KeyRings.UpdateKeyRings;
	  if Result <> 0 then Exit;
	  Result := GetX509CertSetProps(RevokeCertSet, KeySetX509, SignKey, E, 0,
					kPGPKeyOrdering_UserID, CertPropsList);
	  if Result > 0 then Result := 0;
	finally
	  PGPFreeOptionList(OptionList);
	end;
      finally
	PGPFreeKeySet(RevokeCertSet);
      end;
    finally
      if SignKeyAndPassPrompt <> E then PGPFreeData(RevokerPassphrase);
      FinitX509CertKeySet(KeySetX509);
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// specifying a unique CertAttributes.SerialNumber is mandatory
// splitting a user ID into CommonName and Email will certify it
// self-signed certificates cannot be created for an existing user ID
function CreateX509Certificate(const CertKeyPrompt, SignKeyAndPassPrompt, CertKeyHexID, SignCertIASN: RawByteString;
			       IssuerPassphrase: PAnsiChar; var CertAttributes: TX509CertAttributes; ValidDays: Integer;
			       SelfSigned: Longbool; var CertPropsList: TX509CertPropsList; WinHandle: Longint): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetX509	: pPGPKeySet;
  KeySetX509Sign: pPGPKeySet;
  SignKey	: pPGPKeyDBObj;
  SignCert	: pPGPKeyDBObj;
  SNUnique	: Longbool;
  KeySetFiltered: pPGPKeySet;
  KeyToCertify	: pPGPKeyDBObj;
  AttributeCount: Longint;
  AttributeList	: TX509CertAttrArray;
  OptionList	: pPGPOptionList;
  Certificate	: pPGPKeyDBObj;
begin
  if SDKVersion >= PGP7XX then begin
    Result := InitX509CertKeySet(Context, KeySetMain, KeySetX509);
    if Result <> 0 then Exit;
    try
      if SelfSigned then
	KeySetX509Sign := KeySetMain
      else begin
	Result := GetX509CertifierSet(KeySetX509, KeySetX509Sign, FALSE, TRUE, TRUE);
	if Result <> 0 then Exit;
      end;
      try
	Result := X509GetCertAndPassphraseDialog(Context, KeySetX509Sign, SignKey, SignCert, IssuerPassphrase,
						 E, SignCertIASN, SignKeyAndPassPrompt, WinHandle);
      finally
	if SelfSigned then begin
	  if Result = kPGPError_X509NeededCertNotAvailable then Result := 0;
	end
	else PGPFreeKeySet(KeySetX509Sign);
      end;
      if Result <> 0 then Exit;
      Result := IsSerialNumberUnique(Context, KeySetMain, SignKey, CertAttributes.SerialNumber, SNUnique);
      if Result <> 0 then Exit;
      if not SNUnique then begin
	Result := kPGPError_DuplicateCert;
	Exit;
      end;
      if not SelfSigned then begin
	GetKeySetWithoutSigner(Context, KeySetMain, SignKey, KeySetFiltered);
	try
	  if CertKeyPrompt <> E then
	    Result := X509GetKeyDialog(Context, KeySetFiltered, FALSE, KeyToCertify, CertKeyPrompt, WinHandle)
	  else Result := GetKeyByHexID(KeySetFiltered, CertKeyHexID, KeyToCertify);
	finally
	  if KeySetFiltered <> KeySetMain then PGPFreeKeySet(KeySetFiltered);
	end;
      end
      else KeyToCertify := SignKey;
      if Result <> 0 then Exit;
      with CertAttributes do begin
	if KeyToCertify <> SignKey then begin
	  if (CommonName = E) or (Email = E) then begin
	    Result := GetPrimaryUserIDAsAttributes(KeyToCertify, CommonName, Email);
	    if Result <> 0 then Exit;
	  end;
	  // as we already know which certifier key will be used we only have to check with the selected SignKey
	  if HasUserIDCertifierCert(KeyToCertify, CommonName + ' <' + Email + '>', KeySetMain, SignKey) then begin
	    Result := kPGPError_DuplicateUserID;
	    Exit;
	  end;
	end
	else begin
	  if (CommonName = E) or (Email = E) then begin
	    Result := kPGPError_InvalidDistinguishedName;
	    Exit;
	  end
	  else if HasKeyUserID(KeyToCertify, CommonName + ' <' + Email + '>') then begin
	    Result := kPGPError_DuplicateUserID;
	    Exit;
	  end;
	end;
      end;
      Result := GetAttributeListFromRecord(CertAttributes, AttributeList, AttributeCount);
      if Result <> 0 then Exit;
      OptionList := NIL;
      Result := PGPBuildOptionList(Context, OptionList,
	[
	  PGPOAttributeValue(Context, @AttributeList, AttributeCount),
	  PGPOPassphrase(Context, IssuerPassphrase),
	  PGPOExpiration(Context, ValidDays)
	]);
      if Result <> 0 then Exit;
      try
	if KeyToCertify = SignKey then
	  Result := PGPCreateSelfSignedX509Certificate(SignKey, Certificate, OptionList, PGPOLastOption(Context))
	else Result := PGPCreateX509Certificate(SignCert, KeyToCertify, Certificate, OptionList, PGPOLastOption(Context));
	if Result <> 0 then Exit;
	Result := KeyRings.UpdateKeyRings;
	if Result <> 0 then Exit;
	Result := GetX509CertPropsList(Certificate, CertPropsList);
      finally
	PGPFreeOptionList(OptionList);
      end;
    finally
      if SignKeyAndPassPrompt <> E then PGPFreeData(IssuerPassphrase);
      FinitX509CertKeySet(KeySetX509);
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// self-signed certificates can only be removed including the user ID they are attached to
// user IDs of certificates can only be removed if they don't have any other signatures assigned
// if there's only a single user ID left on a key the key will be removed from the keyring all together
function RemoveX509Certificate(const CertKeyPrompt, RemoveCertIASN: RawByteString; IncludeUserID: Longbool;
			       var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetX509	: pPGPKeySet;
  RemoveCertKey : pPGPKeyDBObj;
  RemoveCert	: pPGPKeyDBObj;
  SignKey	: pPGPKeyDBObj;
  RemoveUserID	: pPGPKeyDBObj;
  SigRemoved	: Longbool;
  KeySet	: pPGPKeySet;
  KeyList	: pPGPKeyList;
  KeyIter	: pPGPKeyIter;
  UserIDFound	: pPGPKeyDBObj;
  SignatureFound: pPGPKeyDBObj;
begin
  if SDKVersion >= PGP7XX then begin
    Result := InitX509CertKeySet(Context, KeySetMain, KeySetX509);
    if Result <> 0 then Exit;
    try
      if CertKeyPrompt <> E then begin // prompt for a certified key
	Result := X509GetKeyDialog(Context, KeySetX509, TRUE, RemoveCertKey, CertKeyPrompt, WinHandle);
	if Result <> 0 then Exit;
	if SDKVersion >= PGP81X then begin
	  RemoveCert := RemoveCertKey;
	  RemoveCertKey := PGPPeekKeyDBObjKey(RemoveCert);
	end
	else RemoveCert := GetValidX509CertFromKey(E, RemoveCertKey, NIL, KeySetX509);
      end
      else begin // do we have a hex ID or an IASN?
	Result := GetKeyByHexID(KeySetX509, RemoveCertIASN, RemoveCertKey);
	if Result = 0 then
	  RemoveCert := GetValidX509CertFromKey(E, RemoveCertKey, NIL, KeySetX509)
	else begin
	  RemoveCert := GetCertByCertIASN(KeySetX509, RemoveCertIASN, SignKey);
	  if RemoveCert <> NIL then RemoveCertKey := PGPPeekKeyDBObjKey(RemoveCert);
	end;
      end;
      RemoveUserID := NIL;
      if RemoveCert <> NIL then begin
	RemoveUserID := PGPPeekKeyDBObjUserID(RemoveCert);
	Result := GetX509CertPropsList(RemoveCert, CertPropsList);
      end
      else if Result = 0 then Result := kPGPError_X509NeededCertNotAvailable;
      if Result <> 0 then Exit;
      with CertPropsList do begin
	if not X509CertProps[pred(Count)].x509IsRoot then begin
	  Result := PGPRemoveSig(RemoveCert);
	  if Result <> 0 then Exit;
	  SigRemoved := TRUE;
	end
	else SigRemoved := FALSE;
      end;
      if IncludeUserID then begin
	Result := GetSingleKeyDBObjIter(RemoveUserID, KeySet, KeyList, KeyIter);
	if Result = 0 then begin
	  try
	    while PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_UserID, UserIDFound) = 0 do begin
	      if UserIDFound = RemoveUserID then begin
		if (PGPKeyIterRewindKeyDBObj(KeyIter, kPGPKeyDBObjType_Signature) <> 0)
		or (PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Signature, SignatureFound) <> 0)
		or (
		  (SignatureFound = RemoveCert) and
		  (PGPKeyIterNextKeyDBObj(KeyIter, kPGPKeyDBObjType_Signature, SignatureFound) <> 0)
		) then begin
		  Result := PGPRemoveUserID(RemoveUserID);
		  if Result <> 0 then begin
		    PGPFreeKeySet(KeySet);
		    KeySet := NIL;
		    if (PGPNewSingletonKeySet(RemoveCertKey, KeySet) = 0) and (CountUserIDs(KeySet) = 1) then begin
		      Result := PGPRemoveKeys(KeySet, KeySetMain);
		    end;
		  end;
		  if Result = 0 then SigRemoved := TRUE
		end;
		Break;
	      end;
	    end;
	  finally
	    PGPFreeKeyIter(KeyIter);
	    PGPFreeKeyList(KeyList);
	    PGPFreeKeySet(KeySet);
	  end;
	end;
      end;
      if SigRemoved then
	Result := KeyRings.UpdateKeyRings
      else if not IncludeUserID then Result := kPGPError_X509SelfSignedCert;
    finally
      FinitX509CertKeySet(KeySetX509);
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// import binary or armored standard X509 certificates or PGP keys to keyring(s)
function ImportX509Certificate(const CertKeyPrompt, CertPassPrompt, CertData: RawByteString;
			       PKCS12Passphrase, CertPassphrase: PAnsiChar; FromFile: Longbool;
			       var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeyPropsList	: TKeyPropsList;
  ImportKeySet	: pPGPKeySet;
begin
  if SDKVersion >= PGP7XX then begin
    if CertData <> E then begin
      KeyPropsList := NIL;
      try
	Result := KeyImportDialog(CertKeyPrompt, CertPassPrompt, CertData, PKCS12Passphrase, KeyPropsList, FromFile,
				  IgnoreFlag_None, spgpKeyPropFlag_KeyID, WinHandle);
	if Result <> 0 then Exit;
	if CertPassphrase <> PKCS12Passphrase then begin
	  ChangePassphrase(PAnsiChar(KeyPropsList[0]), PKCS12Passphrase, CertPassphrase);
	end;
	Result := KeyRings.InitKeyRings(Context, KeySetMain);
	if Result <> 0 then Exit;
	try
	  Result := GetKeySetByAnyIDs(Context, KeySetMain, KeyPropsList.CommaText, ImportKeySet);
	  if Result <> 0 then Exit;
	  try
	    Result := GetX509CertSetProps(ImportKeySet, KeySetMain, NIL, E, 0,
					  kPGPKeyOrdering_UserID, CertPropsList);
	    if Result > 0 then Result := 0;
	  finally
	    PGPFreeKeySet(ImportKeySet);
	  end;
	finally
	  KeyRings.FreeKeyRings;
	end;
      finally
	KeyPropsList.Free;
      end;
    end
    else Result := kPGPError_InvalidInputFormat;
    Result := TranslateX509ErrorToPGPError(Result);
  end
  else Result := kPGPError_LazyProgrammer;
end;

// write certificate(s) and size information to chain
function WriteDataToChain(const CertData: PAnsiChar; var Chain: RawByteString;
			  var ChainPos: PGPSize; CertLen: PGPSize): Integer;
var
  pChain	: PAnsiChar;
begin
  Result := 0;
  pChain := Pointer(Chain);
  if ChainPos = 0 then begin
    pChain[0] := AnsiChar(CertLen shr 16);
    pChain[1] := AnsiChar(CertLen shr 8);
    pChain[2] := AnsiChar(CertLen);
    ChainPos := 3;
  end
  else begin
    SetLength(Chain, ChainPos + CertLen + 3);
    pChain := Pointer(Chain);
    pChain[ChainPos + 0] := AnsiChar(CertLen shr 16);
    pChain[ChainPos + 1] := AnsiChar(CertLen shr 8);
    pChain[ChainPos + 2] := AnsiChar(CertLen);
    Move(CertData^, pChain[ChainPos + 3], CertLen);
    inc(ChainPos, CertLen + 3);
    if ChainPos > $FFFFFF then begin
      Result := kPGPError_LDAPSizelimitExceeded;
      Chain := E;
    end;
  end;
end;

// list certificates from oldest to youngest
function GetDateSortedCertListFromKeySet(CertKeySet: pPGPKeySet; var CertList: TRawByteStrings): Integer;
var
  KeyCount	: PGPUInt32;
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
  KeyFound	: pPGPKeyDBObj;
  CreaTime	: PGPTime;
begin
  CertList := NIL;
  Result := PGPCountKeys(CertKeySet, KeyCount);
  if Result <> 0 then Exit;
  if KeyCount = 0 then begin
    Result := kPGPError_X509NeededCertNotAvailable;
    Exit;
  end;
  Result := PGPNewKeyIterFromKeySet(CertKeySet, KeyIter);
  if Result <> 0 then Exit;
  try
    CertList := TRawByteStrings.Create;
    if CertList <> NIL then begin
      while FindCertificate(KeyIter, CertKeySet, Certificate, KeyFound) = 0 do begin
	Result := PGPGetKeyDBObjTimeProperty(Certificate, kPGPSigProperty_Creation, CreaTime);
	if Result <> 0 then Exit;
	CertList.AddObject(IntToHex(CreaTime, 8), TObject(Certificate));
      end;
      CertList.Sort;
    end
    else Result := kPGPError_OutOfMemory;
  finally
    PGPFreeKeyIter(KeyIter);
  end;
  if CertList.Count = 0 then Result := kPGPError_X509NeededCertNotAvailable;
end;

// create a certificate chain from any keyset
function CreateCertChainFromKeySet(CertKeySet: pPGPKeySet; var CertChain: RawByteString;
				   var MostRecentCert: pPGPKeyDBObj): Integer;
var
  CertList	: TRawByteStrings;
  ChainPos	: PGPSize;
  CertData	: TX509CertData;
  CertIndex	: Integer;
  CertLength	: PGPSize;
begin
  MostRecentCert := NIL;
  try
    Result := GetDateSortedCertListFromKeySet(CertKeySet, CertList);
    if Result <> 0 then Exit;
    ChainPos := 0;
    CertData[0] := #0;
    SetString(CertChain, NIL, 3);
    Result := WriteDataToChain(CertData, CertChain, ChainPos, 0);
    if Result <> 0 then Exit;
    CertIndex := pred(CertList.Count);
    MostRecentCert := CertList.Objects[CertIndex];
    while CertIndex >= 0 do begin
      Result := PGPGetKeyDBObjDataProperty(pPGPKeyDBObj(CertList.Objects[CertIndex]), kPGPSigProperty_X509Certificate,
					   @CertData, MAX_509_CERT_SIZE, CertLength);
      if Result <> 0 then Exit;
      try
	Result := WriteDataToChain(CertData, CertChain, ChainPos, CertLength);
	if Result <> 0 then Exit;
      except
	on EOutOfMemory do begin
	  Result := kPGPError_OutOfMemory;
	  CertChain := E;
	  Exit;
	end;
      end;
      dec(CertIndex);
    end;
    if ChainPos > 3 then begin
      CertLength := 0;
      CertData[0] := #0;
      WriteDataToChain(CertData, CertChain, CertLength, ChainPos - 3);
    end
    else begin
      Result := kPGPError_X509NeededCertNotAvailable;
      Exit;
    end;
  finally
    CertList.Free;
  end;
end;

// either export a single certificate or a certificate chain with parent certificate(s) included
// chains suitable for VerifyX509CertificateChain can be created with Armor disabled
function ExportX509Certificate(const CertKeyOrKeyAndPassPrompt, ExportCertIASN: RawByteString; SecKeyPassphrase: PAnsiChar;
			       Armor, Chain, Secret, ToFile: Longbool; var CertData: RawByteString;
			       var CertPropsList: TX509CertPropsList; WinHandle: Integer): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetX509	: pPGPKeySet;
  ExportCert	: pPGPKeyDBObj;
  ExportCertKey	: pPGPKeyDBObj;
  SignKey	: pPGPKeyDBObj;
  CertFile	: RawByteString;
  FileExtension	: String;
  ExportCertSet	: pPGPKeySet;
  OptionList	: pPGPOptionList;
  PassOption	: pPGPOptionList;
  CertOutput	: PAnsiChar;
  CertLength	: PGPSize;
  OptionListCopy: pPGPOptionList;
  CertSigner	: pPGPKeyDBObj;
  MostRecentCert: pPGPKeyDBObj;
begin
  if SDKVersion >= PGP7XX then begin
    if Secret then Chain := FALSE;
    Result := InitX509CertKeySet(Context, KeySetMain, KeySetX509);
    if Result <> 0 then Exit;
    try
      ExportCert := NIL;
      if CertKeyOrKeyAndPassPrompt <> E then begin // prompt for a certified key
	if Secret then begin
	  Result := X509GetCertAndPassphraseDialog(Context, KeySetX509, ExportCertKey, ExportCert, SecKeyPassphrase,
						   E, ExportCertIASN, CertKeyOrKeyAndPassPrompt, WinHandle);
	end
	else begin
	  Result := X509GetKeyDialog(Context, KeySetX509, TRUE, ExportCertKey, CertKeyOrKeyAndPassPrompt, WinHandle);
	  if Result <> 0 then Exit;
	  if SDKVersion >= PGP81X then begin
	    ExportCert := ExportCertKey;
	    ExportCertKey := PGPPeekKeyDBObjKey(ExportCert);
	  end
	  else ExportCert := GetValidX509CertFromKey(E, ExportCertKey, NIL, KeySetX509);
	end;
      end
      else begin // do we have a hex ID or an IASN?
	Result := GetKeyByHexID(KeySetX509, ExportCertIASN, ExportCertKey);
	if Result = 0 then
	  ExportCert := GetValidX509CertFromKey(E, ExportCertKey, NIL, KeySetX509)
	else begin
	  ExportCert := GetCertByCertIASN(KeySetX509, ExportCertIASN, SignKey);
	  if ExportCert <> NIL then begin
	    ExportCertKey := PGPPeekKeyDBObjKey(ExportCert);
	    Result := 0;
	  end;
	end;
	if ExportCert = NIL then Result := kPGPError_X509NeededCertNotAvailable;
      end;
      if Result <> 0 then Exit;
      Secret := (Secret and GetKeyPropIsSecret(ExportCertKey));
      if ToFile then begin
	CertFile := CertData;
	FileExtension := LowerCase(ExtractFileExt(String(CertFile)));
	if Armor then begin
	  if FileExtension <> '.pem' then CertFile := CertFile + '.pem';
	end
	else if Secret then begin
	  if FileExtension <> '.p12' then CertFile := CertFile + '.p12';
	end
	else if Chain then begin
	  if FileExtension <> '.chain' then CertFile := CertFile + '.chain';
	end
	else if (FileExtension <> '.cer') then CertFile := CertFile + '.cer';
      end;
      ExportCertSet := NIL;
      OptionList := NIL;
      try
	if not Chain then begin
	  if Secret then begin
	    if SecKeyPassphrase <> NIL then
	      PassOption := PGPOPassphrase(Context, SecKeyPassphrase)
	    else begin
	      Result := kPGPError_BadPassphrase;
	      Exit;
	    end;
	  end
	  else PassOption := PGPONullOption(Context);
	  Result := PGPBuildOptionList(Context, OptionList,
	    [
	      PassOption,
	      PGPOExportUserID(Context, ExportCert),
	      PGPOArmorOutput(Context, PGPBoolean(Armor = TRUE)),
	      PGPOExportFormat(Context, kPGPExportFormat_X509Cert),
	      PGPOExportPrivateKeys(Context, PGPBoolean(Secret = TRUE)),
	      PGPOExportPrivateSubKeys(Context, PGPBoolean(Secret = TRUE)),
	      PGPOAllocatedOutputBuffer(Context, CertOutput, $FFFFFFFF, CertLength)
	    ]);
	  if Result <> 0 then Exit;
	  try
	    Result := PGPExport(Context, OptionList, PGPOLastOption(Context));
	    if Result <> 0 then Exit;
	    try
	      Result := CopyDataToString(CertOutput, CertData, CertLength);
	      if Result <> 0 then Exit;
	      Result := GetX509CertPropsList(ExportCert, CertPropsList);
	    finally
	      PGPFreeData(CertOutput);
	    end;
	  finally
	    PGPFreeOptionList(OptionList);
	  end;
	end
	else begin
	  Result := PGPNewEmptyKeySet(PGPPeekKeySetKeyDB(KeySetX509), ExportCertSet);
	  if Result <> 0 then Exit;
	  if Armor then begin
	    Result := PGPBuildOptionList(Context, OptionList,
	      [
		PGPOArmorOutput(Context, PGPTrue),
		PGPOExportFormat(Context, kPGPExportFormat_X509Cert),
		PGPOAllocatedOutputBuffer(Context, CertOutput, $FFFFFFFF, CertLength)
	      ]);
	    if Result <> 0 then Exit;
	  end;
	  try
	    CertData := E;
	    repeat
	      if ExportCert <> NIL then begin
		Result := PGPAddKey(ExportCert, ExportCertSet);
		if Result <> 0 then Exit;
		if Armor then begin
		  Result := PGPCopyOptionList(OptionList, OptionListCopy);
		  if Result <> 0 then Exit;
		  try
		    Result := PGPAppendOptionList(OptionListCopy, [PGPOExportUserID(Context, ExportCert)]);
		    if Result <> 0 then Exit;
		    Result := PGPExport(Context, OptionListCopy, PGPOLastOption(Context));
		    if Result <> 0 then Exit;
		    try
		      CertData := CertData + CertOutput + CRLF;
		    finally
		      PGPFreeData(CertOutput);
		    end;
		  finally
		    PGPFreeOptionList(OptionListCopy);
		  end;
		end;
		Result := PGPGetSigCertifierKey(ExportCert, KeySetX509, CertSigner);
		if (Result <> 0) and (Result <> kPGPError_ItemNotFound) then Exit;
		if (Result = 0) and (CertSigner <> ExportCertKey) then
		  ExportCertKey := CertSigner
		else Chain := FALSE;
	      end
	      else begin
		Result := kPGPError_X509NeededCertNotAvailable;
		Exit;
	      end;
	      ExportCert := GetValidX509CertFromKey(GetCertPropIssuerName(ExportCert), ExportCertKey, NIL, KeySetX509);
	    until not Chain;
	  finally
	    PGPFreeOptionList(OptionList);
	  end;
	  if not Armor then begin
	    Result := CreateCertChainFromKeySet(ExportCertSet, CertData, MostRecentCert);
	    if Result <> 0 then Exit;
	  end
	  else Delete(CertData, pred(Length(CertData)), Length(CRLF));
	  Result := GetX509CertSetProps(ExportCertSet, KeySetX509, NIL, E, 0,
					kPGPKeyOrdering_CreationDate, CertPropsList);
	  if Result > 0 then Result := 0;
	end;
      finally
	PGPFreeKeySet(ExportCertSet);
      end;
      if ToFile then begin
	Result := WriteDataToFile(CertData, CertFile);
	CertData := CertFile;
      end;
    finally
      if CertKeyOrKeyAndPassPrompt <> E then PGPFreeData(SecKeyPassphrase);
      FinitX509CertKeySet(KeySetX509);
      if Result <> 0 then CertData := E;
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// create a temporary keyDB for checking certificates
function GetX509CertImportDB(Context: pPGPContext; const CertData: RawByteString;
			     var CertImportDB: pPGPKeyDB; FromFile: Longbool): Integer;
var
  CertInput	: RawByteString;
  LoopCount	: Integer;
  OptionList	: pPGPOptionList;
begin
  Result := 0;
  OptionList := NIL;
  CertImportDB := NIL;
  if FromFile then
    Result := ReadDataFromFile(CertData, CertInput)
  else CertInput := CertData;
  if Result <> 0 then Exit;
  LoopCount := 0;
  repeat
    inc(LoopCount);
    if Result <> kPGPError_InvalidPKCS7Encoding then begin
      Result := PGPBuildOptionList(Context, OptionList,
	[
	  PGPOInputFormat(Context, kPGPInputFormat_X509DataInPKCS7),
	  PGPOInputBuffer(Context, PAnsiChar(CertInput), Length(CertInput))
	]);
    end
    else begin
      Result := PGPBuildOptionList(Context, OptionList,
	[
	  PGPOInputFormat(Context, kPGPInputFormat_PEMEncodedX509Cert),
	  PGPOInputBuffer(Context, PAnsiChar(CertInput), Length(CertInput))
	]);
    end;
    if Result <> 0 then Exit;
    try
      Result := PGPImport(Context, CertImportDB, OptionList, PGPOLastOption(Context));
    finally
      PGPFreeOptionList(OptionList);
      OptionList := NIL;
    end;
  until (Result = 0) or (LoopCount = 2);
  if Result < 0 then Result := kPGPError_X509CertificateParseError;
end;

// the required certificate chain can be created using ExportX509Certificate() with Chain enabled and Armor disabled
function VerifyX509CertificateChain(const CertChain: RawByteString; FromFile: Longbool): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetX509	: pPGPKeySet;
  RootCertsBuf	: RawByteString;
  MostRecentCert: pPGPKeyDBObj;
  CertDataBuf	: RawByteString;
  CertDataDB	: pPGPKeyDB;
begin
  if SDKVersion >= PGP7XX then begin
    Result := InitX509CertKeySet(Context, KeySetMain, KeySetX509);
    if Result <> 0 then Exit;
    try
      Result := CreateCertChainFromKeySet(KeySetX509, RootCertsBuf, MostRecentCert);
      if Result <> 0 then Exit;
      if FromFile then
	Result := ReadDataFromFile(CertChain, CertDataBuf)
      else CertDataBuf := CertChain;
      if Result <> 0 then Exit;
      Result := PGPVerifyX509CertificateChain(Context, PAnsiChar(CertDataBuf), PAnsiChar(RootCertsBuf));
      if Result = kPGPError_X509SelfSignedCert then Exit;
      if Result <> 0 then begin
	if GetX509CertImportDB(Context, CertDataBuf, CertDataDB, FALSE) = 0 then begin
	  Result := CreateCertChainFromKeySet(PGPPeekKeyDBRootKeySet(CertDataDB), CertDataBuf, MostRecentCert);
	  if Result = 0 then Result := PGPVerifyX509CertificateChain(Context, PAnsiChar(CertDataBuf), PAnsiChar(RootCertsBuf));
	end;
      end;
    finally
      FinitX509CertKeySet(KeySetX509);
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// verify certificate chain in given key DB
function VerifyX509CertificateChainDB(CertKeyDB: pPGPKeyDB; var MostRecentCertProps: TX509CertPropsRec): TX509ChainValidity;
var
  CertChain	: RawByteString;
  MostRecentCert: pPGPKeyDBObj;
begin
  Result := CertChain_Invalid;
  if (SDKVersion >= PGP7XX)
  and (CreateCertChainFromKeySet(PGPPeekKeyDBRootKeySet(CertKeyDB), CertChain, MostRecentCert) = 0) then begin
    case PGPVerifyX509CertificateChain(PGPPeekKeyDBContext(CertKeyDB), PAnsiChar(CertChain), PAnsiChar(CertChain)) of
      0: Result := CertChain_Valid;
      kPGPError_X509SelfSignedCert: Result := CertChain_Root;
    end;
    if MostRecentCert <> NIL then GetCertProps(MostRecentCert, MostRecentCertProps);
  end;
end;

// select certificate and return it's properties
function SelectX509CertificateDialog(const CertKeyPrompt: RawByteString;
				     var CertPropsList: TX509CertPropsList;
				     WinHandle: Integer): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  KeySetX509	: pPGPKeySet;
  CertKey	: pPGPKeyDBObj;
  Certificate	: pPGPKeyDBObj;
begin
  if SDKVersion >= PGP7XX then begin
    Result := InitX509CertKeySet(Context, KeySetMain, KeySetX509);
    if Result <> 0 then Exit;
    try
      Result := X509GetKeyDialog(Context, KeySetX509, TRUE, CertKey, CertKeyPrompt, WinHandle);
      if Result <> 0 then Exit;
      if SDKVersion >= PGP81X then
	Certificate := CertKey
      else Certificate := GetValidX509CertFromKey(E, CertKey, NIL, KeySetX509);
      if Certificate <> NIL then
	Result := GetX509CertPropsList(Certificate, CertPropsList)
      else Result := kPGPError_X509NeededCertNotAvailable;
    finally
      FinitX509CertKeySet(KeySetX509);
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// check whether a certificate with the given serial number already exists on the signer's keyring
function IsSerialNumberUnique(Context: pPGPContext; KeySetToCheck: pPGPKeySet; SignKey: pPGPKeyDBObj;
			      const SerialNumber: RawByteString; var Unique: Longbool): Integer;
var
  SignKeyCertSet: pPGPKeySet;
  KeyIter	: pPGPKeyIter;
  Certificate	: pPGPKeyDBObj;
  KeyFound	: pPGPKeyDBObj;
begin
  Result := 0;
  Unique := FALSE;
  if IntToStr(StrToIntDef(SerialNumber, 0)) <> SerialNumber then Exit;
  Result := GetCertKeySetBySignKey(Context, KeySetToCheck, SignKey, TRUE, FALSE, SignKeyCertSet);
  if Result = kPGPError_X509NeededCertNotAvailable then begin
    Unique := TRUE;
    Result := 0;
    Exit;
  end
  else if Result <> 0 then Exit;
  try
    Result := PGPNewKeyIterFromKeySet(SignKeyCertSet, KeyIter);
    if Result <> 0 then Exit;
    try
      Unique := TRUE;
      while (FindCertificate(KeyIter, KeySetToCheck, Certificate, KeyFound) = 0) and Unique do begin
	if (KeyFound = SignKey) and (GetCertPropOwnerSerialNumber(Certificate) = SerialNumber) then Unique := FALSE;
      end;
    finally
      PGPFreeKeyIter(KeyIter);
    end;
  finally
    PGPFreeKeySet(SignKeyCertSet);
  end;
end;

// retrieve properties and IASN of single certificate
function GetX509CertPropsList(Certificate: pPGPKeyDBObj; var CertPropsList: TX509CertPropsList): Integer;
begin
  if CertPropsList = NIL then CertPropsList := TX509CertPropsList.Create;
  if CertPropsList <> NIL then with CertPropsList do begin
    Result := GetCertProps(Certificate, pX509CertPropsRec(Objects[Add(GetCertPropIASN(Certificate))])^);
  end
  else Result := kPGPError_OutOfMemory;
end;

// returns number of certificates found or error
// strings contain the supposedly unique IssuerAndSerialNumber (binary) structure
function GetX509CertSetProps(CertSet, KeySetToSearch: pPGPKeySet; SignKey: pPGPKeyDBObj;
			     const UserID: RawByteString; Filter: DWord; Order: PGPKeyOrdering;
			     var CertPropsList: TX509CertPropsList): Integer;
var
  CertList	: pPGPKeyList;
  CertIter	: pPGPKeyIter;
  IsAddress	: Longbool;
  Certificate	: pPGPKeyDBObj;
  KeyFound	: pPGPKeyDBObj;
  CertIndex	: Integer;
begin
  if SDKVersion >= PGP7XX then begin
    if CertPropsList = NIL then CertPropsList := TX509CertPropsList.Create;
    if CertPropsList = NIL then begin
      Result := kPGPError_OutOfMemory;
      Exit;
    end;
    Result := PGPOrderKeySet(CertSet, Order, PGPFalse, CertList);
    if Result <> 0 then Exit;
    CertIter := NIL;
    try
      Result := PGPNewKeyIter(CertList, CertIter);
      if Result <> 0 then Exit;
      IsAddress := IsEmailAddress(UserID);
      while FindCertificate(CertIter, KeySetToSearch, Certificate, KeyFound) = 0 do begin
	if ((SignKey = NIL) or (KeyFound = SignKey))
	and (
	  (UserID = E) or (GetCertPropOwnerUserID(Certificate) = UserID)
	  or (IsAddress and (GetCertPropOwnerEmail(Certificate) = UserID))
	) then begin
	  CertIndex := CertPropsList.Add(E);
	  if CertIndex = -1 then begin
	    Result := kPGPError_OutOfMemory;
	    Exit;
	  end;
	  Result := GetCertProps(Certificate, pX509CertPropsRec(CertPropsList.Objects[CertIndex])^);
	  if Filter and KeyFilterFlag_X509Root <> 0 then with CertPropsList do begin
	    if not X509CertProps[CertIndex].x509IsRoot then Delete(CertIndex);
	  end
	  else if Filter and KeyFilterFlag_X509CanEncrypt <> 0 then with CertPropsList do begin
	    if not X509CertProps[CertIndex].x509CanEncryptData then Delete(CertIndex);
	  end
	  else if Filter and KeyFilterFlag_X509CanSign <> 0 then with CertPropsList do begin
	    if not X509CertProps[CertIndex].x509CanSignData then Delete(CertIndex);
	  end;
	  with CertPropsList do if CertIndex < Count then Strings[CertIndex] := GetCertPropIASN(Certificate);
	end;
      end;
      if Result = 0 then Result := CertPropsList.Count;
    finally
      PGPFreeKeyIter(CertIter);
      PGPFreeKeyList(CertList);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// returns number of certificates found or error
function GetX509CertificateProperties(const CertData: RawByteString; FromFile: Longbool;
				      var CertPropsList: TX509CertPropsList): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  CertDBToCheck	: pPGPKeyDB;
  CertSetToCheck: pPGPKeySet;
begin
  if SDKVersion >= PGP7XX then begin
    Result := KeyRings.InitKeyRings(Context, KeySetMain);
    if Result <> 0 then Exit;
    try
      Result := GetX509CertImportDB(Context, CertData, CertDBToCheck, FromFile);
      if Result <> 0 then Exit;
      try
	Result := PGPNewKeySet(CertDBToCheck, CertSetToCheck);
	if Result <> 0 then Exit;
	try
	  Result := GetX509CertSetProps(CertSetToCheck, KeySetMain, NIL, E, 0,
					kPGPKeyOrdering_UserID, CertPropsList);
	finally
	  PGPFreeKeySet(CertSetToCheck);
	end;
      finally
	PGPFreeKeyDB(CertDBToCheck);
      end;
    finally
      KeyRings.FreeKeyRings;
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;

// returns number of certificates found or error
// KeyFilterFlags is applied to any kind of search
// SignKeyID can narrow down the search with CertKey...IDs
// IncludeSigner is only valid without CertKey...IDs specified
// in this case you will get the following results depending on SignKeyHexID and IncludeSigner:
// (SignKey = NIL) and (IncludeSigner = FALSE)  => all certified keys on keyring
// (SignKey = NIL) and (IncludeSigner = TRUE)   => all certifier keys on keyring (signing keys only)
// (SignKey <> NIL) and (IncludeSigner = FALSE) => all SignKey-certified keys on keyring without certifier
// (SignKey <> NIL) and (IncludeSigner = TRUE)  => all SignKey-certified keys on keyring including certifier
function FindX509CertProps(const CertKeyUserID, CertKeyHexID, SignKeyHexID: RawByteString; KeyFilterFlags: DWord;
			   IncludeSigner: Longbool; var CertPropsList: TX509CertPropsList): Integer;
var
  Context	: pPGPContext;
  KeySetMain	: pPGPKeySet;
  CertSet	: pPGPKeySet;
  KeyFilter	: pPGPFilter;
  KeySetToCheck	: pPGPKeySet;
  SignKey	: pPGPKeyDBObj;
  CertKey	: pPGPKeyDBObj;
  CertSetBuf	: pPGPKeySet;
  UserIDFilter	: pPGPFilter;
  KeyCount	: PGPUInt32;
begin
  if SDKVersion >= PGP7XX then begin
    Result := KeyRings.InitKeyRings(Context, KeySetMain);
    if Result <> 0 then Exit;
    CertSet := NIL;
    try
      if KeyFilterFlags <> KeyFilterFlag_AllKeys then begin
	Result := GetKeyFilter(Context, KeyFilterFlags, KeyFilter);
	if Result <> 0 then Exit;
	try
	  Result := PGPFilterKeySet(KeySetMain, KeyFilter, KeySetToCheck);
	  if Result <> 0 then Exit;
	finally
	  PGPFreeFilter(KeyFilter);
	end;
      end
      else KeySetToCheck := KeySetMain;
      try
	if SignKeyHexID <> E then begin
	  Result := GetKeyByHexID(KeySetMain, SignKeyHexID, SignKey);
	  if (Result = 0) and not GetKeyPropCanSign(SignKey) then Result := kPGPError_KeyUnusableForSignature;
	  if Result <> 0 then Exit;
	end
	else SignKey := NIL;
	if (CertKeyUserID <> E) or (CertKeyHexID <> E) then begin
	  if CertKeyUserID <> E then begin
	    Result := GetUserIDFilter(Context, CertKeyUserID, TRUE, UserIDFilter);
	    if Result <> 0 then Exit;
	  end
	  else UserIDFilter := NIL;
	  try
	    if CertKeyHexID <> E then begin
	      Result := GetKeyByHexID(KeySetToCheck, CertKeyHexID, CertKey);
	      if Result <> 0 then Exit;
	      Result := PGPNewSingletonKeySet(CertKey, CertSet);
	      if Result <> 0 then Exit;
	      if UserIDFilter <> NIL then begin
		CertSetBuf := CertSet;
		CertSet := NIL;
		try
		  Result := PGPFilterKeySet(CertSetBuf, UserIDFilter, CertSet);
		  if Result <> 0 then Exit;
		finally
		  PGPFreeKeySet(CertSetBuf);
		end;
	      end;
	    end
	    else begin
	      Result := PGPFilterKeySet(KeySetToCheck, UserIDFilter, CertSet);
	      if Result <> 0 then Exit;
	    end;
	  finally
	    PGPFreeFilter(UserIDFilter);
	  end;
	end
	else begin
	  Result := GetCertKeySetBySignKey(Context, KeySetToCheck, SignKey, IncludeSigner, FALSE, CertSet);
	  if Result <> 0 then Exit;
	  Result := PGPCountKeys(CertSet, KeyCount);
	  if (Result <> 0) or (KeyCount = 0) then Exit;
	end;
	Result := GetX509CertSetProps(CertSet, KeySetMain, SignKey, CertKeyUserID,
				      KeyFilterFlags and KeyFilterMask_X509Props,
				      kPGPKeyOrdering_UserID, CertPropsList);
      finally
	if KeySetToCheck <> KeySetMain then PGPFreeKeySet(KeySetToCheck);
      end;
    finally
      PGPFreeKeySet(CertSet);
      KeyRings.FreeKeyRings;
      Result := TranslateX509ErrorToPGPError(Result);
    end;
  end
  else Result := kPGPError_LazyProgrammer;
end;


end.

